
//*******************************************************************
//
//  UNIT:     B l o b s                        03.08.2001, 04.10.2002
//
//  animated 2d-Blobs demo unit, coded by Gerd Platl
//
//  references:
//    http://www.delphi3d.net/articles/printarticle.php?article=metaballs.htm
//    http://www.geisswerks.com/ryan/BLOBS/blobs.html
//    http://www.uni-koblenz.de/~cohnen/applets/ccblob.htm
//    http://www.cfxweb.net/~aggrav8d/tutorials/metaballs.html#top
//    http://www.l-shaped.freeserve.co.uk/computing/blobs/
//    http://flipcode.com/documents/metaball.txt
//    http://www.angelcode.com/home/tutorials/metaballs.asp
//
//*******************************************************************
{$I comp_opt.inc}

unit Blobs;

{$ifopt D+}
  {$define xtest1    main functions}
  {$define xtest2    sub functions}
{$endif}

interface

//*********************************************************
implementation

uses  SysUtils, Windows, Classes, Graphics,
      SysUtil2, Graph, Images, Palette, Values, Vector3d, ColorRGB, Bounce3d,
      GrForm;

type TBall3d = object(TBounce3d)
       ballColor: tColor;        // Position, Richtung
       ballRadius: single;
       ballFaktor: single;       // 1 / radius^2
       constructor Init;
     end;

{*********************************************************}
constructor TBall3d.Init;

begin
  inherited;
  ballColor := RandomColor (60,255);
  ballRadius := 12;
  ballFaktor := 1 / sqr (ballRadius);
end;
{*********************************************************}

const modes = 2;                       // # of demo modes
      maxFkt = 3;                      // # of blob functions
      maxBlobs = 10;                   // max. blobs

var   modeNo:      word = 0;           // actual demo mode
      fktNo:       word = 1;           // actual blob function
      Blob: array [0..maxBlobs] of TBall3d;
      actBlobs:    word = maxBlobs div 2;  // active blobs
      speed:       float = 5.0;
      factor:      float = 16.0;
      attraction:  v3d;

var   attrVec: v3d = (x:0; y:0.1; z:0);

const intMapSize = 128;
      halfMapSize = intMapSize div 2;
var   intMap: tBitmap = nil;   // original background picture

const tileMapSize = 256;       // draw map for a tile
var   tileMap: tBitmap = nil;

type  tBlobFunction = function (x,y: float): float;
var   BlobFktCall: tBlobFunction;

//=========================================================
// blob function result: 0 .. maxVal
//=========================================================

const maxVal = 64;

//*********************************************************
// blob function:  f(x,y) = 1 - (d / (1 + rq))
//                      d = (x^2 + y^2)
//---------------------------------------------------------
function BlobFunction1 (x,y: float): float;
var  d: float;
begin
  d := (sqr(x) + sqr(y)) / factor;
  Result := maxVal * (1 - (d / (1+d)));
end;

//*********************************************************
// blob function:  f(x,y) = 0.4 / sqrt(x^2 + y^2)
//---------------------------------------------------------
function BlobFunction2 (x,y: float): float;
var  d: float;
begin
  d := (sqr(x) + sqr(y)) / factor;
  if d > 0.01 then begin
    d := maxVal * (0.4 / sqrt(d));
    if d > maxVal
    then Result := maxVal
    else Result := d;
  end
  else Result := maxVal;
end;
//*********************************************************
// blob function:  f(x,y) = (r^4 - r^2 + 0.25)
//---------------------------------------------------------
function BlobFunction3 (x,y: float): float;
var  d: float;
begin
  d := (sqr(x) + sqr(y)) / factor;
  Result := maxVal * (1+cos(d)) * (1 - (d / (1+d)));
end;
//*********************************************************
// create draw map for a tile
//---------------------------------------------------------
procedure CreateTileMap;
begin
  {$ifdef test2}
    writeln (tst, 'Blobs.CreateTileMap: ');
  {$endif}
  tileMap := tBitmap.Create;
  tileMap.PixelFormat := pf8Bit;
  tileMap.Width  := tileMapSize;
  tileMap.Height := tileMapSize;
end;
//*********************************************************
// create map with precalculated blob values
//---------------------------------------------------------
procedure CreateBlobMap;
var   intPalette: tPalette;    // intesity palette
begin
  {$ifdef test2}
    writeln (tst, 'Blobs.CreateBlobIntensityMap: ');
  {$endif}
  intMap := tBitmap.Create;
  intMap.PixelFormat := pf16Bit;
  intMap.Width  := intMapSize;
  intMap.Height := intMapSize;
  SetIntensityPalette (intPalette);
  SetBitmapPalette (intMap, intPalette);
  SetBitmapPalette (tileMap, intPalette);
end;
//*********************************************************
// calculate blob map with active blob function
//---------------------------------------------------------
procedure CalcBlobMap;
var   xi,yi, baseVal, val: integer;
      xf, yf: float;
      valPtr: ^smallint;
begin
  {$ifdef test2}
    writeln (tst, 'Blobs.CalcBlobMap: ');
  {$endif}

  baseVal := smallint (trunc(BlobFktCall (0, halfMapSize)));
  for yi := 0 to intMapSize - 1
  do begin
    yf := yi - halfMapSize;
    valPtr := intMap.ScanLine[yi];
    for xi := 0 to intMapSize - 1
    do begin
      xf := xi - halfMapSize;
      val := smallint (trunc(BlobFktCall (xf, yf))) - baseVal;
      if val > 0
      then valPtr^ := val
      else valPtr^ := 0;
      inc (valPtr);
    end;
  end;
end;
//*********************************************************
// init Blobs
//---------------------------------------------------------
procedure InitBlobs;
var   bi: integer;
begin
  {$ifdef test1}
    writeln (tst, 'Blobs.InitBlobs: ');
  {$endif}

  BlobFktCall := BlobFunction1;
  CreateTileMap;
  CreateBlobMap;
  CalcBlobMap;

  // init blobs
  for bi := 0 to maxBlobs
  do with Blob[bi] do begin
    Init;
    attraction := attrVec;
  end;
end;
//*********************************************************
// change Blobs
//---------------------------------------------------------
procedure ChangeBlobs (randomPos, randomDir: Boolean);
var   bi: integer;
      maxValue: v3d;
      moveEndless: boolean;
begin
  moveEndless := (modeNo = 2);
  if moveEndless
  then begin
    maxValue.x := tileMapSize;
    maxValue.y := tileMapSize;
    maxValue.z := tileMapSize;
  end
  else begin
    maxValue.x := pageX;
    maxValue.y := pageY;
    maxValue.z := 10000;
  end;
  for bi := 0 to maxBlobs
  do with Blob[bi] do begin
    attraction := attrVec;
    ballColor := RandomColor (160,255);
    if randomPos then Range (zeroVector3d, maxValue);
    if randomPos then RandomPosition;
    if randomDir then RandomDirection (-speed, speed);
  end;
end;

{********************************************************}
procedure CreateValues;
begin
  with valuePtr^
  do if ValueCount = 0
  then begin
    AppendValue ('blobs', 2, maxBlobs);
    AppendValue ('speed', 10, 100);
    AppendValue ('attraction', 0, 100);
    AppendValue ('factor', 64, 512);
  end;
end;
{********************************************************}
procedure ResetValues;
begin
  valuePtr^.Reset;
  ChangeBlobs (true, true);
end;
//*********************************************************
procedure RandomValues;
begin
  valuePtr^.Reset;
  valuePtr^.RandomValues;
  ChangeBlobs (true, true);
  // MarkTime;
end;

//*********************************************************
procedure NextMode;
begin
  modeNo := 1 + modeNo mod modes;
  ChangeBlobs (true, true);
  MarkTime;
end;

//*********************************************************
procedure NextBlobFunction;
begin
  fktNo := 1 + fktNo mod maxFkt;
  case fktNo of
    1: BlobFktCall := BlobFunction1;
    2: BlobFktCall := BlobFunction2;
    3: BlobFktCall := BlobFunction3;
  end;
end;

//*********************************************************
// create random color palette
//---------------------------------------------------------
procedure RandomColors;

var   intPalette: tPalette;    // intesity palette
      //col: tColor;
begin
//  SetLayerPalette (intPalette, clYellow, 32);
  SetSineWavePalette (intPalette, 0, 200+Random(1250)
                                   , 200+Random(1250)
                                   , 200+Random(1250));
(*
  // for testing: red means no influence
  col := clRed;
  intPalette[0] := tCol(col);
*)
  SetBitmapPalette (bgBitmap, intPalette);
  SetBitmapPalette (tileMap, intPalette);
end;
//*********************************************************
// start new animation
//---------------------------------------------------------
procedure AnimStart;
begin
  {$ifdef test1}
    writeln (tst, #13#10'Blobs.AnimStart: ', modeNo);
  {$endif}
  CreateValues;
//  SetGraphMode (pf32bit);
  SetGraphMode (pf8bit);
  RandomColors;
  NextMode;
end;
//*********************************************************
// resize animation window
//---------------------------------------------------------
procedure AnimResize;
begin
  {$ifdef test1}
    writeln (tst, 'Blobs.AnimResize: ', pageX, '*', pageY);
  {$endif}
  ChangeBlobs (true, true);
end;

//*********************************************************
// example of bouncing blobs with simple distance calculation
// its simple code but is too slow for an animation !
//---------------------------------------------------------
procedure SimpleBlobs;
var   xi,yi, bi: integer;
      val: float;
begin
  BlobFktCall := BlobFunction3;
  ClearViewport;
  for yi := 0 to pageY - 1
  do begin
    for xi := 0 to pageX - 1
    do begin
      val := 0.01;
      for bi := 1 to actBlobs do with blob[bi]
      do val := val + BlobFktCall (pos.x - xi, pos.y - yi);
      PutPixel (xi,yi, byte(trunc(val)));
    end;
  end;

  // move blobs
  for bi := 1 to actBlobs do blob[bi].Move;
end;

//*********************************************************
// draw blob at position xp,yp to active map
//---------------------------------------------------------
procedure DrawBlob (xp,yp: integer);
var   xi,yi, val, xs: integer;
      valPtr: ^smallint;
      drawPtr: pByteArray;
begin
  xs := xp - halfMapSize;
  yp := yp - halfMapSize;
  for yi := 0 to intMapSize - 1
  do begin
    if yp >= pageY then Break;
    if yp >= 0
    then begin
      valPtr := intMap.ScanLine[yi];
      drawPtr := (activeBitmap.ScanLine[yp]);
      xp := xs;
      for xi := 0 to intMapSize - 1
      do begin
        if xp >= 0
        then begin
          if xp >= pageX then Break;
          val := valPtr^;
          if val > 0
          // AddPixel (xp, yp, val);
          then drawPtr^[xp] := byte(drawPtr^[xp] + val);
        end;
        inc (valPtr);
        inc (xp);
      end;
    end;
    inc (yp);
  end;
end;

//*********************************************************
// draw blob at position xp,yp to tileMap
//---------------------------------------------------------
procedure DrawTileBlob (xp,yp: integer);
var   xi,yi, val, xs: integer;
      valPtr: ^smallint;
      drawPtr: ^byte;
      linePtr: pByteArray;
begin
  xs := xp - halfMapSize;
  yp := yp - halfMapSize;
  if xs < 0 then xs := xs + tileMapSize;
  if yp < 0 then yp := yp + tileMapSize;
  for yi := 0 to intMapSize - 1
  do begin
    valPtr := intMap.ScanLine[yi];
    if yp >= tileMapSize then yp := 0;
    linePtr := (tileMap.ScanLine[yp]);
    xp := xs;
    drawPtr := @linePtr^[xs];
    for xi := 0 to intMapSize - 1
    do begin
      if xp >= tileMapSize
      then begin
        xp := 0;
        drawPtr := @linePtr^[0];
      end;
      val := valPtr^;
      if val > 0
      then drawPtr^ := byte(drawPtr^ + val);
      inc (valPtr);
      inc (drawPtr);
      inc (xp);
    end;
    inc (yp);
  end;
end;

//*********************************************************
// example of bouncing blobs with overlapping algorithm
//---------------------------------------------------------
procedure BlobsDemo1;
var   bi: integer;
begin
  ClearViewport;
  for bi := 1 to actBlobs do with blob[bi]
  do begin
    Move;
    DrawBlob (trunc(pos.x), trunc(pos.y));
  end;
end;

//*********************************************************
// example of bouncing blobs with overlapping algorithm
//---------------------------------------------------------
procedure BlobsDemo2;
var   bi: integer;
   //   tileBm: tBitmap;
begin
  ClearViewport;
  BlackBitmap (tileMap);
  for bi := 1 to actBlobs do with blob[bi]
  do begin
    Move;
    DrawTileBlob (trunc(pos.x), trunc(pos.y));
  end;

  mainForm.bufferedDrawing := false;
//  mainForm.imageBm.Canvas.Draw (0,0, tileMap);
  DrawTilesBackground (0,0, tileMap, mainForm.imageBm);
(*
  tileBm := tBitmap.Create;
  tileBm.PixelFormat := screen.pixelFormat;
  tileBm.Height := intMap.Height;
  tileBm.Width  := intMap.Width;
  tileBm.Canvas.Draw (0,0, intMap);
  DrawTilesBackground (0,0, tileBm, bgBitmap);
  tileBm.Free;
*)
end;

//*********************************************************
// draw one animation picture
//---------------------------------------------------------
procedure AnimDraw;
var   bi: integer;
      valPtr: ^smallint;
begin
  {$ifdef test2}
    writeln (tst, 'Blobs.AnimDraw: ', modeNo);
  {$endif}

  CalcBlobMap;  //??? nur fr test

  case modeNo of
    1: BlobsDemo1;
    2: BlobsDemo2;
//    3: SimpleBlobs;
  end;

  // draw middle blob values
  if demoHelp then begin
    valPtr := intMap.ScanLine[halfMapSize];
    for bi := 1 to intMapSize
    do begin
      PutPixel (30+bi, pageY - 10, 128);   // draw max value
      PutPixel (30+bi, pageY - 10 - valPtr^, 100);
      inc (valPtr);
    end;
  end;
end;

//*********************************************************
// handle animation input
//---------------------------------------------------------
procedure AnimInput;
var  sps, avs: float;
begin
  {$ifdef test2}
    writeln (tst, 'Ripple.AnimInput: ', ord(cb));
  {$endif}

  with valuePtr^
  do begin
    sps := speed;
    avs := attrVec.y;
    actBlobs  := IntValue (1);
    speed     := IntValue (2) / 40.0;
    attrVec.y := IntValue (3) / 500.0;
    factor    := IntValue (4);
  end;

  if sps <> speed          Then ChangeBlobs (false, false)
  else if avs <> attrVec.y Then ChangeBlobs (false, false);

  case cb of
    CR : NextMode;
    'r': ResetValues;
    'n': begin   RandomValues;   RandomColors;   end;
    'c': RandomColors;
    'f': NextBlobFunction;
  end;
end;

//*********************************************************
// display animation specific help
//---------------------------------------------------------
procedure AnimHelp;
begin
  HelpText (1, '<ret>  mode ' + IntToStr(modeNo));
  HelpText (2, '   r   reset values');
  HelpText (3, '   n   random values');
  HelpText (4, '   c   random colors');
  HelpText (5, '   f   blob function ' + IntToStr(fktNo));
end;

//*********************************************************
initialization

  {$ifdef InitTest}
    writeln (tst, 'Blobs.initialization: ');
    flush (tst);
  {$endif}

  InitBlobs;

  RegisterAnimation (' Blobs 2d ',
    AnimStart, AnimResize, AnimDraw, AnimInput, AnimHelp);

//---------------------------------------------------------
finalization

  {$ifdef ExitTest}
    writeln (tst, 'Blobs.finalization: ');
    flush (tst);
  {$endif}

end.

