{*******************************************************************}
{*                                                                 *}
{*  UNIT:          B o u n c e 3 D          16.04.98,  12.02.2002  *}
{*                                                                 *}
{*  move points within given 3d room                               *}
{*  coded by Gerd Platl                                            *}
{*                                                                 *}
{*  Methoden:                                                      *}
{*  -------------------------------------------------------------- *}
{*  Init                Objekt initialisieren                      *}
{*  Range               Quaderdimensionen festlegen                *}
{*  RandomPosition      zufllige Position im Quader setzen        *}
{*  RandomDirection     zufllige Richtung einschlagen             *}
{*  Size                Position im Quader verndern               *}
{*  Move                Position im Quader verndern               *}
{*                                                                 *}
{*******************************************************************}
{$I comp_opt.inc}

unit Bounce3d;

interface

uses  Vector3d, Geom3d;

type TBounce3d = object
       pos, dir: v3d;        {Position, Richtung}
       minL, maxL: v3d;      {min/max limits}
       rng: v3d;             {Bereich: max - min}
       objSize: v3d;         {Objektgre x*y*z}
       endless: boolean;
       attraction: v3d;      {Anziehungskraft}
       useAttr: boolean;
       constructor Init;
       procedure Range (minValue, maxValue: v3d);
       procedure RandomPosition;
       procedure RandomDirection (dmin,dmax: float);
       procedure Size (objectSize: v3d);
       procedure Move;
     end;

implementation
{*********************************************************}
constructor TBounce3d.Init;

begin
  pos := zeroVector3d;
  dir := zeroVector3d;
  minL.x := -1000;
  minL.y := -1000;
  minL.z := -1000;
  maxL.x := +1000;
  maxL.y := +1000;
  maxL.z := +1000;
  endless := false;        // bounce from wall
  attraction := zeroVector3d;
  useAttr := false;
end;
{*********************************************************}
procedure TBounce3d.Range (minValue, maxValue: v3d);

begin
  minL := minValue;
  maxL := maxValue;
  VectorDiff3d (rng, maxL, minL);
  pos := zeroVector3d;
end;
{*********************************************************}
procedure TBounce3d.RandomPosition;

begin
  pos.x := minL.x + random (trunc(rng.x));
  pos.y := minL.y + random (trunc(rng.y));
  pos.z := minL.z + random (trunc(rng.z));
end;
{*********************************************************}
procedure TBounce3d.RandomDirection (dmin,dmax: float);

var   r: float;
begin
  r := dmax - dmin;
  dir.x := dmin + r * random;
  dir.y := dmin + r * random;
  dir.z := dmin + r * random;
end;
{*********************************************************}
procedure TBounce3d.Size (objectSize: v3d);

begin
  objSize := objectSize;
end;
{*********************************************************}
procedure FlipMax (var lim, pos, dir: float);

begin
  pos := 2*lim - pos;
  dir := -dir;
end;
{---------------------------------------------------------}
procedure FlipMin (var lim, pos, dir: float);

begin
  pos := 2*lim - pos;
  dir := -dir;
end;
{*********************************************************}
procedure TBounce3d.Move;

var   minM, maxM: v3d;
begin
  if useAttr then dir.y := dir.y + attraction.y;
  VectorAdd3d (pos, dir);

  if endless      // appear on the other site on outer limit
  then begin
    if      pos.x > maxL.x then pos.x := minL.x
    else if pos.x < minL.x then pos.x := maxL.x;
    if      pos.y > maxL.y then pos.y := minL.y
    else if pos.y < minL.y then pos.y := maxL.y;
    if      pos.z > maxL.z then pos.z := minL.z
    else if pos.z < minL.z then pos.z := maxL.z;
  end
  else begin
    VectorSum3d (minM, minL, objSize);  // reduce limits with
    VectorSum3d (maxM, maxL, objSize);  //   object dimension
    if      pos.x > maxM.x then FlipMax (maxM.x, pos.x, dir.x)
    else if pos.x < minM.x then FlipMin (minM.x, pos.x, dir.x);
    if      pos.y > maxM.y then FlipMax (maxM.y, pos.y, dir.y)
    else if pos.y < minM.y then FlipMin (minM.y, pos.y, dir.y);
    if      pos.z > maxM.z then FlipMax (maxM.z, pos.z, dir.z)
    else if pos.z < minM.z then FlipMin (minM.z, pos.z, dir.z);
  end;
end;

end.

