
//*******************************************************************
//
//  UNIT:     B u m p L i                        06.08.99, 05.06.2003
//
//  bump lightning demo, coded by Gerd Platl
//
//*******************************************************************
{$I 'comp_opt.inc'}

unit BumpLi;

{$ifopt D+}
  {$define test1    main functions}
  {$define test2    sub functions}
{$endif}

interface

//*********************************************************
implementation

uses  SysUtils, Windows, Graphics,
      SysUtil2, Graph, ColorRGB, Palette, Images, Values,
      GrForm, FractMap, Dance, Blur;

const modes  = 3;            // # of lightning modes
      minLI  = 33;           // minimum of light intensity
      maxLI  = 215;          // maximum of light intensity

var   modeNo:      word    =  1;       // actual demo mode
      lightNo:     word    =  1;       // actual light source
      moveSpeed:   integer = 30;       // moving speed
      moveRadius:  integer = 70;       // moving radius
      lightInt:    integer = 16;       // light intensity
      roughness:   integer = 40;       // map roughness: 40..80

      heightMap:   tBitmap = nil;      // 2^n maps
      colorMap:    tBitmap = nil;

      heightBm:    tBitmap = nil;      // height bitmap (0..255)
      colorBm:     tBitmap = nil;      // color index bitmap (0..255)
      palIntBm:    tBitmap = nil;      // palette color intensity map
                                       // colors 0..255, shades 0..255
      palColPtr:   pLongIntArray;      // palette intensity map start pointer
      palIntPtr:   array[byte] of pLongIntArray;
      showMap:     boolean = false;

      lsPal: tPalette;                 // landscape palette

{********************************************************}
procedure GetValues;
begin
  with valuePtr^ do begin
    moveSpeed  := IntValue (1);
    moveRadius := IntValue (2);
    lightInt   := IntValue (3);
    roughness  := IntValue (4);
  end;
end;
{********************************************************}
procedure CreateValues;
begin
  with valuePtr^
  do if ValueCount = 0
  then begin
    AppendValue ('moving speed',      10,  70);
    AppendValue ('moving radius',     20, 160);
    AppendValue ('light intensity',    8,  48);
    AppendValue ('fractal dimension', 50,  70);
  end;
  GetValues;
end;

//*********************************************************
// start new animation
//---------------------------------------------------------
procedure AnimStart;
begin
  {$ifdef test1}
    writeln (tst, #13#10'BumpLi.AnimStart: ');
  {$endif}

  CreateValues;
  SetGraphMode (pf32bit);

  modeNo := 1 + modeNo mod modes;
  SelectMap (heightMap, colorMap);
  GenerateMap (roughness);
end;

//*********************************************************
// create intensity color bitmap with
// and pointer array for each scan line
//---------------------------------------------------------
procedure CreatePaletteIntensityMap;
var   yi: longint;
begin

  {$ifdef test1}
    writeln (tst, 'BumpLi.CreatePaletteIntensityMap: '
    , HexString(heightMap), ', ', HexString(colorMap));
  {$endif}

  GetBitmapPalette (colorMap, lsPal);
  MakePalIntensityMap (palIntBm, lsPal);
  // set pointer to begin of palette intensity map
  palColPtr := palIntBm.ScanLine[255];

  // calculate scanline pointer
  for yi := 0 to 255
  do palIntPtr[yi] := palIntBm.ScanLine[yi];
end;
//*********************************************************
// create / resize height and color bitmaps
//---------------------------------------------------------
procedure AnimResize;
begin
{$ifdef test1}
  writeln (tst, 'BumpLi.ResizeBumpLightning: ', pageX, '*', pageY);
{$endif}

  // create heightBm with overscan window size
  // and fill map with heightMap tiles
  if heightBm = nil
  then begin
    heightBm := tBitmap.Create;
    heightBm.PixelFormat := pf8Bit;
  end;
  heightBm.Width  := pageX + 2;
  heightBm.Height := pageY + 2;
  SetBitmapPalette (heightBm, lsPal);
  DrawTilesBackground (0,0, heightMap, heightBm);

  // create colorBm with overscan window size
  // and fill map with colorMap tiles
  if colorBm = nil
  then begin
    colorBm := tBitmap.Create;
    colorBm.PixelFormat := pf8Bit;
  end;
  colorBm.Width  := (pageX + 8) and $FFFFFFF8;    // raster step to 8 pixel
  colorBm.Height := pageY + 2;
  SetBitmapPalette (colorBm, lsPal);
  DrawTilesBackground (0,0, colorMap, colorBm);
end;
//*********************************************************
// Bump-Lightning Animation 1
// uses only light direction
// input:      lx,ly: light position
//             sm: height bitmap
//             cm: color bitmap
//             dm: destination bitmap
//---------------------------------------------------------
procedure BL1Anim (lx,ly: integer; var sm,cm,dm: tBitmap);

var   xi,yi,                      // pixel position
      li: longint;                // light intensity
      hLeft,hMid,hRight: byte;    // actual line heights
      h1, h2, h3: ^byte;          // height line pointer
      cip: ^byte;                 // color index pointer
      pix: ^tColor;               // pointer to color pixel
begin
{$ifdef test2}
  writeln (tst, ' BumpLi.BL1Anim:');
{$endif}

  // calculate light position
  lx := round(cos(picNo * moveSpeed / 256) * lightInt);
  ly := round(sin(picNo * moveSpeed / 256) * lightInt);

  for yi := 0 to GetMaxY
  do begin
    h1 := sm.ScanLine[yi];   // set height line pointer
    h2 := sm.ScanLine[yi+1];
    h3 := sm.ScanLine[yi+2];
    hLeft  := h2^;
    hMid   := h2^;
    hRight := h2^;
    cip := cm.ScanLine[yi];  // ptr to color bitmap line
    pix := dm.ScanLine[yi];  // ptr to destination bitmap pixel
    for xi := 0 to GetMaxX
    do begin
      // calculate light intensity
      li := 160 - ly*(h1^ - h3^)
                - lx*(hRight - hLeft);
      // set pixel color to intensity palette color
//      pix^ := palColPtr^[cip^ shl 8 + Clamp (li, minLi, maxLi)];
      if      li > maxLI then pix^ := palColPtr^[cip^ shl 8 + maxLI]
      else if li < minLI then pix^ := palColPtr^[cip^ shl 8 + minLI]
      else                    pix^ := palColPtr^[cip^ shl 8 + li];

      inc (cip);             // inc color index source pointer
      inc (pix);             // inc pixel pointer
      inc (h1);              // inc source pointer
      inc (h2);
      inc (h3);

      hLeft  := hMid;        // shift pixel value
      hMid   := hRight;
      hRight := h2^;
    end;
  end;
  // draw light position if demo help is on
  if demoHelp
  then PutPixel (pageX div 2 + lx, pageY div 2 - ly, clYellow);
end;

//*********************************************************
// Bump-Lightning Animation 2
// uses online calculated light intensity
// input:      sm: height bitmap
//             cm: color bitmap
//             dm: destination bitmap (24bit)
//---------------------------------------------------------
procedure BL2Anim (lx,ly: integer; var sm,cm,dm: tBitmap);

var   vx,vy,                      // distance to light
      xi,yi,                      // pixel position
      li: integer;                // light intensity
      hLeft,hMid,hRight: byte;    // actual line heights
      h1, h2, h3: ^byte;          // height line pointer
      cip: ^byte;                 // color index pointer
      pix: ^longInt;              // color pixel pointer
begin
{$ifdef test2}
  writeln (tst, ' BumpLi.BL2Anim:');
{$endif}

  vy := -ly;                 // init y-distance to light

  palColPtr := palIntBm.ScanLine[255];
  for yi := 0 to GetMaxY
  do begin
    h1 := sm.ScanLine[yi];   // set height line pointer
    h2 := sm.ScanLine[yi+1];
    h3 := sm.ScanLine[yi+2];
    vx := -lx;
    hLeft  := h2^;
    hMid   := h2^;
    hRight := h2^;
    cip := cm.ScanLine[yi];
    pix := dm.ScanLine[yi];
    for xi := 0 to GetMaxX
    do begin
      // calculate light intensity
      li := 255 - (abs(vy + (h1^ - h3^) shl 4)
                  +abs(vx + (hRight - hLeft) shl 4)) shl 5 div lightInt;
      if li < minLI then li := minLI;

      // set pixel color to intensity palette color
//ToDo: correct this bug
(*
write (tst, '  li=',li:3, cip^:4);
writeln (tst, '  colorIndex=', word(palIntPtr[cip^]^));
flush (tst);
*)
{$R-,Q-}
      pix^ := palColPtr^[cip^ shl 8 + li];
//      pix^ := palIntPtr[cip^]^[li];
      inc (cip);             // inc color index source pointer
      inc (pix);             // inc pixel pointer
      inc (vx);              // update y-distance
      inc (h1);              // inc source pointer
      inc (h2);
      inc (h3);

      hLeft  := hMid;        // shift pixel value
      hMid   := hRight;
      hRight := h2^;
    end;
    inc(vy);                 // update y-distance
  end;
end;

//*********************************************************
// Bump-Lightning Animation 3
// uses precalculated intensity light map
// input:      sm: height bitmap
//             cm: color bitmap
//             dm: destination bitmap
//---------------------------------------------------------
procedure BL3Anim (lx,ly: integer; var sm,cm,dm: tBitmap);
var
    xi,yi,                   // pixel position
    li: longint;             // light intensity
    hLeft,hMid,hRight: byte; // actual line heights
    h1, h2, h3: ^byte;       // height line pointer
    cip: ^byte;              // color index pointer
    lp: ^byte;               // light intensity pointer
    pix: ^tColor;            // color pixel pointer
begin
{$ifdef test2}
  writeln (tst, ' BumpLi.BL3Anim:');
{$endif}

  for yi := 0 to GetMaxY
  do begin
    h1 := sm.ScanLine[yi];   // set height line pointer
    h2 := sm.ScanLine[yi+1];
    h3 := sm.ScanLine[yi+2];
    hLeft  := h2^;
    hMid   := h2^;
    hRight := h2^;
    cip := cm.ScanLine[yi];
    pix := dm.ScanLine[yi];
    lp := LMLinePtr (yi + pageY div 2 - ly);
    inc (lp, pageX div 2 - lx);
    for xi := 0 to GetMaxX
    do begin
      // calculate light intensity
      li := lp^ * lightInt shr 5
          - 16 * ((h1^ - h3^)+(hRight - hLeft));
      if      li > maxLI then li := maxLI
      else if li < minLI then li := minLI;

      // set pixel color to intensity palette color
      pix^ := palColPtr^[cip^ * 256 + li];
//      pix^ := palIntPtr[cip^]^[li];      // set pixel
      inc (cip);             // inc color index source pointer
      inc (pix);             // inc pixel pointer
      inc (h1);              // inc source pointer
      inc (h2);
      inc (h3);
      inc (lp);              // intensity map pointer

      hLeft  := hMid;        // shift pixel value
      hMid   := hRight;
      hRight := h2^;
    end;
  end;
end;
//*********************************************************
// Bump-Lightning Animation 4
// uses online calculated light intensity
// input:      sm: height bitmap
//             cm: color bitmap
//             dm: destination bitmap (24bit)
//---------------------------------------------------------
procedure BL4Anim (lx,ly: integer; var sm,cm,dm: tBitmap);
var
    xi,yi,                   // pixel position
    li: longint;             // light intensity
    hLeft,hMid,hRight: byte; // actual line heights
    h1, h2, h3: ^byte;       // height line pointer
    cip: ^byte;              // color index pointer
    lp: ^byte;               // light intensity pointer
    pix: ^tColor;            // color pixel pointer
begin
{$ifdef test2}
  writeln (tst, ' BumpLi.BL4Anim:');
{$endif}

  for yi := 0 to GetMaxY
  do begin
    h1 := sm.ScanLine[yi];   // set height line pointer
    h2 := sm.ScanLine[yi+1];
    h3 := sm.ScanLine[yi+2];
    hLeft  := h2^;
    hMid   := h2^;
    hRight := h2^;
    cip := cm.ScanLine[yi];
    pix := dm.ScanLine[yi];
    lp := LMLinePtr ((yi + pageY div 2 - ly) and $FF);
    inc (lp, pageX div 2 - lx);
    for xi := 0 to GetMaxX
    do begin
      // calculate light intensity
      li := ((lp^ + (h1^ - h3^) shl 3)
            +lp^ + (hRight - hLeft) shl 3)  shl 5 div lightInt;
      if      li > maxLI then li := maxLI
      else if li < minLI then li := minLI;

      // set pixel color to intensity palette color
      pix^ := palColPtr^[cip^ shl 8 + li];
//      pix^ := palIntPtr[cip^]^[li];
      inc (cip);             // inc color index source pointer
      inc (pix);             // inc pixel pointer
      inc (h1);              // inc source pointer
      inc (h2);
      inc (h3);
      inc (lp);              // intensity map pointer

      hLeft  := hMid;        // shift pixel value
      hMid   := hRight;
      hRight := h2^;
    end;
  end;
end;
//*********************************************************
// create new maps
//---------------------------------------------------------
procedure NewMap;
begin
  GenerateMap (roughness);
  AnimResize;
end;
//*********************************************************
procedure FractalMap;
begin
  GenerateFractal (modeNo);
  AnimResize;
end;
//*********************************************************
procedure SmoothHeightMap;
begin
  Blur4C (heightBm, heightBm, 0,heightBM.height-1);
end;
//*********************************************************
// draw one animation picture
//---------------------------------------------------------
procedure AnimDraw;

var   lx,ly: integer;        // light position
begin
  {$ifdef test2}
    writeln (tst, 'BumpLi.AnimDraw: ');
  {$endif}

  // calculate light position
  lx := pageX div 2 + round(cos(picNo*moveSpeed / 256) * moveRadius);
  ly := pageY div 2 + round(sin(picNo*moveSpeed / 200) * moveRadius);

  case modeNo of
    1: BL1Anim (lx,ly, heightBm, colorBm, bgBitmap);
    2: BL2Anim (lx,ly, heightBm, colorBm, bgBitmap);
    3: BL3Anim (lx,ly, heightBm, colorBm, bgBitmap);
    4: BL4Anim (lx,ly, heightBm, colorBm, bgBitmap);
  end;
  if showMap then DrawBitmap (0,0, palIntBm, bgBitmap);
end;

//*********************************************************
// handle animation input
//---------------------------------------------------------
procedure AnimInput;
begin
  {$ifdef test2}
    writeln (tst, 'BumpLi.AnimInput: ', ord(cb));
  {$endif}

  GetValues;

  case cb of
    CR : modeNo := 1 + modeNo mod modes;
    'r': valuePtr^.Reset;
    'n': valuePtr^.RandomValues;
    '#': lightNo := UseLightmap (lightNo + 1);
    'a': FractalMap;
    'm': NewMap;
    'o': SmoothHeightMap;
    'R': begin LoadMap ('BumpLi'); AnimResize; end;
    'W': SaveMap ('BumpLi');
    'M': showMap := not showMap;
  end;
end;
//*********************************************************
// display animation specific help
//---------------------------------------------------------
procedure AnimHelp;
begin
  HelpText (1, ' <ret>  effect ' + IntToStr(modeNo));
  case modeNo of
    1:   HelpText (2, '  directonal light:');
    2,4: HelpText (2, '  fixed light map:');
    3:   HelpText (2, '  # light map ' + IntToStr(lightNo));
  end;
  HelpText (3, '  a   apple map');
  HelpText (4, '  n   random values');
  HelpText (5, '  m   create new map');
  HelpText (6, '  o   smooth heights');
  HelpText (7, ' R/W  read/write map');
end;
//*********************************************************
initialization

  FractMap.CreateMap (heightMap, colorMap);

  {$ifdef InitTest}
    writeln (tst, 'BumpLi.initialization: ');
  {$endif}

  CreatePaletteIntensityMap;

  RegisterAnimation (' Bump Lightning ',
    AnimStart, AnimResize, AnimDraw, AnimInput, AnimHelp);

//---------------------------------------------------------
finalization

  {$ifdef ExitTest}
    writeln (tst, 'BumpLi.finalization: ');
    flush (tst);
  {$endif}

  DeleteMap (heightMap, colorMap);
end.

