
//*********************************************************
//
//  UNIT:     C o l o r R G B          04.04.95, 17.06.2003
//
//  color definitions, routines and conversion for Delphi-5/6
//  coded by Gerd Platl
//
//---------------------------------------------------------
//  WinColor         convert arguments to tColor value
//  Color2RGB        convert color value -> RGB value
//  ShadeColor       return color with given intensity
//  MixColor         mix 2 colors with given intensity
//  RandomColor      return random color
//  GrayValue        return gray value of color
//
//  LongColor        convert color -> long color
//  LongColor2Color  convert long color -> color
//
//  RGBRealToByte    convert real color -> color
//  RGBByteToReal    convert color -> real color
//  RGBRealGray      return gray value of real color
//  RGBNorm          nomalize real color
//  RGBRandom        return random real color
//  ColorString      convert color values to string
//  ColorHexString   convert color values to hex string
//
//*********************************************************
{$I comp_opt.inc}

UNIT ColorRGB;

INTERFACE

uses  Graphics, SysUtil2, Gr32;

{$I float.inc}

const maxColorRGB = 255;               {global color contants}
      intMask   = $FF000000;
      redMask   = $00FF0000;
      greenMask = $0000FF00;
      blueMask  = $000000FF;
      rgbMask   = $00FFFFFF;

//=== t C o l o r 3 2 ===

const maxInt32   = $FF000000;     // max. intensity

//type  tColor32 = Cardinal;        {A R G B}
type  tColor32 = gr32.tColor32;   {A R G B}

      pColor32 = ^tColor32;
      tColor32Array = array [0..9999] of tColor32;
      pColor32Array = ^tColor32Array;
      tArrayOfColor32 = array of tColor32;

type  pColorRGBI = ^tColorRGBI;
      tColorRGBI = packed record
        rb,gb,bb, ii: byte;   end;

//=== t L o n g C o l o r ===
type  tLongColor = record              {color with longint values}
        r,g,b: longint;   end;         {  rgb range of 24.8 fix point}

//=== t R e a l C o l o r ===
type  pRealColor = ^tRealColor;
      tRealColor = record              {color: red-green-blue}
        r,g,b: float;   end;           {  rgb range: 0.0 .. 1.0}

const maxRealByte = 255.99;
      blackRealColor: tRealColor = (r: 0.0; g: 0.0; b: 0.0);
      whiteRealColor: tRealColor = (r: 1.0; g: 1.0; b: 1.0);

{ 8 x 8 - dither matrix for 64 intensity patterns }
const intMatrix8: array [0..7,0..7] of byte =
     (( 0,32, 8,40, 2,34,10,42),(48,16,56,24,50,18,58,26)
     ,(12,44, 4,36,14,46, 6,38),(60,28,52,20,62,30,54,22)
     ,( 3,35,11,43, 1,33, 9,41),(51,19,59,27,49,17,57,25)
     ,(15,47, 7,39,13,45, 5,37),(63,31,55,23,61,29,53,21));

//=== t C o l o r ===   {declared in standard unit Graphics}
 function WinColor      (r,g,b: byte): tColor;              overload;
procedure Color2RGB     (color: tColor; var r,g,b: byte);   overload;
 function ShadeColor    (color: tColor; i1: byte): tColor;  overload;
 function MixColor      (c1,c2: tColor; i2: byte): tColor;  overload;
 function MixHalfColor  (c1,c2: tColor): tColor;            overload;
 function RandomColor   (min, max: byte): tColor;
 function GrayValue     (color: tColor): byte;
 function ColorString   (color: tColor): string;            overload;
 function ColorHexString(color: tColor): string;            overload;

//=== t C o l o r 1 5 ===
 function RGB15 (r,g,b: byte): word;

//=== t C o l o r 1 6 ===
 function RGB16 (r,g,b: byte): word;

//=== t C o l o r 3 2 ===
                                       // like windows.TRGBQuad
type  tCol32 = packed record  blue, green, red, alpha: byte;   end;

// function Color32 (color: tColor): tColor32;                    overload;
 function Color32 (r,g,b: Byte): tColor32;                      overload;
 function WinColor (color: tColor32): tColor;                   overload;
 function ShadeColor(color: tColor32; i1: byte): tColor32;      overload;
 function MixColor (c1,c2: tColor32; i2: byte): tColor32;       overload;
 function MixHalfColor (c1,c2: tColor32): tColor32;             overload;
 function ColorMin (c1,c2: tColor32): tColor32;
 function ColorMax (c1,c2: tColor32): tColor32;
 function RandomColor32 (min, max: byte): tColor32;
 function ColorString    (color: tColor32): string;             overload;
 function ColorHexString (color: tColor32): string;             overload;

{ additional colors:   a r g b}
const
      Red          = $FFFF0000;
      DarkRed      = $FF800000;
      LightRed     = $FFFF8080;

      Green        = $FF00FF00;
      DarkGreen    = $FF008000;
      LightGreen   = $FF80FF80;

      Blue         = $FF0000FF;
      DarkBlue     = $FF000080;
      LightBlue    = $FF8080FF;

      Cyan         = $FF00FFFF;
      DarkCyan     = $FF008080;
      LightCyan    = $FF80FFFF;

      Magenta      = $FFFF00FF;
      DarkMagenta  = $FF800080;
      LightMagenta = $FFFF80FF;

      Gray         = $FF808080;
      DarkGray     = $FF404040;
      LightGray    = $FFC0C0C0;

      Black        = $FF000000;
      Yellow       = $FFFFFF00;
      White        = $FFFFFFFF;

// frei erfunden...
      Invisible    = $00000000;
      Brown        = $FF774422;
      Violet       = $FFAA00FF;
      Pink         = $FFFF9999;
      WineRed      = $FF990033;
      Orange       = $FFFF8833;
      Silver       = $FFDEDEDE;
      Gold         = $FFFAC852;
      LightYellow  = $FFFFFF80;

//=== t L o n g C o l o r ===
 function LongColor (color: tColor): tLongColor;
 function Color32   (color: tLongColor): tColor32;                   overload;
 function WinColor  (color: tLongColor): tColor;                     overload;
 function ShadeColor(color: tLongColor; i1: byte): tLongColor;       overload;
procedure IncColor  (var c1,c2: tLongColor);                         overload;
 function MixHalfColor (c1,c2: tLongColor): tLongColor;              overload;

//=== t R e a l C o l o r ===
function Color32 (var rgbReal: tRealColor): tColor32;                overload;
function RealColor (r,g,b: float): tRealColor;                   overload;
function RealColor (var color: tColor): tRealColor;                  overload;
function RealColor (var color: tColor32): tRealColor;                overload;
procedure RGBRealGray   (var rgbReal: tRealColor);
procedure RGBNorm       (var rgbReal: tRealColor);
procedure RGBRandom     (var rgbReal: tRealColor);

//*********************************************************
IMPLEMENTATION   uses SysUtils;

//===============
//  t C o l o r
//===============

//*********************************************************
// convert r g b values to tColor
// input:     r,g,b:   rgb values (0..255)
// return:    dWord:   24 bit color
// format:    24bit (888) 00000000 bbbbbbbb gggggggg rrrrrrrr
//---------------------------------------------------------
function WinColor (r,g,b: byte): tColor;
begin
  WinColor := r + g shl 8 + b shl 16;
end;
//*********************************************************
// convert color to rgb values
// input:     color:  mode specific color value
// output:    r,g,b:  rgb values of color (0..255)
//---------------------------------------------------------
procedure Color2RGB (color: tColor; var r,g,b: byte);
begin
  r :=  longint(color)         and $FF;
  g := (longint(color) shr  8) and $FF;
  b := (longint(color) shr 16) and $FF;
end;
//*********************************************************
// calculate intensity color
//            ShadeColor = (color * int) / 255
// input:     color:  base colors
//            i1:     intensity of color (0..255)
// return:    tColor: intensity color
//---------------------------------------------------------
function ShadeColor (color: tColor; i1: byte): tColor; overload;
begin
{$Q-}
  Result := ((longWord(color) and $00FF00FF) * i1) shr 8 and $FF00FF
           +((longWord(color) and $0000FF00) * i1) shr 8 and $FF00;
end;
//*********************************************************
// mix 2 colors with given intensity
//            MixColor = (c1 * (255-i2) + c2 * i2) / 256
// input:     c1,c2:  colors to mix
//            i2:     intensity of color 2 (0..255)
// return:    tColor: mixed color
//---------------------------------------------------------
function MixColor (c1,c2: tColor; i2: byte): tColor;
var  i1: longWord;
begin
  i1 := 255 - i2;
  Result := ((longWord(c1) and $00FF00FF) * i1
            +(longWord(c2) and $00FF00FF) * i2) shr 8 and $FF00FF
           +((longWord(c1) and $0000FF00) * i1
            +(longWord(c2) and $0000FF00) * i2) shr 8 and $FF00;
end;
//*********************************************************
// mix 2 colors with half intensity together
// note: this function is faster than MixColor(c1,c2,127)
//            MixHalfColor := (c1 + c2) / 2
// input:     c1,c2:  colors to mix
// return:    tColor: mixed color
//---------------------------------------------------------
function MixHalfColor (c1,c2: tColor): tColor;
begin
  Result := ((c1 shr 1) and $7F7F7F7F
            +(c2 shr 1) and $7F7F7F7F);
end;
//*********************************************************
// return a random color with rgb values between min and max
//---------------------------------------------------------
function RandomColor (min, max: byte): tColor;
var rng: word;
begin
  rng := max - min + 1;
  Result := WinColor (min + random(rng)
                     ,min + random(rng)
                     ,min + random(rng));
end;
//*********************************************************
// return gray value (0..255) of given color
//---------------------------------------------------------
function GrayValue (color: tColor): byte;

begin
  with tColorRGBI(color)
  do Result := (77 * rb + 150 * gb + 29 * bb) shr 8;
end;
//*********************************************************
// return color values as string
//---------------------------------------------------------
function ColorString (color: tColor): string;
begin
  with tColorRGBI(color)
  do Result := Format('r=%3d  g=%3d  b=%3d',[rb,gb,bb]);
end;
//*********************************************************
// return color values as hex string
//---------------------------------------------------------
function ColorHexString (color: tColor): string;
begin
  with tColorRGBI(color)
  do Result := HexString (rb) + ','
             + HexString (gb) + ','
             + HexString (bb);
end;

//===================
//  t C o l o r 1 5
//===================

{*********************************************************}
{ convert r g b values to word color
{ input:     r,g,b:   rgb values (0..255)
{ return:    word:    15 bit color
{ format:    15bit (555) 0rrrrrgg gggbbbbb
{---------------------------------------------------------}
function RGB15 (r,g,b: byte): word;
begin
  RGB15 := b shr 3 or (g shl 2) and $03E0
                   or (r shl 7) and $FC00;
end;

//===================
//  t C o l o r 1 6
//===================

{*********************************************************}
{ convert r g b values to word color
{ input:     r,g,b:   rgb values (0..255)
{ return:    word:    16 bit color
{ format:    16bit (565) rrrrrggg gggbbbbb
{---------------------------------------------------------}
function RGB16 (r,g,b: byte): word;
begin
  RGB16 := b shr 3 or (g shl 3) and $07E0
                   or (r shl 8) and $F800;
end;

//===================
//  t C o l o r 3 2
//===================

//*********************************************************
// combines arguments to 4 byte tColor32
//---------------------------------------------------------
function Color32 (r,g,b: Byte): tColor32;
begin
  Result := $FF000000 + B shl 16 + G shl 8 + R;
 // Result := (((((A shl 8) or B) shl 8) or G) shl 8) or R;
end;
//*********************************************************
// convert tColor32 to windows color (see GR32...)
//---------------------------------------------------------
function WinColor (color: tColor32): tColor;
asm
  // the alpha channel byte is set to zero!
  ROL    EAX,8  // ABGR  ->  BGRA
  XOR    AL,AL  // BGRA  ->  BGR0
  BSWAP  EAX    // BGR0  ->  0RGB
end;
//*********************************************************
// mix 2 colors with given intensity
//            MixColor = (c1 * (255-i2) + c2 * i2) / 256
// input:     c1,c2:  colors to mix
//            i2:     intensity of color 2 (0..255)
// return:    tColor: mixed color
//---------------------------------------------------------
function MixColor (c1,c2: tColor32; i2: byte): tColor32;
var  i1: longWord;
begin
  i1 := 255 - i2;                         // ??? Alpha mix fehlt
  Result := ((c1 and $00FF00FF) * i1
            +(c2 and $00FF00FF) * i2) shr 8 and $00FF00FF
           +((c1 and $0000FF00) * i1
            +(c2 and $0000FF00) * i2) shr 8 and $0000FF00;
end;
//*********************************************************
// calculate intensity color,
//   similar to Gr32.SetAlpha but changes RGB values
//            ShadeColor = (color * int) / 255
// input:     color:  base colors
//            i1:     intensity of color (0..255)
// return:    tColor: intensity color
//---------------------------------------------------------
function ShadeColor(color: tColor32; i1: byte): tColor32;
begin
  Result := ((longWord(color) and $00FF00FF) * i1) shr 8 and $FF00FF
           +((longWord(color) and $0000FF00) * i1) shr 8 and $FF00
           +(color and $FF000000);
end;
//*********************************************************
// mix 2 colors with half intensity together
// note: this function is faster than MixColor(c1,c2,127)
//            MixHalfColor := (c1 + c2) / 2
// input:     c1,c2:  colors to mix
// return:    tColor32: mixed color
//---------------------------------------------------------
function MixHalfColor (c1,c2: tColor32): tColor32;
begin
  Result := ((c1 shr 1) and $7F7F7F7F
            +(c2 shr 1) and $7F7F7F7F);
end;
//*********************************************************
// return color with minimum rgb values
//---------------------------------------------------------
function ColorMin (c1,c2: tColor32): tColor32;
begin
  Result := Min(c1 and intMask,   C2 and intMask)
          + Min(c1 and redMask,   C2 and redMask)
          + Min(c1 and greenMask, C2 and greenMask)
          + Min(c1 and blueMask,  C2 and blueMask);
end;
//*********************************************************
// return color with maximum rgb values
//---------------------------------------------------------
function ColorMax (c1,c2: tColor32): tColor32;
begin
  Result := Max(c1 and intMask,   C2 and intMask)
          + Max(c1 and redMask,   C2 and redMask)
          + Max(c1 and greenMask, C2 and greenMask)
          + Max(c1 and blueMask,  C2 and blueMask);
end;
//*********************************************************
// return a random color with rgb values between min and max
//---------------------------------------------------------
function RandomColor32 (min, max: byte): tColor32;
var rng: word;
begin
  rng := max - min + 1;
  Result :=  maxInt32 or word(min + random(rng)) shl 16
                      or word(min + random(rng)) shl 8
                      or word(min + random(rng));
end;
//*********************************************************
function ColorString (color: tColor32): string;
begin
  with tCol32(color)
  do Result := Format('r=%3d  g=%3d  b=%3d  alpha=%3d',[red,green,blue,alpha]);
end;
//*********************************************************
// return color values as hex string
//---------------------------------------------------------
function ColorHexString (color: tColor32): string;
begin
  Result := '$' + HexString (color);
end;
(*
//*********************************************************
// mix 2 colors with given intensity
//            MixColor = (c1 * (255-i2) + c2 * i2) / 256
//            Rearranged formula saves 1 multiplication:
//            MixColor := c1 + (c2-c1) * i2 / 256
// input:     c1,c2:  colors to mix
//            i2:     intensity of color 2 (0..255)
// return:    tColor: mixed color
//---------------------------------------------------------
function MixColor (c1,c2: tCol; i2: byte): tCol;
var  i1: longint;
begin
  i1 := 255 - i2;
  MixColor.rb := (c1.rb * i1 + c2.rb * i2) shr 8;
  MixColor.gb := (c1.gb * i1 + c2.gb * i2) shr 8;
  MixColor.bb := (c1.bb * i1 + c2.bb * i2) shr 8;
end;
{
begin
  MixColor.bb := c1.bb + ((longint(c2.bb) - c1.bb) * i2) div 256;
  MixColor.gb := c1.gb + ((longint(c2.gb) - c1.gb) * i2) div 256;
  MixColor.rr := c1.rr + ((longint(c2.rr) - c1.rr) * i2) div 256;
end;
}
*)

//=======================
//  t L o n g C o l o r
//=======================

//*********************************************************
// convert color from byte to 24.8 fix point values
//---------------------------------------------------------
function LongColor (color: tColor): tLongColor;
begin
  with tColorRGBI(color)
  do begin
    result.r := rb shl 8;
    result.g := gb shl 8;
    result.b := bb shl 8;
  end;
end;
//*********************************************************
// convert long color to tColor32
//---------------------------------------------------------
function Color32 (color: tLongColor): tColor32;
begin
  result := color.r shl 16 and redMask
           +color.g        and greenMask
           +color.b shr 8  and blueMask;
end;
//*********************************************************
// convert color from 24.8 fix point to windows color
//---------------------------------------------------------
function WinColor (color: tLongColor): tColor;
begin
  result := color.r shl 8 and redMask
           +color.g       and greenMask
           +color.b shr 8 and blueMask;
end;
//*********************************************************
// calculate intensity color
//            ShadeColor = (color * int) / 255
// input:     color:  base colors
//            i1:     intensity of color (0..255)
// return:    tColor: intensity color
//---------------------------------------------------------
function ShadeColor(color: tLongColor; i1: byte): tLongColor; overload;
begin
  Result.r := (color.r * i1) shr 8;
  Result.g := (color.g * i1) shr 8;
  Result.b := (color.b * i1) shr 8;
end;
//---------------------------------------------------------
procedure IncColor  (var c1,c2: tLongColor);  //c1 = c1 + c2;
begin
//  c1.r := c1.r + c2.r;
//  c1.g := c1.g + c2.g;
//  c1.b := c1.b + c2.b;
  inc (c1.r, c2.r);
  inc (c1.g, c2.g);
  inc (c1.b, c2.b);
end;
//---------------------------------------------------------
function MixHalfColor (c1,c2: tLongColor): tLongColor;
begin
  Result.r := (c1.r + c2.r) shr 1;
  Result.g := (c1.g + c2.g) shr 1;
  Result.b := (c1.b + c2.b) shr 1;
end;

//=======================
//  t R e a l C o l o r
//=======================

(*
//=== t R G B ===                      {used for palette handling}

type  pRGB = ^tRGB;
      tRGB = packed record             {color with longint values}
        rb,gb,bb: byte;   end;         {  rgb range: 0..255}

//*********************************************************
// convert color from tRealColor to tRGB format
//---------------------------------------------------------
procedure RGBRealToByte (var rgbReal: tRealColor;
                         var rgb: tRGB);
begin
  with rgbReal, rgb
  do begin
    rb := byte (trunc (r * maxRealByte));
    gb := byte (trunc (g * maxRealByte));
    bb := byte (trunc (b * maxRealByte));
  end;
end;
//*********************************************************
// convert color from tRGB to tRealColor format
//---------------------------------------------------------
procedure RGBByteToReal (var rgb: tRGB;
                         var rgbReal: tRealColor);
begin
  with rgbReal, rgb
  do begin
    r := rb / maxColorRGB;
    g := gb / maxColorRGB;
    b := bb / maxColorRGB;
  end;
end;
*)
//*********************************************************
// convert color from tRealColor to tColor32 format
//---------------------------------------------------------
function Color32 (var rgbReal: tRealColor): tColor32;
begin
  with rgbReal
  do result := Color32 (byte (trunc (r * maxRealByte))
                       ,byte (trunc (g * maxRealByte))
                       ,byte (trunc (b * maxRealByte)),$FF);
end;
//*********************************************************
// convert color values to tRealColor format
//---------------------------------------------------------
function RealColor (r,g,b: float): tRealColor;
begin
  Result.r := r;
  Result.g := g;
  Result.b := b;
  RGBNorm (Result);
end;
//*********************************************************
// convert color from tColor to tRealColor format
//---------------------------------------------------------
function RealColor (var color: tColor): tRealColor;
begin
  Result.r := tColorRGBI(color).rb / 255.0;
  Result.g := tColorRGBI(color).gb / 255.0;
  Result.b := tColorRGBI(color).bb / 255.0;
end;
//*********************************************************
// convert color from tColor32 format to tRealColor format
//---------------------------------------------------------
function RealColor (var color: tColor32): tRealColor;

var   alpha: float;
begin
  with Result do begin
  alpha := (color shr 24) * 256.0;
    r := (color and $00FF0000) shr 8 / alpha;
    g := (color and $0000FF00)       / alpha;
    b := (color and $000000FF) shl 8 / alpha;
  end;
end;
//*********************************************************
// convert rgb values to gray scaled color
//---------------------------------------------------------
procedure RGBRealGray (var rgbReal: tRealColor);

begin
  with rgbReal
  do begin
    r := 0.299 * r + 0.587 * g + 0.114 * b;
    g := r;
    b := r;
  end;
end;
//*********************************************************
// normalize RGB values to range 0..1
//---------------------------------------------------------
procedure RGBNorm (var rgbReal: tRealColor);

var  max: float;

begin
  with rgbReal
  do begin
    if r >= g
    then
      if r >= b
      then max := r
      else max := b
    else   {g > r}
      if g >= b
      then max := g
      else max := b;

    if max > 1.00
    then begin
      r := r / max;
      g := g / max;
      b := b / max;
    end;
  end;
end;
//*********************************************************
// create random color
//---------------------------------------------------------
procedure RGBRandom (var rgbReal: tRealColor);
begin
  rgbReal.r := random(256) / 255.0;
  rgbReal.g := random(256) / 255.0;
  rgbReal.b := random(256) / 255.0;
end;

end.

