
//*********************************************************
//
//  UNIT:     D a n c e                03.09.99, 05.04.2002
//
//  dancing RGB lights defined by a lightning map
//  coded by Gerd Platl
//
//*********************************************************
{$I 'comp_opt.inc'}

unit Dance;

{$ifopt D+}
  {$define xtest1    main functions}
  {$define xtest2    sub functions}
{$endif}

interface

// used by unit BumpLi
function UseLightmap (mode: word): word;
function LMLinePtr (y:integer): pointer;

//*********************************************************
implementation

uses  SysUtils, Windows, Graphics,
      SysUtil2, ColorRGB, Graph, Palette, Values, GrForm;

var   lightBm: tBitmap = nil;

const modes =   4;           // # of lightning maps
      mapX  = 512;           // width lightning map
      mapY  = 512;           // height lightning map
      mapOs = mapY + 10;     // height with overscan lines
      minLI =  22;           // minimum of light intensity
      maxLI = 255;           // maximum of light intensity

var   modeNo:      word    =  0;       // actual demo mode
      moveSpeed:   integer = 20;       // light moving speed
      radius:      integer = 65;       // moving radius

{********************************************************}
procedure CreateValues;
begin
  with valuePtr^
  do if ValueCount = 0
  then begin
    AppendValue ('speed',  1, 21);
    AppendValue ('radius', 5, 95);
  end;
end;

//*********************************************************
// return pointer of light map line
//---------------------------------------------------------
function LMLinePtr (y:integer): pointer;
begin
  if      y < 0    then result := lightBm.ScanLine[0]
  else if y < mapY then result := lightBm.ScanLine[y]
  else                  result := lightBm.ScanLine[mapY];
end;
//*********************************************************
function LightIntensity (x,y:integer): byte;
var   lp: PByteArray;
begin
  lp := lightBm.ScanLine[y mod mapY];
  result := lp^[x mod mapX];
end;
//*********************************************************
// start new animation
//---------------------------------------------------------
procedure AnimStart;
var   helpPalette: tPalette;
begin
  {$ifdef test1}
    writeln (tst, #13#10'Dance.AnimStart: ', modeNo);
  {$endif}

  CreateValues;
//  SetGraphMode (pf15bit);
//  SetGraphMode (pf16bit);
  SetGraphMode (pf24bit);
//  SetGraphMode (pfDevice);
//  SetGraphMode (pf8bit);

  if lightBm = nil
  then begin
    lightBm := tBitmap.Create;
    lightBm.PixelFormat := pf8Bit;
    lightBm.Width  := mapX;
    lightBm.Height := mapOs;      // overscan lines needed !!!
    SetIntensityPalette (helpPalette);
    SetBitmapPalette (lightBm, helpPalette);
  end;

  modeNo := UseLightmap (1);
end;
//*********************************************************
// calculate simple circle lightning map
//---------------------------------------------------------
procedure CalculateLightBitmap1;
var   xd,yd, xi,yi, ly,li: longint;
      lp: ^byte;
begin
  xd := mapX div 2;
  yd := mapY div 2;
  for yi := 0 to mapOs - 1          //         ___  _____maxLI
  do begin                          //       -    -
    ly := 3 * sqr(yd - yi);         //     /        \
    lp := lightBm.ScanLine[yi];     // __/            \__minLI
    for xi := 0 to mapX - 1
    do begin
      li := maxLI - trunc(sqrt(ly + 3*sqr(xi-xd)));
      if li <= minLI
      then lp^ := minLI
      else lp^ := li;
      inc (lp);
    end;
  end;
end;
//*********************************************************
// calculate simple diamont lightning map
//---------------------------------------------------------
procedure CalculateLightBitmap2;
var   xd,yd, xi,yi, ly,li: longint;
      lp: ^byte;
begin
  xd := mapX div 2;
  yd := mapY div 2;
  for yi := 0 to mapOs - 1          //        ______maxLI
  do begin                          //       ^
    ly := maxLi - 3*abs(yd-yi);     //     /   \
    lp := lightBm.ScanLine[yi];     // __/       \__minLI
    for xi := 0 to mapX - 1
    do begin
      li := (ly + maxLI - 3* abs(xd-xi)) div 2;
      if li <= minLI
      then lp^ := minLI
      else lp^ := li;
      inc (lp);
    end;
  end;
end;
//*********************************************************
// calculate simple star lightning map
//---------------------------------------------------------
procedure CalculateLightBitmap3;
var   xd,yd, xi,yi, li: longint;
      lp: ^byte;   ly: double;
begin
  xd := mapX div 2;
  yd := mapY div 2;
  for yi := 0 to mapOs - 1          //       |  _____maxLI
  do begin                          //      | |
    ly := sqrt(abs(yi - yd));       //     /   \
    lp := lightBm.ScanLine[yi];     // __/       \__minLI
    for xi := 0 to mapX - 1
    do begin
      li := maxLi - trunc (16*(sqrt(abs(xi-xd))+ly));
      if li <= minLI
      then lp^ := minLI
      else lp^ := li;
      inc (lp);
    end;
  end;
end;
//*********************************************************
// calculate simple hole lightning map
//---------------------------------------------------------
procedure CalculateLightBitmap4;
var   xd,yd, xi,yi, ly,li: longint;
      lp: ^byte;
begin
  xd := mapX div 2;
  yd := mapY div 2;
  for yi := 0 to mapOs - 1          // __      _____maxLI
  do begin                          //   \    /
    ly := sqr(yi - yd);             //    \  /
    lp := lightBm.ScanLine[yi];     //    |__|    __minLI
    for xi := 0 to mapX - 1
    do begin
      li := maxLI - trunc(1000000 / (1+sqr(xi-xd)+ly));
      if li <= minLI
      then lp^ := minLI
      else lp^ := li;
      inc (lp);
    end;
  end;
end;
//*********************************************************
// calculate light bitmap values
//---------------------------------------------------------
function UseLightmap (mode: word): word;
begin
  mode := 1 + (mode - 1) mod modes;

{$ifdef test2}
  writeln (tst, 'UseLightmap: ', mode);
{$endif}

  case mode of
    1: CalculateLightBitmap1;
    2: CalculateLightBitmap2;
    3: CalculateLightBitmap3;
    4: CalculateLightBitmap4;
  end;
  result := mode;
end;

//*********************************************************
// draw one animation picture
//---------------------------------------------------------
procedure AnimDraw;

var
    mx,my,                   // middle point
    lx1,ly1,li1,             // light position
    lx2,ly2,li2,             // and line index
    lx3,ly3,li3,
    xi,yi: longint;          // pixel position
    rgb3Ptr: ^byte;          // drawing pointer (3 byte)
    rgb2Ptr: ^word;          // Color15 or Color16 coded (2 byte)
    lp1,lp2,lp3: ^byte;      // light intensity pointer
    dm: tBitmap;             // destination bitmap
begin
  {$ifdef test2}
    writeln (tst, 'Dance.AnimDraw: ');
  {$endif}

  dm := bgBitmap;
  mx := (mapX - pageX) div 2;
  my := (mapY - pageY) div 2;
  // calculate light position
  lx1 := mx + round(cos(1.1*picNo*moveSpeed / 256) * radius);
  ly1 := my + round(sin(1.2*picNo*moveSpeed / 200) * radius);
  lx2 := mx + round(cos(1.3*picNo*moveSpeed / 256) * radius);
  ly2 := my + round(sin(1.4*picNo*moveSpeed / 200) * radius);
  lx3 := mx + round(cos(1.5*picNo*moveSpeed / 256) * radius);
  ly3 := my + round(sin(1.6*picNo*moveSpeed / 200) * radius);

  for yi := 0 to GetMaxY
  do begin
    // get scanline indizes
    my := 2 * mapY + yi;   // to get always values > 0
    li1 := (my + ly1) mod mapY;
    li2 := (my + ly2) mod mapY;
    li3 := (my + ly3) mod mapY;

    // get scanline pointers
    lp1 := lightBm.ScanLine[li1];  inc (lp1, lx1);
    lp2 := lightBm.ScanLine[li2];  inc (lp2, lx2);
    lp3 := lightBm.ScanLine[li3];  inc (lp3, lx3);

    case dm.PixelFormat of
      pf24Bit: begin
          rgb3Ptr := dm.ScanLine[yi];   // 3 byte
          for xi := 0 to GetMaxX
          do begin
            rgb3Ptr^ := lp1^;  inc (rgb3Ptr);   inc(lp1);
            rgb3Ptr^ := lp2^;  inc (rgb3Ptr);   inc(lp2);
            rgb3Ptr^ := lp3^;  inc (rgb3Ptr);   inc(lp3);
          end;
        end;

      pf16Bit: begin
          rgb2Ptr := dm.ScanLine[yi];   // 2 byte
          for xi := 0 to GetMaxX
          do begin
            rgb2Ptr^ := RGB16(lp1^,lp2^,lp3^);
            inc(rgb2Ptr);
            inc(lp1);
            inc(lp2);
            inc(lp3);
          end;
        end;

      pf15Bit: begin
          rgb2Ptr := dm.ScanLine[yi];   // 2 byte
          for xi := 0 to GetMaxX
          do begin
            rgb2Ptr^ := RGB15(lp1^,lp2^,lp3^);
            inc(rgb2Ptr);
            inc(lp1);
            inc(lp2);
            inc(lp3);
          end;
        end;
      else RaiseGraphError (sPixelFormat);
    end;
  end;
end;

//*********************************************************
// handle animation input
//---------------------------------------------------------
procedure AnimInput;
begin
  {$ifdef test2}
    writeln (tst, 'Dance.AnimInput: ', ord(cb));
  {$endif}

  with valuePtr^ do begin
    moveSpeed  := IntValue (1);
    radius     := IntValue (2);
  end;

  case cb of
    CR : modeNo := UseLightmap (modeNo + 1);
    'r': valuePtr^.Reset;
    'n': valuePtr^.RandomValues;
  end;
end;
//*********************************************************
// display animation specific help
//---------------------------------------------------------
procedure AnimHelp;
begin
  HelpText (1, '<ret> light map ' + IntToStr(modeNo));
  HelpText (2, '  n   random values');
end;
//*********************************************************
initialization

  {$ifdef InitTest}
    writeln (tst, 'Dance.initialization: ');
    flush (tst);
  {$endif}

  RegisterAnimation (' Lissa Dance ',
    AnimStart, nil, AnimDraw, AnimInput, AnimHelp);

//----------------------------------------------------------
finalization

  {$ifdef ExitTest}
    writeln (tst, 'Dance.finalization: ');
    flush (tst);
  {$endif}

end.

