//*******************************************************************
//
//  UNIT:     E m p t y                     22.12.2000,  03.12.2001
//
//  basic template unit for creating new demo units, coded by Gerd Platl
//
//*******************************************************************
{$I 'comp_opt.inc'}

unit Empty;

{$ifopt D+}
  {$define xtest1    main functions}
  {$define xtest2    sub functions}
{$endif}

interface

//*********************************************************
implementation

uses  SysUtils, Windows, Classes, Graphics,
      SysUtil2, Graph, Images, Values, GrForm;

const modes  = 2;                      // # of demo modes

var   modeNo:      word = 1;           // actual demo mode
      backBm:      tBitmap = nil;      // background bitmap

{********************************************************}
procedure CreateValues;
begin
  with valuePtr^
  do if ValueCount = 0
  then begin
    Add ('value 1',  10,  50);
    Add ('value 2', 100, 500);
  end;
end;

//*********************************************************
// create double sized background bitmap
// size: 2*screen.X * 2*screen.Y+256
//---------------------------------------------------------
procedure InitEmpty;
begin
  backBm := tBitmap.Create;
  backBm.PixelFormat := screen.pixelFormat;
  backBm.Width  := screen.X;
  backBm.Height := screen.Y;
end;
//*********************************************************
procedure NextMode;
begin
  modeNo := 1 + modeNo mod modes;
  ClearBitmap (bgBitmap);
  MarkTime;
end;
//*********************************************************
// start new animation
//---------------------------------------------------------
procedure AnimStart;
begin
  {$ifdef test1}
    writeln (tst, #13#10'Empty.AnimStart: ', modeNo);
  {$endif}

  CreateValues;
  SetGraphMode (screen.pixelFormat);
  NextMode;
end;
//*********************************************************
// resize animation window
//---------------------------------------------------------
procedure AnimResize;
begin
  {$ifdef test1}
    writeln (tst, 'Shelly.AnimResize: ',pageX,'*',pageY);
  {$endif}

  ClearBitmap (backBm);
end;
//*********************************************************
// example of slow pixel drawing using canvas.Pixels[..]
//---------------------------------------------------------
procedure PixelDemo1;
var   xi: integer;
      image: tCanvas;
begin
  image := animWindow.Canvas;
  for xi := 1 to valuePtr^.IntValue(2)
  // draw direct to screen canvas (not buffered !!!)
  do image.Pixels[random(pageX),random(pageY)] := random ($FFFFFF);
end;

//*********************************************************
// direct paint to window image1 !!!
//---------------------------------------------------------
procedure LineDemo1;
var   xi: integer;
      image: tCanvas;
begin
  image := animWindow.Canvas;
  for xi := 1 to valuePtr^.IntValue(1)
  do with animWindow do begin
    image.Pen.Color := random($FFFFFF);
    image.MoveTo (mouseX, mouseY);
    image.LineTo (random (pageX),random(pageY));
  end;
end;

//*********************************************************
// draw one animation picture
//---------------------------------------------------------
procedure AnimDraw;

begin
  {$ifdef test2}
    writeln (tst, 'Empty.AnimDraw: ');
  {$endif}

  case modeNo of
    1: LineDemo1;
    2: PixelDemo1;
  end;
end;

//*********************************************************
// handle animation input
//---------------------------------------------------------
procedure AnimInput;
begin
  {$ifdef test2}
    writeln (tst, 'Empty.AnimInput: ', ord(cb));
  {$endif}

  case cb of
    CR : NextMode;
    'n': valuePtr^.RandomValues;
    'r': valuePtr^.Reset;
    else valuePtr^.KeyDown (vKey, shiftState);
  end;
end;

//*********************************************************
// display animation specific help
//---------------------------------------------------------
procedure AnimHelp;
begin
  HelpText (1, '<ret>  mode ' + IntToStr(modeNo));
  HelpText (2, '   n   random values');
  HelpText (3, '   r   reset values');
end;

//*********************************************************
initialization

  {$ifdef InitTest}
    writeln (tst, 'Empty.initialization: ');
    flush (tst);
  {$endif}

  InitEmpty;

  RegisterAnimation (' Empty demo ',
    AnimStart, AnimResize, AnimDraw, AnimInput, AnimHelp);

//----------------------------------------------------------
finalization

  {$ifdef ExitTest}
    writeln (tst, 'Empty.finalization: ');
    flush (tst);
  {$endif}

  backBM.Free;
end.

