//*********************************************************
//
//  UNIT:     F i r e                14.02.2000, 05.06.2003
//
//  fire animation, coded by Gerd Platl
//
//*********************************************************
{$I comp_opt.inc}

unit Fire;

{$ifopt D+}
  {$define xtest1    main functions}
  {$define xtest2    sub functions}
{$endif}

interface

//*********************************************************
implementation

uses  SysUtils, Windows, Graphics, ColorRGB, Fonts,
      SysUtil2, Math2, Graph, Palette, Blur, Values,
      GrForm, Rot3d;

var   backBm: tBitmap = nil;
      cPal: tPalette;

{================================================}
{ palette handling
{================================================}

const spreadColors = 8;

      fireColors: array [0..spreadColors-1] of tColorRGBI
      =((rb:  0; gb:  0; bb:  0; ii:   0)   // black
       ,(rb: 33; gb: 33; bb: 33; ii:  50)   // smoke
       ,(rb:155; gb: 50; bb:  0; ii: 150)   // red
       ,(rb:255; gb:127; bb:  0; ii: 200)   // lightRed
       ,(rb:255; gb:255; bb:  0; ii: 220)   // yellow
       ,(rb:255; gb:255; bb:255; ii: 240)   // white
       ,(rb:  0; gb:  0; bb:155; ii: 250)   // blue
       ,(rb:  0; gb:  0; bb:100; ii: 255)); // dark blue

      clFire: tColor = $01FF0F000;

      maxRnd = 2047;
var   randomArray: array[0..2047] of byte;

{********************************************************}
procedure CreateValues;
begin
  with valuePtr^
  do if ValueCount = 0
  then begin
    AppendValue ('root',         0,  50);
    AppendValue ('new fire',     0,  20);
    AppendValue ('minFire',      0, 100);
    AppendValue ('more fire',   -2,  18);
    AppendValue ('fire inc',     3,  90);
    AppendValue ('smooth base',  0,   9);
    AppendValue ('decay',        0,  10);
    AppendValue ('sparks',       0,  40);
    AppendValue ('sprinkle',     0,  20);
    AppendValue ('base color',  80, 255);
  end;
end;
//*********************************************************
// start new animation
//---------------------------------------------------------
procedure AnimStart;
var   ni: integer;
begin
  {$ifdef test1}
    writeln (tst, #13#10'Fire.AnimStart: ');
  {$endif}

  Randomize;
  CreateValues;
  SetGraphMode (pf8bit);
  SetSpreadPalette (cPal, fireColors, spreadColors);
  for ni := 0 to maxRnd do randomArray[ni] := random(3);
  ChangeEffect (3);
end;
//*********************************************************
// resize animation window
//---------------------------------------------------------
procedure AnimResize;
begin
  {$ifdef test1}
    writeln (tst, 'Fire.AnimResize: ',pageX,'*',pageY);
  {$endif}

  if backBm = nil
  then begin
    backBm := tBitmap.Create;
    backBm.PixelFormat := pf8Bit;
  end;
  // resize and clear air bitmap
  backBm.Width  := pageX;
  backBm.Height := pageY;

  // nice effect to burn down last pf8bit demo picture
  CopyBitmap (bgBitmap, backBm);
  CopyBitmap (backBm, bgBitmap);

  // set fire palette
  SetAllPalette (cPal);
  SetBitmapPalette (backBm, cPal);

  ResizeRot3d;
  SetBaseColor (180);
end;
{*********************************************************}
{ fast 2x2 rectangle drawing
{---------------------------------------------------------}
procedure Put4Pixel (x,y: integer; color: byte);

var   wCol: word;   pixPtr: ^word;
begin
  wCol := color + color shl 8;
  pixPtr := GetPixelPointer(x,y);
  pixPtr^ := wCol;
  pixPtr := GetPixelPointer(x,y+1);
  pixPtr^ := wCol;
end;

{================================================}
{ fire simulation
{================================================}

const burnText = 'fire simulation';

var   bluring:    boolean = true;      {bluring fire pixels}
      anim3d:     boolean = true;      {draw 3d figur}
      animText:   boolean = false;     {draw moving text}

var   rootRand,         {Max/Min decrease of the root of the flames}
      newFlame,         {create new flames at base}
      minFire,          {limit between the "starting to burn"
                         and the "is burning" routines }
      moreFire,         {change fire intensity}
      fireIncrease,     {3 = Wood, 90 = Gazolin}
      decay,            {How far should the flames go up on the screen ?}
      smooth,           {How descrete can the flames be?}
      sparks,           {new sparks per picture}
      sprinkle: integer;  {put water to fire}
      baseColor: byte = 192;

var   flameArray: array[0..1600] of byte;

{********************************************************}
procedure AnimateText;

var   xp,yp: integer;
begin
  SetFont ('Desdemona', 20, []);
  xp := Max(20, (20 + GetMaxX - TextWidth(burnText)) div 2);
  xp := 20 + xp + round(sin(picNo*0.0281) * xp);
  yp := GetMaxY div 2 - 2 * TextHeight;
  yp := yp + round(cos(picNo*0.0214) * yp);
  TextColor (baseColor);
  DisplayText (xp,yp, burnText);
  ResetFont;
end;
{********************************************************}
procedure SprinkleIntoFire;

var   xi: integer;
begin
  for xi := 1 to 10
  do flameArray[Random(pageX)] := 50;
end;
{********************************************************}
procedure DrawBaseLine;

const fireSize  = 4;
      waterSize = 4;
var   xi,yi,ci: integer;
begin
  if Random(newFlame) = 0       {new fire ?}
  then FillChar (flameArray[Random(GetMaxX - fireSize)], fireSize, 255);

  if Random(sprinkle) > 0       {put water ?}
  then FillChar (flameArray[Random(GetMaxX - waterSize)], waterSize, 100);

  {This loop controls the "root" of the flames (values in flameArray)}
  for xi := 0 to GetMaxX
  do begin
    ci := flameArray[xi];
    if ci < MinFire               {increase by the "burnability"}
    then begin                    {start burning}
      if ci > 10 then Inc (ci, Random (fireIncrease));
    end
    else {otherwise randomize and increase by intensity (is burning)}
      Inc (ci, Random (rootRand * 2 + 1) - rootRand + moreFire);
    if ci > 255 then ci := 255;   {limit fire value}
    flameArray[xi] := byte(ci);
  end;

  {smooth the flameArray values to avoid "descrete" flames}
  for xi := smooth to GetMaxX-smooth
  do begin
    ci := 0;
    for yi := -smooth to smooth do Inc (ci, flameArray[xi+yi]);
    flameArray[xi] := ci div (2*smooth+1);
  end;

  {clear fire on both edges}
  FillChar (flameArray[0], 10, 0);
  FillChar (flameArray[pageX-10], SizeOf(flameArray)-pageX, 0);

  {put values from flameArray to the bottom screen line}
  for xi := 0 to GetMaxX do PutPixel (xi, GetMaxY, flameArray[xi]);

  // draw random sparks
  for ci := 1 to sparks
  do begin
    xi := 2 + random (GetMaxX - 4);
    yi := 2 + random (GetMaxY - 4);
    pByte(GetPixelPointer(xi,yi))^ := byte (100 + random(140));
  end;
end;

{********************************************************}
procedure FireAnimation0;
var   yi: integer;
begin
  for yi := 1 to GetMaxY
  do CopyLine (backBm, 0,yi, 0,yi-1, pageX);
  Blur3H (backBm, backBm, 1, GetMaxY);
end;
{********************************************************}
procedure FireAnimation1;

var   pix: byte;
      ri, xi,yi: integer;
      pixPtr, drwPtr: pByte;
begin
  {=== This loop makes the actual flames ===}
  for yi := 1 to GetMaxY
  do begin
    pixPtr := backBm.ScanLine[yi];
    drwPtr := backBm.ScanLine[yi-1];
    ri := picNo + random (10000);
    for xi := 0 to GetMaxX - 2
    do begin
      ri := (ri + 1) and $3F;
      inc (pixPtr);
      pix := pixPtr^;
      if pix <= decay
//      then pByteArray(drwPtr)^[Random(3)] := 0
//      else pByteArray(drwPtr)^[Random(3)] := byte(pix - decay);
      then pByteArray(drwPtr)^[randomArray[ri]] := 0
      else pByteArray(drwPtr)^[randomArray[ri]] := byte(pix - decay);
      inc (drwPtr);
    end;
  end;
end;
{********************************************************}
procedure FireAnimation2;

var   xi,yi: integer;
      pixPtr: pByteArray;
      drwPtr: pByte;
begin
(*
  for yi := GetMaxY-1 to GetMaxY
  do begin
    pixPtr := backBm.ScanLine[yi];
    for xi := 10 to GetMaxX - 10
    do pixPtr[xi] := Random(150) + 100;
  end;
*)
  {=== This loop makes the actual flames ===}
  for yi := 1 to GetMaxY
  do begin
    pixPtr := backBm.ScanLine[yi];
    drwPtr := backBm.ScanLine[yi-1];
    inc (drwPtr);
    for xi := 1 to GetMaxX - 1
    do begin
      case Random(15) of
        0:   drwPtr^ := pixPtr[xi];
        1:   drwPtr^ := pixPtr[xi+1];
        2:   drwPtr^ := pixPtr[xi-1];
        else drwPtr^ := (pixPtr[xi-1] + pixPtr[xi] shl 1 + pixPtr[xi+1]) shr 2;
      end;
      inc (drwPtr);
    end;
  end;
end;

{********************************************************}
const charList: string[96]
      = ' !"#$%&''()*+,-./' +
        '0123456789:;<=>?' +
        '@ABCDEFGHIJKLMNO' +
        'PQRSTUVWXYZ[\]^_' +
        '`abcdefghijklmno' +
        'pqrstuvwxyz{|}~';

//*********************************************************
// draw blured character table
//---------------------------------------------------------
procedure DrawCharList;

const fSize = 16;
var   ci,xi,yi: integer;
begin
  SetFont ('Desdemona', fSize, []);
  TextColor (clYellow);
  for ci := 0 to sizeof(charList) - 2
  do begin
    xi := 16 + 20 * (ci mod 16);
    yi := 16 + 2 * fSize * (ci shr 4);
    DisplayText (xi, yi, charList[ci+1]);
  end;
end;
//*********************************************************
// draw one animation picture
//---------------------------------------------------------
procedure AnimDraw;

begin
  {$ifdef test2}
    writeln (tst, 'Fire.AnimDraw: ');
  {$endif}

  SelectBitmap ('backBm', backBm);

//  DrawCharList;
//  Blur4C (backBm, backBm, 0, GetMaxY);

  DrawBaseline;
//  FireAnimation0;
  FireAnimation1;

  if bluring  then Blur3S (backBm, backBm, 1, GetMaxY,1);
  if anim3d   then Rotate3d;
  if animText then AnimateText;

  CopyBitmap (backBm, bgBitmap);

end;
//*********************************************************
// handle animation input
//---------------------------------------------------------
procedure AnimInput;
begin
  {$ifdef test2}
    writeln (tst, 'Fire.AnimInput: ', ord(cb));
  {$endif}

  with valuePtr^ do begin
    rootRand     :=  IntValue (1);
    newFlame     :=  IntValue (2);
    minFire      :=  IntValue (3);
    moreFire     :=  IntValue (4);
    fireIncrease :=  Round (sqr (FloatValue(5)));
    smooth       :=  IntValue (6);
    decay        :=  IntValue (7);
    sparks       :=  IntValue (8);
    sprinkle     :=  IntValue (9);
    baseColor    :=  ByteValue (10);
  end;

  case cb of
    #0 : ;
    'r': valuePtr^.Reset;
    CR : valuePtr^.RandomValues;
    'a': animValues := not animValues;
    'b': bluring    := not bluring;
    't': animText   := not animText;
    'u': anim3d     := not anim3d;
    'w': SprinkleIntoFire;
    'e': FillChar (flameArray, SizeOf(flameArray), 0);
  end;
end;

//*********************************************************
// display animation specific help
//---------------------------------------------------------
procedure AnimHelp;
begin
  HelpText (1, '<ret> random values');
  HelpText (2, '  a   animate values');
  HelpText (3, '  b   blur ' + OnOff (bluring));
  HelpText (4, '  w   put water');
  HelpText (5, '  e   extinguish');
  HelpText (6, '  t    text');
  HelpText (7, '  u   3d object');
end;
//*********************************************************
initialization

  {$ifdef InitTest}
    writeln (tst, 'Fire.initialization: ');
    flush (tst);
  {$endif}

  RegisterAnimation (' Fire Simulation ',
    AnimStart, AnimResize, AnimDraw, AnimInput, AnimHelp);

//----------------------------------------------------------
finalization

  {$ifdef ExitTest}
    writeln (tst, 'Fire.finalization: ');
    flush (tst);
  {$endif}

end.

