//*******************************************************************
//
//  UNIT:     F l a s h                           30.09.99,  30.09.99
//
//  flash lightning demo
//
//-------------------------------------------------------------------
//
//  StartFlash             start flash demo
//  ResizeFlash            called on new window size
//  AnimateFlash           animate flash
//
//*******************************************************************
unit Flash;

interface

function StartFlash: string;
procedure ResizeFlash;
procedure AnimateFlash;

//*********************************************************
implementation

uses  SysUtils, Windows, Graphics, ColorRGB,
      GrForm, Graph, Palette, Blur;

type  tbyte = ^byte;

var   airBm: tBitmap = nil;
      cPal: tPalette;

const minLI =  33;      // minimum of light intensity
      maxLI = 200;      // maximum of light intensity

var   voltage:    integer = 55;
      resistance: integer = 20;

{================================================}
{ palette handling
{================================================}

const spreadColors = 4;

      black: tCol = (bb:255; gb:122; rb:000; ii:   0);
      flashColors: array [0..spreadColors-1] of tCol
      =((bb:255; gb:122; rb:  0; ii:   0)   // black
       ,(bb:127; gb:  0; rb:  0; ii: 210)   // darkBlue
       ,(bb:255; gb:127; rb:  0; ii: 233)   // lightBlue
       ,(bb:255; gb:255; rb:255; ii: 255)); // white

//*********************************************************
function StartFlash: string;
begin
  result := 'flash simulation';
  SetGraphMode (pf8bit);
  ResizeFlash;
end;
//*********************************************************
// create / resize lightning map
//---------------------------------------------------------
procedure ResizeFlash;
begin
  if airBm = nil
  then begin
    airBm := tBitmap.Create;
    airBm.PixelFormat := pf8Bit;
  end;
  // resize and clear air bitmap
  airBm.Width  := pageX;
  airBm.Height := pageY;
  airBm.Canvas.Brush.Color := 127;
  ClearBitmap (airBM);
  SetIntensityPalette (helpPalette);
  SetBitmapPalette (airBm, helpPalette);

  // clear voltage map and set flash palette
  bgBitmap.Canvas.Brush.Color := 127;
  ClearBitmap (bgBitmap);
  SetSpreadPalette (cPal, flashColors, spreadColors);
  SetAllPalette (cPal);
end;

//*********************************************************
//---------------------------------------------------------
procedure RandomMoleculeMoves (var dm: tBitmap);

var   xi,yi: longint;          // pixel position
      linePtr, pixPtr: ^byte;
begin
  for yi := 0 to GetMaxY
  do begin
    linePtr := airBm.ScanLine[yi];
    for xi := 0 to GetMaxX div 4
    do begin
      pixPtr := linePtr;
      inc (pixPtr, random(pageX));
      pixPtr^ := pixPtr^ + integer(random(50)) - 25;
    end;
  end;
end;

//*********************************************************
// add new loadings
//---------------------------------------------------------
procedure AddVoltage;
var
    xi: longint;          // pixel position of top line
    skyPtr: ^byte;
    volt: integer;
begin

  skyPtr := bgBitmap.ScanLine[0];
  for xi := 0 to GetMaxX
  do begin
    volt := skyPtr^ + random (voltage);
    if volt > 255
    then skyPtr^ := 255
    else skyPtr^ := volt;
    inc (skyPtr);
  end;

 // FillChar (bgBitmap.ScanLine[10]^, pageX, 255);
  FillChar (bgBitmap.ScanLine[GetMaxY]^, pageX, 0);
end;
//*********************************************************
procedure ByteSwap (var b1,b2: byte);
var x: byte;
begin
x:= b1;
b1 := b2;
b2 := x;
end;
//*********************************************************
//---------------------------------------------------------
procedure FlashVDown;

var
    xi,yi: longint;          // pixel position
    d1,d2,cx: integer;
    a0,a1: tbyte;
    p0,p1: tbyte;
begin
  for yi := 0 to GetMaxY-1
  do begin
    a0 := airBm.ScanLine[yi];
    a1 := airBm.ScanLine[yi+1];
    p0 := bgBitmap.ScanLine[yi];
    p1 := bgBitmap.ScanLine[yi+1];
    for xi := 0 to GetMaxX
    do begin
      d1 := integer(a0^) - integer(a1^);
      d2 := integer(p0^) - integer(p1^);
      if abs(d2)-abs(d1) > 0
      then begin
        cx := d2 div 2;
        p0^ := byte(integer(p0^) - cx);
        p1^ := byte(integer(p1^) + cx);
      end;
      inc (p0);
      inc (p1);
    end;
  end;
end;
//---------------------------------------------------------
procedure FlashVUp;
var
    xi,yi: longint;          // pixel position
    d1,d2, cx: integer;
    air1Ptr, air2Ptr: ^byte;
    p0, p1: ^byte;
begin
  for yi := GetMaxY-1 downto 0
  do begin
    air1Ptr := airBm.ScanLine[yi];
    air2Ptr := airBm.ScanLine[yi+1];
    p0 := bgBitmap.ScanLine[yi];
    p1 := bgBitmap.ScanLine[yi+1];
    for xi := 0 to GetMaxX
    do begin
//      d1 := integer(air1Ptr^) - integer(air2Ptr^);
      d2 := integer(p0^) - integer(p1^);
      if air1Ptr^ - random(55) + resistance > 0
      then begin
//        if abs (d2) >= resistance
//        then begin
          cx := d2 div 2;
          p0^ := p0^ - cx;
          p1^ := p1^ + d2 - cx;
//        end;
      end;
      inc (p0);
      inc (p1);
      inc (air1Ptr);
      inc (air2Ptr);
    end;
  end;
end;
//*********************************************************
//---------------------------------------------------------
procedure FlashH (offset: byte);
var
    xi,yi: longint;          // pixel position
    d1,d2, cx: integer;
    air1Ptr, air2Ptr: ^byte;
    p0, p1: ^byte;
begin
  for yi := offset to GetMaxY div 2
  do begin
    air1Ptr := airBm.ScanLine[yi];    inc (air2Ptr, offset);
    air2Ptr := air1Ptr;               inc (air2Ptr);
    p0 := bgBitmap.ScanLine[yi]; inc (p1, offset);
    p1 := p0;                    inc (p1);
    for xi := 0 to GetMaxX - 1
    do begin
//    d1 := integer(air1Ptr^) - integer(air2Ptr^);
      d2 := integer(p0^) - integer(p1^);
      if air1Ptr^ - random(55) + resistance > 0
      then begin
//        if abs (d2) >= resistance
//        then begin
          cx := d2 div 2;
          p0^ := p0^ - cx;
          p1^ := p1^ + d2 - cx;
//        end;
      end;
      inc (p0,2);
      inc (p1,2);
      inc (air1Ptr,2);
      inc (air2Ptr,2);
    end;
  end;
end;

//*********************************************************
// display transparent shadowed help text
//---------------------------------------------------------
procedure DisplayHelp;
var   h: word;
begin
  TextColor (clWhite);
  h := TextHeight;
  ShadowTextXY (10,1*h, '<ret> reset');
  ShadowTextXY (10,2*h, ' +/-  flash intensity ' + IntToStr(voltage));
  ShadowTextXY (10,3*h, ' o/k  resistance ' + IntToStr(resistance));
end;
//*********************************************************
// draw one picture and handle keyboard input
//---------------------------------------------------------
procedure AnimateFlash;
begin
  RandomMoleculeMoves (airBm);
//  Blur3H (airBm, airBm, 0, GetMaxY);
  Blur4C (airBm, airBm, 0, GetMaxY);
  AddVoltage;
  CopyBitmap (airBm, bgBitmap);
{
  case picNo and 1 of
    0: FlashVDown;
    1: FlashH (0);
    2: FlashVDown;
    3: FlashH (1);
  end;
}
  if animHelp then DisplayHelp;
  case cb of
    CR : ResizeFlash;
    '+': inc(voltage);
    '-': dec(voltage);
    'o': inc(resistance);
    'k': dec(resistance);
  end;
end;

end.

