
//*********************************************************
//
//  UNIT:     G r F o r m              10.06.99, 24.06.2003
//
//  main window for flicker free graphics animation
//  coded by Gerd Platl
//
//---------------------------------------------------------
//
//  --- global routines ---
//
//  RegisterAnimation     set animation callback routines
//  HelpText              display help text
//  ClearInput            clear input values
//
//  --- objects ---
//
//  mainForm: TMainForm   main animation window
//
//    FlashInfo              show info text for a while
//    FlashInfo2             show info text 2 for a while
//    ClearFlashInfo         clear flash information
//    SetBackground          set background bitmap
//    FillTexture            fill background with texture
//    SelectTexture          get background bitmap
//    GetRotatePicture       return copy of rotation picture
//    Restart                restart time measurement
//
//  --- global variables ---
//
//  keyPressed: boolean       if key input was detected
//  shiftState: TShiftState   key shift state
//  vKey      : word          last virtual key input
//  cb        : char          last character input
//  picNo     : longInt       endless picture counter
//
//*********************************************************
{$I comp_opt.inc}

unit GrForm;

{$ifopt D+}
  {$define xtest1    main functions}
  {$define xtest2    sub functions}
  {$define xtest3    animation functions}
  {$define xtest4    key handling}
  {$define xtest5    toggle handling}
  {$define xtest6    mouse handling}
  {$define xtest7    image handling}
  {$define xtest8    window message handling}
  {$define xtest9    exit handling}
{$endif}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Forms,Dialogs
  ,Controls, StdCtrls, ExtCtrls, Menus, JPEG, ShellAPI
  ,Fonts, Values;

procedure RegisterAnimation (windowCaption: string;
                             startProc:   tProcedure;
                             resizeProc:  tProcedure;
                             animateProc: tProcedure;
                             inputProc:   tProcedure;
                             helpProc:    tProcedure);

procedure HelpText (y: integer; textString: string);
procedure ClearInput;

type
  tMainForm = class(TForm)
    Image1: TImage;
    Texture1: TImage;
    Texture2: TImage;
    Texture3: TImage;
    Texture4: TImage;
    Texture5: TImage;
    Texture6: TImage;
    NextDemoTimer: TTimer;
    RefreshTimer: TTimer;
    x20Timer: TTimer;
    PopupMenu1: TPopupMenu;
    NextDemo: TMenuItem;
    PreviousDemo: TMenuItem;
    Size1: TMenuItem;
    Size2: TMenuItem;
    Size3: TMenuItem;
    Center1: TMenuItem;
    About: TMenuItem;
    exit: TMenuItem;
    rotate2: TImage;
    rotate3: TImage;
    rotate1: TImage;
    FullScreen: TMenuItem;
    Pause: TMenuItem;
    Step: TMenuItem;

    // form events
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);

    procedure FormMouseWheelDown(Sender: TObject; Shift: TShiftState;
      MousePos: TPoint; var Handled: Boolean);
    procedure FormMouseWheelUp(Sender: TObject; Shift: TShiftState;
      MousePos: TPoint; var Handled: Boolean);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);

    // image1 mouse events
    procedure Image1MouseMove(Sender: TObject;
                        Shift: TShiftState; X,Y: Integer);
    procedure Image1MouseDown(Sender: TObject; Button: TMouseButton;
                        Shift: TShiftState; X, Y: Integer);
    procedure Image1MouseUp(Sender: TObject; Button: TMouseButton;
                        Shift: TShiftState; X, Y: Integer);
    procedure Image1Click(Sender: TObject);
    procedure Image1DblClick(Sender: TObject);

    // timer events
    procedure RefreshPicture(Sender: TObject);
    procedure HideFlashInfo(Sender: TObject);

    // commands
    procedure SelectDemoClick(Sender: TObject);
    procedure NextDemoClick(Sender: TObject);
    procedure PreviousDemoClick(Sender: TObject);
    procedure FormCenter(Sender: TObject);
    procedure Size1Click(Sender: TObject);
    procedure Size2Click(Sender: TObject);
    procedure Size3Click(Sender: TObject);
    procedure FullScreenClick(Sender: TObject);
    procedure StepClick(Sender: TObject);
    procedure PauseClick(Sender: TObject);
    procedure AboutClick(Sender: TObject);
    procedure ExitClick(Sender: TObject);

    // public functions
    procedure WindowReplacement (x,y,w,h: integer;
                                 center: Boolean);
    procedure FlashInfo (infoText: string);
    procedure FlashInfo2(infoText: string);
    procedure ClearFlashInfo;
    procedure SetBackground (var bm: tBitmap);
    procedure FillTexture      (picIndex: word; var bm: tBitmap);
    procedure GetRotatePicture (picIndex: word; var bm: tBitmap);
    procedure Restart;

   protected
    procedure ResizePicture;

   private
    procedure WMEraseBkgnd(var msg: TWMEraseBkgnd); message WM_ERASEBKGND;
//    procedure WMCommand   (var msg: TWMCommand);    message WM_COMMAND;
//    procedure WMHitTest   (var msg: TWMNCHitTest);  message WM_NCHITTEST;
//    procedure WMMouseWheel(var msg: TWMMouseWheel); message WM_MOUSEWHEEL;
    procedure IdleHandler(Sender: TObject; var Done: Boolean);
    procedure FileIsDropped (var Msg: TMessage); Message WM_DropFiles;
    function CreateFilename (fileExt: string): string;

    function AddMainMenuItem (itemText: string): tMenuItem;
    procedure AddSubMenuItem (mainItem: tMenuItem; itemText: string);
    procedure RestartTimer;
    procedure SetRefreshMode (refMode: word);
    procedure ShowDemo(no: integer);
    procedure ShowPrevDemo;
    procedure ShowNextDemo;
    procedure ToggleHelp;
    procedure ToggleValueInput;
    procedure TogglePause;
    procedure SetStepMode;
    procedure ToggleInfo;
    procedure ToggleSyncRedraw;
    procedure TogglePalette;
    procedure ToggleHeapInfo;
    procedure ToggleColorAnimation;

    procedure DrawNextPicture;
    procedure OnePicture;

   public
    imageBm        :TBitmap;      // visible bitmap
    mouseButton    :TMouseButton; // mouse key state
    mouseX, mouseY :integer;      // mouse position
    actDemo        :integer;      // actual demo number
    x20Counter     :integer;      // 20 ticks per second counter
    bufferedDrawing:boolean;
   private
    refreshCounter :integer;      // refresh timer tick counter
    refreshMode    :word;         // refresh syncronisation mode
    refreshRate    :word;         // refresh rate [hertz]
    flashText1: string[40];       // hold info text 1
    flashText2: string[40];       // hold info text 2
  end;

var   MainForm     :tMainForm = nil;   // main window form

var   // some global flags
      demoHelp     :boolean = false;   // show specific help
      animColors   :boolean = false;   // cycle colors
      showValues   :boolean = false;   // values input
      animValues   :boolean = false;   // animate values
      heapInfo     :boolean = false;   // show heap info

      // key input handling: reseted on ...
var   keyPressed   :boolean = false;   // if key input was detected
      shiftState   :TShiftState = [];  // key shift state
      vKey         :word = 0;          // last vitual key input
      cb           :char = #0;         // last character input
      picNo        :longInt = 0;       // endless picture counter
      valuePtr     :pValues;           // pointer to actual demo values

//---------------------------------------------------------
// callback procedure handling
//---------------------------------------------------------
type
  tCallProcs = record             // -- callback procedure record --
    winCaption:  string;          // window caption
    CallStart:   tProcedure;      // start procedure
    CallResize:  tProcedure;      // resizing window
    CallAnimate: tProcedure;      // animation
    CallInput:   tProcedure;      // input handling
    CallHelp:    tProcedure;      // display help
    xxxValues:   tValues;         // demo values
  end;

var
  maxDemo        :integer = 0;    // max. registered demos
  cProcList      :array [0..50] of tCallProcs;
  cProcs         :tCallProcs;     // actual calling procedures

//*********************************************************
implementation

uses  Clipbrd, SysUtil2, Graph, Images, Palette, TpString, GrAbout;

{$R *.DFM}

const picFilePath  = 'Pics';      // path of save files
      picFileName  = 'GrAnDe';    // name of save files

type  tInfo = (noInfo, picCntInfo, fpsInfo, maxInfo);
      tHelp = (noHelp, prgHelp, specHelp, maxHelp);

const sInfo: array[tInfo] of string = ('info off','pic#','fps','max');

var
      isPaused       :boolean = false;
      showPalette    :boolean = false;
      showInfo       :tInfo = fpsInfo;
      showHelp       :tHelp = noHelp;
      selectMenuItem :tMenuItem = nil;

//*********************************************************
// clear all input values
//---------------------------------------------------------
procedure ClearInput;
begin
  // reset key input
  keyPressed := false;
  shiftState := [];
  cb := #0;        // clear character input
  vKey := 0;       // clear key input

end;
//*********************************************************
// add menu item to main menu
//---------------------------------------------------------
function TMainForm.AddMainMenuItem (itemText: string): tMenuItem;
var   newItem: tMenuItem;
begin
  {$ifdef test3}
    writeln (tst, 'GrForm.AddMainMenuItem: ', itemText);
  {$endif}

  newItem := tMenuItem.Create (PopupMenu1);
  newItem.caption := itemText;
  PopupMenu1.items.insert (0, newItem);
  result := newItem;
end;
//*********************************************************
// add sub menu item to main menu item
//---------------------------------------------------------
procedure TMainForm.AddSubMenuItem (mainItem: tMenuItem; itemText: string);
var   newItem: tMenuItem;
begin
{$ifdef test3}
  writeln (tst, 'GrForm.AddSubMenuItem: ', itemText);
{$endif}

  newItem := tMenuItem.Create (mainItem);
  newItem.caption := itemText;
  newItem.OnClick := SelectDemoClick;
  mainItem.OnClick := nil;
  mainItem.insert (mainItem.count, newItem);
end;

//*********************************************************
// execute callback function on resize window
//---------------------------------------------------------
procedure TMainForm.ResizePicture;
begin
  {$ifdef test2}
    writeln (tst, 'GrForm.ResizePicture: '
                , HexString (@cProcs.CallResize));
  {$endif}

  //-------------------------------------------------------
  with cProcs do if @CallResize <> nil then CallResize;
  //-------------------------------------------------------
  Restart;
end;

//*********************************************************
procedure TMainForm.FormCreate(Sender: TObject);
var i1: longint;
begin
  {$ifdef test1}
    writeln (tst, 'GrForm.FormCreate: ');
  {$endif}

  // use idle handler for paint events
  Application.OnIdle := IdleHandler;   // add idle handler
  DragAcceptFiles (handle, True);      // to drag and use pictures

  // init attributes
  imageBm := image1.picture.bitmap;    // set image bitmap pointer
  shiftState := [];
  mouseButton := TMouseButton(0);
  actDemo := 0;
  demoHelp := showHelp = specHelp;
  x20Counter := 0;
  bufferedDrawing := True;             // =true: use bgBitmap

  refreshCounter := 0;                 // 100 times per second
  refreshMode := 1;                    // 50 Hertz
  refreshRate := 50;                   // refresh rate [hertz]

  // create demo selection menu
  selectMenuItem := AddMainMenuItem ('select demo');
  for i1 := 0 to maxDemo - 1
  do with cProcList[i1]
  do AddSubMenuItem (selectMenuItem, winCaption);

  MarkTime;
end;
//*********************************************************
procedure TMainForm.FormDestroy(Sender: TObject);
begin
  {$ifdef test1}
    writeln (tst, 'GrForm.FormDestroy: ');
  {$endif}
end;
//*********************************************************
procedure TMainForm.FormActivate(Sender: TObject);
begin
  {$ifdef test1}
    writeln (tst, 'GrForm.FormActivate: ');
  {$endif}

  clientWidth := image1.Width;
  clientHeight := image1.Height;
  SetBitmapSize (clientWidth, clientHeight);
  mouseX := clientWidth div 2;
  mouseY := clientHeight div 2;
  SelectBitmap ('bg', bgBitmap);
  ShowDemo (0);
  FlashInfo ('F1 help     F10 about');
end;
//*********************************************************
procedure TMainForm.FormPaint(Sender: TObject);
begin
  {$ifdef test1}
    writeln (tst, 'GrForm.FormPaint: ');
  {$endif}
end;
//*********************************************************
procedure TMainForm.FormResize(Sender: TObject);
var   cw: integer;
begin
  // round to 4 pixel steps
  cw := (self.clientWidth + 3) and $FFFFFFFC;

  {$ifdef test1}
    writeln (tst, 'GrForm.FormResize: ',cw,'*',clientHeight);
  {$endif}

  image1.Width  := cw;
  image1.Height := clientHeight;
  image1.picture.bitmap.width  := cw;
  image1.picture.bitmap.height := clientHeight;
  SetBitmapSize (cw, clientHeight);
  SetBitmapSize (imageBm,  cw, clientHeight);
  SetBitmapSize (bgBitmap, cw, clientHeight);
  SelectBitmap ('bg', bgBitmap);
  //-------------------------------------------------------
  ResizePicture;
  //-------------------------------------------------------
end;
//*********************************************************
// event on pressing a key
//---------------------------------------------------------
procedure TMainForm.FormKeyDown (Sender: TObject;
                        var Key: Word; Shift: TShiftState);
begin
  keyPressed := true;
  vKey := key;
  shiftState := shift;

  {$ifdef test4}
    writeln (tst, 'GrForm.FormKeyDown: ',key:5,byte(Shift):4);
  {$endif}

  if Shift = []
  then case key of
    vk_F1:      ToggleHelp;
    vk_F2:      ToggleValueInput;
    vk_F3:      TogglePalette;
    vk_F4:      ToggleSyncRedraw;
    vk_F5:      ToggleInfo;
    vk_F6:      ToggleColorAnimation;

    vk_F8:      SetStepMode;
    vk_F9,
    vk_Pause:   TogglePause;
    vk_F11:     ToggleHeapInfo;
    vk_Escape,
    vk_Cancel:  Close;                 // quit application
    vk_Back:    ShowPrevDemo;          // show previous demo
  end;
  if isPaused then DrawNextPicture;

  // type TShiftState = set of (ssShift, ssAlt, ssCtrl,
//                    ssLeft, ssRight, ssMiddle, ssDouble);
//  if Shift = [ssAlt, ssCtrl]
//
//  end;

  {$ifdef test4}
    // show key value - only for test purpose
    SelectCanvas (image1.Canvas);
    TextColor (clLightMagenta);
    OutTextXY (46, pageY - 19, IntToStr (byte(shift)) + ' ' + IntToStr(key));
    RestoreCanvas;
  {$endif}

  RestartTimer;    // reset demo timeout
end;
//*********************************************************
// create file name for saving pictures
//---------------------------------------------------------
function TMainForm.CreateFilename (fileExt: string): string;
begin
  if DirectoryExists (picFilePath)
  then Result := picFilePath + '\' + picFileName
  else Result := picFileName;
  Result := Result + IntString (actDemo, -2) + fileExt;
end;

//*********************************************************
// event on key is pressed (key up)
//---------------------------------------------------------
procedure TMainForm.FormKeyPress(Sender: TObject; var Key: Char);
begin
  cb := key;

  {$ifdef test4}
    writeln (tst, 'GrForm.FormKeyPress: ',byte(key):4,byte(shiftState):4);
  {$endif}

  if ssCtrl in shiftState              // Ctrl pressed
  then case cb of
    '5': ClearDevice;
    #3: Clipboard.Assign(bgBitmap);    // copy bitmap to clipboard
    #2: SaveBMPBitmap (bgBitmap, CreateFilename ('.BMP'));  //Ctrl-b
    #19: SaveJPGBitmap(bgBitmap, CreateFilename ('.JPG'),100,true); //Ctrl-s
    #16: Graph.Print;                  // call bitmap print function      Ctrl-p
  end
  else if shiftState = []
  then case cb of
    ' ': ShowNextDemo;
    'h': ToggleHelp;
    'p': TogglePause;
    else if (key >= '0') and (key <= '9') then ShowDemo (ord(key)-48);
  end
  else if isPaused then DrawNextPicture;

  {$ifdef test4}
    // show key value - only for test purpose
    SelectCanvas (image1.Canvas);
    TextColor (clLightCyan);
    OutTextXY (120,pageY - 19, key + ' ' +IntToStr(ord(key)));
    RestoreCanvas;
  {$endif}

  RestartTimer;    // reset demo timeout
end;

//*********************************************************
procedure TMainForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  {$ifdef test9}
    writeln (tst, 'GrForm.FormClose: action=', ord(Action));
  {$endif}
end;
//*********************************************************
procedure TMainForm.Image1MouseMove(Sender: TObject;
                               Shift: TShiftState; X,Y: Integer);
begin
  {$ifdef test6}
    writeln (tst, '  mouseMove: ',x:3,',',y:3, '  ',byte(Shift):2);
  {$endif}
  mouseX := x;
  mouseY := y;
end;
//*********************************************************
procedure TMainForm.Image1MouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  {$ifdef test6}
    writeln (tst, '  mouseDown: ',x:3,',',y:3
                ,'  ', byte(Shift):2, '  ', ord(Button));
  {$endif}
  mouseButton := Button;
  shiftState  := shift;
  mouseX := x;
  mouseY := y;
end;
//*********************************************************
procedure TMainForm.Image1MouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  {$ifdef test6}
    writeln (tst, '  mouseUp:   ',x:3,',',y:3
                ,'  ', byte(Shift):2, '  ', ord(Button));
  {$endif}
  mouseButton := button;
//  shiftState  := shift;  don't overwrite button down shift state !
  mouseX := x;
  mouseY := y;
end;
//*********************************************************
procedure TMainForm.Image1Click(Sender: TObject);
begin
  {$ifdef test6}
    writeln (tst, '  mouseClick: ');
  {$endif}
  case mouseButton of
    mbLeft: ShowNextDemo;
//    mbRight:
    mbMiddle: ShowDemo (actDemo);    // restart with next picture
  end;
end;
//*********************************************************
procedure TMainForm.Image1DblClick(Sender: TObject);
begin
  {$ifdef test6}
    writeln (tst, '  mouseDoubleClick: ');
  {$endif}
  ShowDemo (actDemo);    // restart with next picture
end;
//*********************************************************
// show previous demo
//---------------------------------------------------------
procedure TMainForm.FormMouseWheelDown(Sender: TObject;
  Shift: TShiftState; MousePos: TPoint; var Handled: Boolean);
begin
  if shift = [ssMiddle]
  then ShowNextDemo
  else begin
    Cursor := crSizeNS;    // arrow up/down
    PrevDemoPicture;       // get back 2 picture files
    ShowDemo (actDemo);    // restart with next picture
    Cursor := crDefault;   // default arrow
  end;
end;
//*********************************************************
procedure TMainForm.FormMouseWheelUp(Sender: TObject; Shift: TShiftState;
  MousePos: TPoint; var Handled: Boolean);
begin
  if shift = [ssMiddle]
  then ShowPrevDemo
  else begin
    Cursor := crSizeNS;    // arrow up/down
    ShowDemo (actDemo);    // restart with next picture
    Cursor := crDefault;   // default arrow
  end;
end;

//*********************************************************
// handle picture refresh timer event
//---------------------------------------------------------
procedure TMainForm.RefreshPicture(Sender: TObject);
begin
  {$ifdef test3}
    writeln (tst, 'GrForm.RefreshPicture: ');
  {$endif}
  inc (refreshCounter);
end;
//*********************************************************
// handle x20Timer event (20 times per second)
//---------------------------------------------------------
procedure TMainForm.HideFlashInfo(Sender: TObject);
begin
  inc (x20Counter);
  if x20Counter = 40     // 2 secs ?
  then ClearFlashInfo;
end;

//=================
// public functions
//=================

//*********************************************************
// change position and / or size of window and center
// input:   x, y    window position
//          w, h    size of client window
//          center  center window on screen
//---------------------------------------------------------
procedure TMainForm.WindowReplacement (x,y,w,h: integer;
                                         center: Boolean);
var   bs,th: longint;
begin
  bs := (width - clientWidth) div 2;
  th := height - clientHeight;
  w := Clamp (w, 160, screenInfo.X);
  h := Clamp (h, 120, screenInfo.Y);
  if center
  then begin
    x := (screenInfo.X - w) div 2 - bs;
    y := (screenInfo.Y - h) div 2 - th;
  end;
  {$ifdef test2}
    writeln (tst, 'GrForm.WindowReplacement: ', x:5,y:5,w:5,h:5);
  {$endif}
  SetBounds (x,y, w+2*bs, h+th);
  FlashInfo ('window ' + IntToStr (clientWidth)
                 + '*' + IntToStr (clientHeight));
  FlashInfo2('screen ' + IntToStr (screenInfo.X)
                 + '*' + IntToStr (screenInfo.Y));
end;
//*********************************************************
// show info text for a while
//---------------------------------------------------------
procedure TMainForm.FlashInfo (infoText: string);
begin
  {$ifdef test1}
    writeln (tst, 'GrForm.FlashInfo: ');
  {$endif}
  x20Counter := 0;
  flashText1 := infoText;
end;
//*********************************************************
// show info text 2 for a while
//---------------------------------------------------------
procedure TMainForm.FlashInfo2 (infoText: string);
begin
  x20Counter := 0;
  flashText2 := infoText;
end;
//*********************************************************
// clear flash information text
//---------------------------------------------------------
procedure TMainForm.ClearFlashInfo;
begin
  flashText1 := '';
  flashText2 := '';
end;
//*********************************************************
// set given bitmap as background picture
//---------------------------------------------------------
procedure TMainForm.SetBackground (var bm: tBitmap);
begin
  {$ifdef test6}
    writeln (tst, 'GrForm.SetBackground: ');
  {$endif}
  if bm.PixelFormat = pf8bit then UseBitmapPalette (bm);
  bgBitmap.Canvas.StretchDraw (Rect(0,0,pageX,pageY), bm);
end;
//*********************************************************
//  select texture image included in this form (DFM file)
//  with given index and fill background with this texture
//---------------------------------------------------------
procedure TMainForm.FillTexture (picIndex: word;
                                   var bm: tBitmap);
var   texImage: tImage;
      tileBm: tBitmap;
begin
  {$ifdef test6}
    writeln (tst, 'GrForm.FillTexture: ', picIndex);
  {$endif}
  case picIndex of
    2: texImage := texture2;
    3: texImage := texture3;
    4: texImage := texture4;
    5: texImage := texture5;
    6: texImage := texture6;
    else texImage := texture1;
  end;
  tileBm := tBitmap.Create;
  tileBm.PixelFormat := screenInfo.pixelFormat;
  CopyImageToBitmap (texImage, tileBm);
  DrawTilesBackground (0,0, tileBm, bm);
  tileBm.Free;
end;
//*********************************************************
//  select rotation image of this form (DFM file)
//---------------------------------------------------------
procedure TMainForm.GetRotatePicture (picIndex: word;
                                        var bm: tBitmap);
var   rotImage: tImage;
begin
  {$ifdef test6}
    writeln (tst, 'GrForm.GetRotatePicture: ', picIndex);
  {$endif}
  case picIndex of
    1: rotImage := rotate1;
    2: rotImage := rotate2;
    else rotImage := rotate3;
  end;
//  bm.Free;
  bm := tBitmap.Create;
  bm.PixelFormat := pf8Bit;
  CopyImageToBitmap (rotImage, bm);
  UseBitmapPalette (bm);
end;

//================
// command section
//================

//*********************************************************
procedure TMainForm.SelectDemoClick(Sender: TObject);

var   demoNo: integer;
begin
  if sender is tMenuItem               // check object type
  then begin
    demoNo := tMenuItem(sender).command;
    demoNo := demoNo - selectMenuItem.items[0].command;
    {$ifdef test6}
      writeln (tst, 'GrForm.SelectDemoClick=', demoNo);
    {$endif}
    ShowDemo (demoNo);
  end;
end;
//*********************************************************
procedure TMainForm.NextDemoClick(Sender: TObject);
begin
  ShowNextDemo;
end;
//*********************************************************
procedure TMainForm.PreviousDemoClick(Sender: TObject);
begin
  ShowPrevDemo;
end;
//*********************************************************
procedure TMainForm.FormCenter(Sender: TObject);
begin
  WindowReplacement (top,left, ClientWidth, ClientHeight, TRUE);
end;
//*********************************************************
procedure TMainForm.Size1Click(Sender: TObject);
begin
  WindowReplacement (top, left, 480, 360, TRUE);
end;
//*********************************************************
procedure TMainForm.Size2Click(Sender: TObject);  // grow
begin
  WindowReplacement (top,left, ClientWidth  * 3 div 2,
                               ClientHeight * 3 div 2, TRUE);
end;
//*********************************************************
procedure TMainForm.Size3Click(Sender: TObject); // shrink
begin
  WindowReplacement (left,top, (ClientWidth  div 3) * 2,
                               (ClientHeight div 3) * 2, TRUE);
end;
//*********************************************************
procedure TMainForm.FullScreenClick(Sender: TObject);
begin
  WindowReplacement (0,0,screenInfo.X, screenInfo.Y, TRUE);
end;
//*********************************************************
procedure TMainForm.StepClick(Sender: TObject);
begin
  SetStepMode;
end;
//*********************************************************
procedure TMainForm.PauseClick(Sender: TObject);
begin
  TogglePause;
end;
//*********************************************************
procedure TMainForm.AboutClick(Sender: TObject);
begin
  AboutBox.Show;
end;
//*********************************************************
procedure TMainForm.ExitClick(Sender: TObject);
begin
{$ifdef test9}
  writeln (tst, 'GrForm.ExitClick: ');
{$endif}

  Close;
end;

//================
// private section
//================

//*********************************************************
// since we're going to be painting the whole form,
// handling this message will suppress the unneccessary
// repainting of the background which can result in flicker
//---------------------------------------------------------
procedure TMainForm.WMEraseBkgnd(var msg: TWMEraseBkgnd);
begin
  {$ifdef test8}
    writeln (tst, 'GrForm.WMEraseBkgnd: ', msg.msg);
  {$endif}
  msg.Result := 1;
end;

(*
//---------------------------------------------------------
procedure TMainForm.WMCommand(var msg: TWMCommand);
begin
{$ifdef test8}
  writeln ('GrForm.WMCommand: msg.itemid=', msg.ItemID);
{$endif}
  Inherited;
end;
//---------------------------------------------------------
procedure TMainForm.WMHitTest(var msg: TWMNCHitTest);
begin
{$ifdef test2}
  writeln (tst, 'GrForm.WMHitTest: ', msg.Result);
{$endif}
  DefaultHandler (msg);
  if msg.Result = HTCLIENT
  then if mouseButton = mbRight
  then if mouseY < (pageY div 2)
  then msg.Result := HTCAPTION;
  //writeln (tst, '  hitMouse: ',mousex,'*',mousey);
end;
*)
//*********************************************************
// idle handler is called e.g. if about box is open
//---------------------------------------------------------
procedure TMainForm.IdleHandler(Sender: TObject; var Done: Boolean);
begin
  {$ifdef test3}
    writeln (tst, 'GrForm.IdleHandler: ');
  {$endif}
  done := false;
  OnePicture;
end;

//*********************************************************
// this routine will be called if a file is dropped
//---------------------------------------------------------
procedure TMainForm.FileIsDropped (var Msg: TMessage);

var   hDrop: THandle;
      cDrop: array[0..255] of char;
      filename: string;
begin
  hDrop := Msg.WParam;
  DragQueryFile (hDrop, 0, cDrop, 254);
  DragFinish (hDrop);
  filename := string (cDrop);
  Add2PictureList (filename);
  FlashInfo (filename);
  ShowDemo (actDemo);        // restart actual demo with new picture
  BringToFront;
  // todo: after drop 256 color bmp picture jpg picture will
  // be reduced to 256 color first and than expanded to true color ???
end;

//*********************************************************
// restart demo timer and picture counter
//---------------------------------------------------------
procedure TMainForm.Restart;
begin
  {$ifdef test1}
    writeln (tst, 'GrForm.Restart: ');
  {$endif}
  nextDemoTimer.enabled := false;
  nextDemoTimer.enabled := true;
  MarkTime;
end;
//*********************************************************
// restart demo timer (reset of timer counter)
//---------------------------------------------------------
procedure TMainForm.RestartTimer;
begin
  {$ifdef test1}
    writeln (tst, 'GrForm.RestartTimer: ');
  {$endif}
  nextDemoTimer.enabled := false;
  nextDemoTimer.enabled := true;
end;

//*********************************************************
const maxFreq = 3;      // number of frequency modes
                        // 4: include monitor frequence
                        //    do not work with this timer !
//*********************************************************
// set animation refresh frequency mode
// refMode: 0=off 1=50 2=100 3=refRate
//---------------------------------------------------------
procedure TMainForm.SetRefreshMode (refMode: word);
begin
  refreshMode := refMode mod maxFreq;
  {$ifdef test2}
    writeln (tst, 'GrForm.SetRefreshMode: ', refreshMode);
  {$endif}
  if refreshMode = 0
  then FlashInfo ('Sync = off')
  else begin
    case refreshMode of
      1: refreshRate := 50;
      2: refreshRate := 100;
      3: if  (screenInfo.refreshRate > 50)
         and (screenInfo.refreshRate < 200)
         then refreshRate := screenInfo.refreshRate
         else refreshRate := 75;
    end;
    RefreshTimer.interval := 1000 div refreshRate;
    FlashInfo ('Sync = ' + IntToStr (refreshRate) + ' Hz');
  end;
  Restart;
end;
//*********************************************************
// show demo with given number
//---------------------------------------------------------
procedure TMainForm.ShowDemo (no: integer);
begin
  {$ifdef test2}
    writeln (tst, 'GrForm.ShowDemo: ', no);
  {$endif}
  actDemo := (no + maxDemo) mod maxDemo;
  isPaused := false;
  flashText1 := '';
  flashText2 := '';
  cProcs := cProcList[actDemo];   // set actual calling procedures
  caption := cProcs.winCaption;
  valuePtr := addr(cProcs.xxxValues);
  //--------------------------------------------------------
  with cProcs do begin
    if @CallStart  <> nil then CallStart;
    if @CallResize <> nil then CallResize;
  end;
  //--------------------------------------------------------
  Restart;
end;
//*********************************************************
procedure TMainForm.ShowNextDemo;
begin
{$ifdef test1}
  writeln (tst, 'GrForm.ShowNextDemo: ', actDemo + 1);
{$endif}
  if (cb = ' ') and isPaused
  then TogglePause
  else ShowDemo (actDemo + 1);
end;
//*********************************************************
procedure TMainForm.ShowPrevDemo;
begin
{$ifdef test1}
  writeln (tst, 'GrForm.ShowPrevDemo: ', actDemo - 1);
{$endif}
  if (cb = ' ') and isPaused
  then TogglePause
  else ShowDemo (actDemo - 1);
end;
//*********************************************************
procedure TMainForm.ToggleHelp;
begin
  showHelp := tHelp((ord(showHelp)+1) mod ord(maxHelp));
  demoHelp := showHelp = specHelp;
  nextDemoTimer.enabled := (showHelp = noHelp);
end;
//*********************************************************
procedure TMainForm.ToggleValueInput;
begin
  showValues := not showValues;
end;
//*********************************************************
procedure TMainForm.TogglePause;
begin
  {$ifdef test5}
    writeln (tst, 'GrForm.TogglePause: ', not isPaused);
  {$endif}
  isPaused := not isPaused;
  nextDemoTimer.Enabled := not isPaused;
  if isPaused
  then begin
    DrawNextPicture;    // to clear freq. display
    caption := cProcs.winCaption + '  p.a.u.s.e.d';
  end
  else caption := cProcs.winCaption;
end;
//*********************************************************
procedure TMainForm.SetStepMode;
begin
  caption := cProcs.winCaption + '   <F8> next picture';
  isPaused := true;
  nextDemoTimer.Enabled := false;
  DrawNextPicture;
end;
//*********************************************************
procedure TMainForm.ToggleInfo;
begin
  showInfo := tInfo((ord(showInfo)+1) mod ord(maxInfo));
  FlashInfo (sInfo[showInfo]);
end;
//*********************************************************
// switch synchronisation on / off
// if sync is on it will be set to 3 different frequency
//---------------------------------------------------------
procedure TMainForm.ToggleSyncRedraw;
begin
  {$ifdef test5}
    writeln (tst, 'GrForm.ToggleSyncRedraw: ');
  {$endif}
  SetRefreshMode (refreshMode + 1);
end;
//*********************************************************
procedure TMainForm.TogglePalette;
begin
  showPalette := not showPalette;
  {$ifdef test5}
    writeln (tst, 'GrForm.TogglePalette: ', showPalette);
  {$endif}
  if not showPalette
  then ResizePicture;   // to restore background
  DrawNextPicture;
end;
//*********************************************************
procedure TMainForm.ToggleHeapInfo;
begin
  heapInfo := not heapInfo;
end;
//*********************************************************
procedure TMainForm.ToggleColorAnimation;
begin
  animColors := not animColors;
end;

//*********************************************************
procedure DisplayGeneralHelp;
begin
  HelpText (1, '  0..9   direct demo selection');
  HelpText (2, '    f       fps / pic# / off');
  HelpText (3, '    p      pause on/off');
  HelpText (4, '    r       reset parameters');
  HelpText (5, ' <F10>     about this demo');
  HelpText (6, '<back>   previous demo');
  HelpText (7, '<blanc>  next demo');
  HelpText (8, ' <esc>    quit');
end;
//*********************************************************
// draw next picture
// - show palette if activated and palette exists to bgBitmap
// - copy background image bgBitmap to image1
// - call drawing procedure
// - display help if activated
// - trigger redraw if step mode is active
//---------------------------------------------------------
procedure TMainForm.DrawNextPicture;
var   infoText: string;
begin
  {$ifdef test3}
    writeln (tst, 'GrForm.DrawNextPicture: ', picNo);
  {$endif}

  inc (picNo);    // inc endless counter - reset at demo start or change
  inc (picCnt);   // inc picture counter - used to get FPS value

  // copy background bitmap to form image of window
//  image1.Canvas.Draw (0,0, bgBitmap);

  // draw actual picture: bgBitmap -> images1

//  BitmapMirrorHorizontal (bgBitmap, image1.picture.bitmap);
//  BitmapMirrorVertical   (bgBitmap, image1.picture.bitmap);

  try
    with cProcs do begin

      CallAnimate;                     // ==> call drawing procedure

      if showPalette then DrawRGBPalette (32,0,16);

      if bufferedDrawing
      then BitBlt (image1.Canvas.handle, 0,0, pageX, pageY
                ,bgBitmap.Canvas.handle, 0,0, SrcCopy);

      SelectBitmap ('image1', imageBm);   // prepare text output
      ResetFont;
      case showHelp of
        prgHelp:  DisplayGeneralHelp;
        specHelp: if @CallHelp <> nil
                  then CallHelp;       // ==> call drawing procedure
      end;

      if showValues
      then begin
        xxxValues.Paint;
        xxxValues.KeyDown (vKey, shiftState);
      end;
      if animValues then valuePtr^.Animate;

      if keyPressed and (@CallInput <> nil)
      then CallInput;                  // ==> call input procedure

      if not isPaused Then
      case showInfo of
        picCntInfo: infoText := IntToStr (picCnt);
        fpsInfo:    infoText := FPS;
        else        infoText := '';
//        else       infoText := IntToStr (refreshCounter);
//        else       infoText := 'fic='+IntToStr (flashInfoCounter);
      end;
      // draw info as transparent shaded text
      if infoText <> ''
      then begin
        TextColor (clYellow);
        ShadowText (5,pageY - 20,infoText);
      end;
      if flashText1 <> ''
      then begin
        TextColor (clCyan);
        ShadowText ((pageX - TextWidth(flashText1)) div 2,
                     pageY div 2,  flashText1);
        if flashText2 <> ''
        then ShadowText ((pageX - TextWidth(flashText2)) div 2,
                          pageY div 2 + 24, flashText2);
      end;
      SelectBitmap ('bg', bgBitmap);
    end;
  except
    ShowNextDemo;
  end;
  mainForm.bufferedDrawing := true;
  ClearInput;
  Invalidate;      // redraw trigger for step mode
end;

//*********************************************************
// draw one picture and show active info text
//---------------------------------------------------------
procedure TMainForm.OnePicture;
begin
  {$ifdef test3}
    writeln (tst, 'GrForm.OnePicture: ');
  {$endif}
  try
    if isPaused
    then begin
      Sleep(100);
      MarkTime;
    end
    else DrawNextPicture;

  finally
    application.ProcessMessages;  // !!!
    if refreshMode > 0
    then begin
      while refreshCounter = 0
      do application.ProcessMessages;  // !!!
      refreshCounter := 0;   // reset refresh counter
    end;
  end;
end;

//*********************************************************
// display shadowed help text with default font
//---------------------------------------------------------
procedure HelpText (y: integer; textString: string);
begin
  ShadowText (10,y*16, textString);
end;

//*********************************************************
// central routine to register all callback routines
// of one animation unit
// input:      windowCaption: window title
//             startProc:     called once on starting next demo
//             resizeProc:    called, if window size changes
//             animateProc:   draw next animation picture
//             inputProc:     handle input
//             helpProc:      show demo specific help
//---------------------------------------------------------
procedure RegisterAnimation (windowCaption: string;
                             startProc:   tProcedure;
                             resizeProc:  tProcedure;
                             animateProc: tProcedure;
                             inputProc:   tProcedure;
                             helpProc:    tProcedure);
begin
  {$ifdef Test1}
    writeln (tst, 'GrForm.RegisterAnimation: ',maxDemo:3, windowCaption);
    flush (tst);
  {$endif}

  try
    if @animateProc = nil
    then Raise EInvalidPointer.Create(#13#10'Animation procedure is empty ?' +
      #13#10'Please correct RegisterAnimation call of "' +
      windowCaption + '"'#13#10);
    with cProcList[maxDemo]
    do begin
      winCaption  := windowCaption;
      CallStart   := startProc;
      CallResize  := resizeProc;
      CallAnimate := animateProc;
      CallInput   := inputProc;
      CallHelp    := helpProc;
      xxxValues   := tValues.Create;
      xxxValues.Init;
      valuePtr    := addr(xxxValues);
    end;
    inc (maxDemo);
  except
    // oh no
  end;
end;

//*********************************************************
procedure TMainForm.FormMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  Cursor := crSizeAll;
end;

//*********************************************************
procedure TMainForm.FormMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  Cursor := crDefault;
end;

//*********************************************************
initialization
  {$ifdef InitTest}
    writeln (tst, 'GrForm.initialization: ');
    flush (tst);
  {$endif}

//  syncRedraw := false;

//---------------------------------------------------------
finalization
  {$ifdef ExitTest}
    writeln (tst, 'GrForm.finalization: ');
    flush (tst);
  {$endif}
end.

