
//*******************************************************************
//
//  UNIT:     G r a p h                          10.06.99, 02.07.2002
//
//  Good old Borland Pascal replacement unit for delphi 5.
//  coded by Gerd Platl
//
//  This should only be a helpful unit for programmers who want
//  make a fast adaptation from the old Borland Pascal sources.
//  And you can see what new statements you have to use
//  instead of the old graph-unit function calls.
//
//  Some functions are only used for experimental purposes.
//
//-------------------------------------------------------------------
//  SetGraphMode        select window size
//  ClearDevice         clear window
//  ClearViewport       clear viewport with background color
//  GetMaxX             return window width
//  GetMaxY             return window height
//  GetPixel            return pixel color
//  PutPixel            set pixel with given color
//  SetPixel            set pixel with active color
//  XorPixel            xor pixel with given color
//  SetAspectRatio      set aspect ratio
//  GetAspectRatio      get aspect ratio
//
//  -- additional --
//  SetFillColor        set fill color
//  SetIntensity        set fill color intensity
//
//  GetStartPointer     return bitmap data starting address
//  GetLinePointer      return scanline address
//  GetPixelPointer     return pixel address
//
//*******************************************************************
{$I comp_opt.inc}

unit Graph;

{$ifopt D+}
  {$define xtest1    main functions}
  {$define xtest2    sub functions}
  {$define xtest3    help functions}
  {$define xtest4    color functions}
{$endif}

interface

uses  SysUtils, Windows, Graphics, Consts, Classes, Printers,
      ColorRGB;

const NUL = #0;
      BEL = #7;
      DEL = #8;
      TAB = #9;
      LF  = #10;
      FF  = #12;
      CR  = #13;
      ESC = #27;
      BLANC = #32;
      CRLF = #13#10;

{-- graph defines ---}

      { palette indizes for pixelFormat = pf8bit }
const iBlack     = 0;    iDarkGray     = 8;
      iBlue      = 1;    iLightBlue    = 9;
      iGreen     = 2;    iLightGreen   = 10;
      iCyan      = 3;    iLightCyan    = 11;
      iRed       = 4;    iLightRed     = 12;
      iMagenta   = 5;    iLightMagenta = 13;
      iBrown     = 6;    iYellow       = 14;
      iLightGray = 7;    iWhite        = 15;

const maxColor  = 255;

{true colors values:     b g r}
const clBlack        = $000000;
      clMaroon       = $000080;
      clDkGreen      = $008000;
      clOlive        = $008080;
      clNavy         = $800000;
      clPurple       = $800080;
      clTeal         = $808000;
      clGray         = $808080;
      clSilver       = $C0C0C0;
      clRed          = $0000FF;
      clLime         = $00FF00;
      clYellow       = $00FFFF;
      clBlue         = $FF0000;
      clFuchsia      = $FF00FF;
      clAqua         = $FFFF00;
      clLtGray       = $C0C0C0;
      clDkGray       = $808080;
      clWhite        = $FFFFFF;

{ additional colors:     b g r}
      clGreen        = $00FF00;
      clDarkRed      = $000080;
      clLightRed     = $8080FF;
      clDarkCyan     = $808000;
      clCyan         = $FFFF00;
      clLightCyan    = $FFFF80;
      clDarkMagenta  = $800080;
      clMagenta      = $FF00FF;
      clLightMagenta = $FF80FF;
      clLightBlue    = $FF8080;
      clLightGreen   = $80FF80;
      clBrown        = $004444;
      clDarkGray     = $404040;
      clLightGray    = $C0C0C0;
      clOrange       = $0080FF;

      { BitBlt operators for PutImage: }
const NormalPut     = 0;    { MOV }       { left for 1.0 compatibility }
      CopyPut       = 0;    { MOV }
      XORPut        = 1;    { XOR }
      OrPut         = 2;    { OR  }
      AndPut        = 3;    { AND }
      NotPut        = 4;    { NOT }
      XNORPut       = 5;    {additional definitions}
      NORPut        = 6;
      NANDPut       = 7;
      TransCopyPut  = 8;

const
      clipOn = true;
      clipOff = false;

type
  LineSettingsType = record
      LineStyle : word;
      Pattern   : word;
      Thickness : word;
    end;

  TextSettingsType = record
      Font      : word;
      Direction : word;
      CharSize  : word;
      Horiz     : word;
      Vert      : word;
    end;

  FillSettingsType = record               { Pre-defined fill style }
      Pattern : word;
      Color   : word;
    end;

  FillPatternType = array[1..8] of byte;  { User defined fill style }

  PointType = record
      X, Y : smallInt;
    end;

  ViewPortType = record
      x1, y1, x2, y2 : integer;
      Clip           : boolean;
    end;

  ArcCoordsType = record
      X, Y           : integer;
      Xstart, Ystart : integer;
      Xend, Yend     : integer;
    end;

var   GetMaxX           :integer = 639;     // max. X-Koordinate
      GetMaxY           :integer = 479;     // max. Y-Koordinate
      pageX             :word    = 640;     // window size
      pageY             :word    = 480;
      GetMaxColor       :word    = 255;     // max. color index
      GetColor          :tColor  =clYellow; // drawing color
      GetColorIndex     :byte    = 15;      // palette index drawing color
      GetBkColor        :tColor  = 0;       // brush color
      GetBkColorIndex   :byte    = 0;       // brush color palette index
      GetFillColor      :tColor  =clBlue;   // fill color
      GetFillColorIndex :byte    = 0;       // fill color palette index
      GetIntensity      :byte    = 255;     // fill color intensity
      GetTextColor      :tColor  =clYellow;
      GetTextBackground :tColor  =clBlack;
      GetGraphMode      :tPixelFormat = pf8Bit; {aktueller Grafikmodus}

procedure SetGraphMode (graphMode: tPixelFormat);

{ *** Screen, viewport, page routines *** }
procedure ClearDevice;
procedure ClearViewPort;
procedure SetViewPort (x1,y1, x2,y2: integer; clip: boolean);
procedure GetViewSettings (var viewPort: viewPortType);
procedure GetAspectRatio (var Xasp, Yasp: Word);
procedure SetAspectRatio (var Xasp, Yasp: Word);

function GetX: integer;
function GetY: integer;

function  GetPixel (x,y: integer): tColor;
procedure PutPixel (x,y: integer; color: tColor);          overload;
procedure PutPixel (x,y: integer; r,g,b: byte);            overload;
procedure PutPixel (x,y: integer; colorIndex: byte);       overload;
procedure AddPixel (x,y: integer; color: byte);
procedure SetPixel (x,y: integer);
procedure XorPixel (x,y: integer; color: byte);

{ *** colors *** }
procedure SetColor (color: tColor);                        overload;
procedure SetColor (colorIndex: byte);                     overload;
procedure SetBkColor (color: tColor);                      overload;
procedure SetBkColor (colorIndex: byte);                   overload;
procedure SetFillColor (color: tColor);                    overload;
procedure SetFillColor (colorIndex: byte);                 overload;
procedure SetIntensity (intensity, color: byte);
procedure SetWriteMode (writeMode: integer);
var       GetAntiAliasing: boolean = false;
procedure SetAntiAliasing (isOn: boolean);

{ *** polygon, fills and figures *** }
procedure Bar (x1,y1, x2,y2: integer);
procedure Rectangle (x1,y1, x2,y2: integer);
procedure DrawPoly (pn: word; var polydata);
procedure FillPoly (pn: word; var polydata);

{ *** line-oriented routines *** }
procedure MoveTo (x, y: integer);
procedure MoveRel(dx,dy:integer);
procedure LineTo (x, y: integer);
procedure LineRel(dx,dy:integer);
procedure Line (x1,y1, x2,y2: integer);

{ *** text routines *** }
procedure TextColor (color: tColor);                       overload;
procedure TextColor (color: byte);                         overload;
procedure TextBackground (color: tColor);                  overload;
procedure TextBackground (color: byte);                    overload;
function TextHeight: word;
function TextWidth (TextString: string): word;
procedure OutText (TextString: string);
procedure OutTextXY (x,y: integer; TextString: string);

{ *** additional routines *** }
(*
function RGB   (r,g,b: byte): tColor;
function RGB15 (r,g,b: byte): word;
function RGB16 (r,g,b: byte): word;
*)

procedure SelectBitmap (name: string; bitmap: tBitmap);
procedure RestoreBitmap;
procedure SelectCanvas (canvas: tCanvas);
procedure RestoreCanvas;
procedure PrintBitmap (bitmap: tBitmap);
procedure Print;
procedure SelectDesktop;

procedure SetBitmapSize (xr,yr: integer);                            overload;
procedure SetBitmapSize (bm: tBitmap; xr,yr: integer);               overload;
procedure CloneBitmapSize (bm: tBitmap);
function GetStartPointer (bm: tBitmap): pointer;
function GetLinePointer (y: integer): pointer;
function GetPixelPointer (x,y: integer): pointer;                    overload;
function GetPixelPointer (bm: tBitmap;  x,y: integer): pointer;      overload;
procedure GetLine (hLine: pointer; x,y,w: integer);
procedure PutLine (hLine: pointer; x,y,w: integer);
procedure CopyLines(src, dest: tBitmap; y1,y2, h: integer);
procedure CopyRect (src, dest: tCanvas; x1,y1, x2,y2, w,h: integer); overload;
procedure CopyRect        (bm: tBitmap; x1,y1, x2,y2, w,h: integer); overload;
procedure CopyLine (src, dest: tCanvas; x1,y1, x2,y2, w: integer);   overload;
procedure CopyLine       (src: tBitmap; x1,y1, x2,y2, w: integer);   overload;
procedure CopyBitmap   (src, dest: tBitmap);
procedure CloneBitmap  (src, dest: tBitmap);
function LoadBitmap2   (var bm: tBitmap; filename: string): boolean;

procedure DrawHLine (x1,x2,y: integer);
procedure DrawVLine (x,y1,y2: integer);
procedure _DrawHLine256 (x1,x2,y: integer; col: byte);
procedure _DrawVLine256 (x,y1,y2: integer; col: byte);
function PixelFormatString (pixelFormat: tPixelFormat): string;

procedure ShadowText (x,y: integer; textString: string);

procedure RaiseGraphError (errStr: string);

var   bgBitmap       : tBitmap = nil;  // background bitmap
      desktop        : tCanvas = nil;  // desktop canvas
      activeBitmap   : tBitmap = nil;  // select drawing bitmap
      activeCanvas   : tCanvas = nil;  // select drawing surface
      activeRect, viewRect: tRect;     // drawing area
      bytesPerPixel  : byte = 1;       // of active bitmap
      transparentText: boolean = true;

const
  sPixelFormat     = 'wrong pixel format';
  sDimension       = 'wrong dimension';
  sLineIndex       = 'wrong line index';
  sBitmapCreation  = 'error at bitmap creation';
  sBitmapOperation = 'error at bitmap operation';

//*********************************************************
implementation

uses  MMSystem, SysUtil2, Math2, Images, Palette;

var  activeViewPort: viewPortType
        = (x1:   0;  y1:   0;
           x2: 640;  y2: 480;  clip: false);

const bytesPerPixels: array[0..ord(pfCustom)] of byte
      = (1,1,1,1,2,2,3,4,3);

var   savedBitmap: tBitmap;          // saved drawing bitmap
      savedCanvas: tCanvas;          // saved drawing canvas
      savedBmName: string;

//*********************************************************
// raise graphics error
//---------------------------------------------------------
procedure RaiseGraphError (errStr: string);
begin
{$ifopt D+}
  writeln (tst,'>>> Graph.RaiseGraphError: ', errStr);
  flush (tst);
{$endif}
  // if crashing -> see call stack for error position
  Raise EInvalidGraphicOperation.Create(errStr);
end;

//*********************************************************
// clear whole desktop
//---------------------------------------------------------
procedure ClearDevice;
begin
{$ifdef test1}
  writeln (tst,'Graph.ClearDevice: ');
{$endif}
  SelectDesktop;
  activeCanvas.Brush.Color := clBlack;
  activeCanvas.FillRect (activeRect);
  SetBkColor(GetBkColor);
  RestoreCanvas;
end;
//*********************************************************
procedure SetViewPort (x1,y1, x2,y2: integer; clip: boolean);
begin
  viewRect.left   := x1;
  viewRect.top    := y1;
  viewRect.right  := x2+1;
  viewRect.bottom := y2+1;
  activeViewPort.clip := clip;
  if clip then activeRect := viewRect;
end;
//---------------------------------------------------------
procedure ClearViewPort;
begin
{$ifdef test2}
  writeln (tst,'Graph.ClearViewPort: ', ColorString (GetBkColor));
{$endif}
  activeCanvas.Brush.Color := GetBkColor;
  if activeViewPort.clip
  then activeCanvas.FillRect (viewRect)
  else ClearBitmap (bgBitmap);
end;
//---------------------------------------------------------
procedure GetViewSettings (var viewPort: viewPortType);
begin
  activeViewPort.x1 := viewRect.left;
  activeViewPort.y1 := viewRect.top;
  activeViewPort.x2 := viewRect.right;
  activeViewPort.y2 := viewRect.bottom;
  viewPort := activeViewPort;
end;
//*********************************************************
procedure GetAspectRatio (var Xasp, Yasp: Word);
begin
  Xasp := screenInfo.xAspect;
  Yasp := screenInfo.yAspect;
end;
//---------------------------------------------------------
procedure SetAspectRatio (var Xasp, Yasp: Word);
begin
  screenInfo.xAspect := Xasp;
  screenInfo.yAspect := Yasp;
end;
//*********************************************************
function GetX: integer;
begin
  GetX := activeCanvas.PenPos.x;
end;
//---------------------------------------------------------
function GetY: integer;
begin
  GetY := activeCanvas.PenPos.y;
end;
//*********************************************************
// draw pixel with active color set by SetColor
//---------------------------------------------------------
procedure SetPixel (x,y: integer);
begin
  if (x < 0) or (x > GetMaxX) then exit;
  if (y < 0) or (y > GetMaxY) then exit;
  activeCanvas.Pixels[x,y] := GetColor;
end;
//*********************************************************
// return pixel color at position x,y from active canvas
//---------------------------------------------------------
function  GetPixel (x,y: integer): tColor;
begin
  if      (x < 0) or (x > GetMaxX) then result := clRed
  else if (y < 0) or (y > GetMaxY) then result := clRed
  else result := activeCanvas.Pixels[x,y];
end;
//*********************************************************
// draw pixel with given true color
// input:     x,y: integer;       pixel position
//            color: tColor;      true pixel color
//---------------------------------------------------------
procedure PutPixel (x,y: integer; color: tColor);
begin
{$ifopt D+}
  if activeBitmap.PixelFormat <= pf8Bit
  then RaiseGraphError (sPixelFormat);
{$endif}
  if (x < 0) or (x > GetMaxX) then exit;
  if (y < 0) or (y > GetMaxY) then exit;
  activeCanvas.Pixels[x,y] := color;
end;
//*********************************************************
// draw pixel with given rgb values
// input:     x,y: integer;       pixel position
//            r,g,b: byte;        rgb values of pixel color
//---------------------------------------------------------
procedure PutPixel (x,y: integer; r,g,b: byte);
begin
{$ifopt D+}
  if activeBitmap.PixelFormat <= pf8Bit
  then RaiseGraphError (sPixelFormat);
{$endif}
  if (x < 0) or (x > GetMaxX) then exit;
  if (y < 0) or (y > GetMaxY) then exit;
  activeCanvas.Pixels[x,y] := RGB(r,g,b);
end;
//*********************************************************
// draw pixel - only for pf8bit format with a palette !!!
// input:     x,y: integer;       pixel position
//            colorIndex: byte;   color index of palette
//---------------------------------------------------------
procedure PutPixel (x,y: integer; colorIndex: byte);
begin
  if (x < 0) or (x > GetMaxX) then exit;
  if (y < 0) or (y > GetMaxY) then exit;
  if activeBitmap.PixelFormat <= pf8Bit
  then pByte(GetPixelPointer(x,y))^ := colorIndex
  else activeCanvas.Pixels[x,y] := Palette.GetPaletteColor(colorIndex);
end;
//*********************************************************
// add value to active pixel - only for pf8bit format with a palette !!!
// input:     x,y: integer;    pixel position
//            color: byte;     add color value
//---------------------------------------------------------
procedure AddPixel (x,y: integer; color: byte);
var      linePtr: pByteArray;
begin
  if (x < 0) or (x > GetMaxX) then exit;
  if (y < 0) or (y > GetMaxY) then exit;
  linePtr := (activeBitmap.ScanLine[y]);
  linePtr^[x] := byte (linePtr^[x] + color);
end;
//---------------------------------------------------------
procedure XorPixel (x,y: integer; color: byte);
var   pixPtr: pByte;
begin
{$ifopt D+}
  if activeBitmap.PixelFormat <> pf8Bit
  then RaiseGraphError (sPixelFormat);
{$endif}
  if (x < 0) or (x > GetMaxX) then exit;
  if (y < 0) or (y > GetMaxY) then exit;
  pixPtr := GetPixelPointer(x,y);
  pixPtr^ := pixPtr^ xor color;
end;
//*********************************************************
// set active drawing color
//---------------------------------------------------------
procedure SetColor (color: tColor);
begin
  GetColor := color;
  activeCanvas.Pen.Color := color;
end;
//*********************************************************
// set active drawing color given by palette index
//---------------------------------------------------------
procedure SetColor (colorIndex: byte);
begin
  GetColorIndex := colorIndex;
  GetColor := Palette.GetPaletteColor(colorIndex);
  activeCanvas.Pen.Color := GetColor;
end;
//*********************************************************
// set active background color
//---------------------------------------------------------
procedure SetBkColor (color: tColor);
begin
{$ifdef test4}
  writeln (tst,'Graph.SetBkColor: ', ColorString(color));
{$endif}
  GetBkColor := color;
  activeCanvas.Brush.Color := color;
end;
//*********************************************************
// set active background color given by palette index
//---------------------------------------------------------
procedure SetBkColor (colorIndex: byte);
begin
{$ifdef test4}
  writeln (tst,'Graph.SetBkColor: ', colorIndex);
{$endif}
  GetBkColorIndex := colorIndex;
  GetBkColor := Palette.GetPaletteColor(colorIndex);
  activeCanvas.Brush.Color := GetBkColor;
end;
//*********************************************************
// set active background color
//---------------------------------------------------------
procedure SetFillColor (color: tColor);
begin
{$ifdef test4}
  writeln (tst,'Graph.SetFillColor: ', ColorString(color));
{$endif}
  GetFillColor := color;
  activeCanvas.Brush.Color := color;  //???
end;
//*********************************************************
// set active background color given by palette index
//---------------------------------------------------------
procedure SetFillColor (colorIndex: byte);
begin
{$ifdef test4}
  writeln (tst,'Graph.SetFillColor: ', colorIndex);
{$endif}
  GetFillColorIndex := colorIndex;
  GetFillColor := Palette.GetPaletteColor(colorIndex);
  activeCanvas.Brush.Color := GetFillColor;
end;
{*********************************************************}
{ Der Helligkeitswert (0..255) fr die aktuelle
{ Fllfarbe (0..15) wird als Farbindex (16..255) gesetzt
{ Achtung: Nur fr 256 Farben !
{--------------------------------------------------------}
procedure SetIntensity (intensity, color: byte);
begin
  GetIntensity := intensity;
  GetFillColor := color and $F;         {0..15}
  intensity := intensity shr 4;         {0..15}
  if (intensity = 0) or (GetFillColor = 0)
  then GetFillColor := 0
  else GetFillColor := GetFillColor * 16 + intensity; {Palettenindex}
{
writeln (tst, 'c=',color,' i=',intensity, '  fc=',GetFillColor);
}
end;
//---------------------------------------------------------
procedure SetWriteMode (writeMode: integer);
begin
  case writeMode of
    CopyPut:    activeCanvas.Pen.Mode := pmCopy;
    XORPut:     activeCanvas.Pen.Mode := pmXor;
    else        activeCanvas.Pen.Mode := pmCopy;
  end;
end;
//*********************************************************
procedure SetAntiAliasing (isOn: boolean);
begin
  GetAntiAliasing := isOn;
end;
//*********************************************************
// draw filled rectangle
//---------------------------------------------------------
procedure Bar (x1,y1, x2,y2: integer);
begin
  activeCanvas.Brush.Style := bsSolid;
  activeCanvas.Brush.Color := GetColor;
  activeCanvas.FillRect (Rect (x1, y1, x2+1, y2+1));
  activeCanvas.Brush.Color := GetBkColor;
end;
//*********************************************************
procedure Rectangle (x1,y1, x2,y2: integer);
begin
  activeCanvas.Brush.Style := bsClear;
  activeCanvas.Rectangle (x1,y1, x2,y2);
end;
{*********************************************************}
{ Zeichnet eine geschlossene Polygonlinie
{ mit der aktuellen Zeichenfarbe GetColor
{---------------------------------------------------------}
procedure DrawPoly (pn: word; var polydata);
begin
// ???  activeCanvas.Polyline(pn, addr(polydata));
end;
{*********************************************************}
{ Zeichnet ein geflltes Polygon mit der Farbe GetFillColor
{ und der Randfarbe GetColor, falls diese nicht 0 ist.
{---------------------------------------------------------}
procedure FillPoly (pn: word; var polydata);
begin
// ???  activeCanvas.Polygon (pn, polydata);
end;
//*********************************************************
procedure MoveTo (x, y: integer);
begin
  activeCanvas.MoveTo (x,y);
end;
//---------------------------------------------------------
procedure MoveRel (dx, dy: integer);
var  p: tPoint;
begin
  p := activeCanvas.PenPos;
  activeCanvas.MoveTo (p.x + dx, p.y + dy);
end;
//*********************************************************
procedure LineTo (x, y: integer);
begin
  activeCanvas.LineTo (x, y);
end;
//---------------------------------------------------------
procedure LineRel (dx, dy: integer);
var  p: tPoint;
begin
  p := activeCanvas.PenPos;
  activeCanvas.LineTo (p.x + dx, p.y + dy);
end;
//*********************************************************
procedure Line (x1,y1, x2,y2: integer);
begin
  activeCanvas.MoveTo (x1,y1);
  activeCanvas.LineTo (x2,y2);
end;
//*********************************************************
// set active text color
//---------------------------------------------------------
procedure TextColor (color: tColor);
begin
  GetTextColor := color;
  activeCanvas.Font.Color := color;
//  writeln (tst, 'fcol=',GetTextColor);
end;
//---------------------------------------------------------
procedure TextColor (color: byte);
begin
  GetTextColor := Palette.GetPaletteColor(color);
  activeCanvas.Font.Color := GetTextColor;
//  writeln (tst, 'fcol=',GetTextColor);
end;
//*********************************************************
// set active text background color
//---------------------------------------------------------
procedure TextBackground (color: tColor);
begin
  GetTextBackground := color;
  activeCanvas.Brush.Color := GetTextBackground;
end;
//---------------------------------------------------------
procedure TextBackground (color: byte);
begin
  GetTextBackground := Palette.GetPaletteColor(color);
  activeCanvas.Brush.Color := GetTextBackground;
end;
//*********************************************************
function TextHeight: word;
begin
  // ToDo: font.size is wrong ???
//  result := activeCanvas.Font.Size;
  result := 16;
end;
//---------------------------------------------------------
function TextWidth (textString: string): word;
begin
  result := activeCanvas.Font.Size * length (textstring);
end;
//*********************************************************
procedure ResetFont;
begin
  with activeCanvas.Font
  do begin
    name := 'MS Sans Serif';
    size := 8;
    style := [fsBold];
  end;
end;
//*********************************************************
// write text with system font
//---------------------------------------------------------
procedure OutText (textString: string);
begin
  ResetFont;
  if transparentText
  then SetBkMode (activeCanvas.Handle, Transparent);
  with activeCanvas.PenPos
  do activeCanvas.TextOut (x,y, textString);
end;
//---------------------------------------------------------
procedure OutTextXY (x,y: integer; textString: string);
begin
  ResetFont;
  if transparentText
  then SetBkMode (activeCanvas.Handle, Transparent);
  activeCanvas.TextOut (x,y, textString);
end;

//=========================================================
// additional routines
//=========================================================

(*
{*********************************************************}
{ convert r g b values to longint color                   }
{ input:     r,g,b:    rgb values of pixel color (0..255) }
{ return:    longint:  mode specific color value          }
{ example: 15bit (555) 00000000 00000000 0rrrrrgg gggbbbbb}
{          16bit (565) 00000000 00000000 rrrrrggg gggbbbbb}
{          24bit (888) 00000000 rrrrrrrr gggggggg bbbbbbbb}
{---------------------------------------------------------}
function RGB2Color (r,g,b: byte): tColor;
begin
  case screen.pixelFormat of
    pf15bit: RGB2Color := RGB15(r,g,b);
    pf16bit: RGB2Color := RGB16(r,g,b);
    pf24bit,
    pf32bit: RGB2Color := WinColor (r,g,b);
    else     RGB2Color := clRed;         // error color
  end;
end;
*)

//---------------------------------------------------------
procedure ShadowText (x,y: integer; textString: string);
begin
{$ifdef test3}
  writeln (tst,'Graph.ShadowText: ', textString);
{$endif}
  with activeCanvas
  do begin
    if transparentText
    then SetBkMode (Handle, Transparent);
    Font.Style := [fsBold];
    Font.Color := clBrown;
    TextOut (x+1,y+1, textString);
    Font.Color := GetTextColor;
    TextOut (x,y, textString);
  end;
end;
//*********************************************************
// select drawing bitmap
//---------------------------------------------------------
procedure SelectBitmap (name: string; bitmap: tBitmap);
begin
{$ifdef test1}
  with bitmap do
  writeln (tst,'Graph.SelectBitmap: ', HexString(bitmap), name:10,'  '
              , Width:5, Height:5, ord(PixelFormat):3);
{$endif}
  savedBmName := name;
  savedBitmap := activeBitmap;
  activeBitmap := bitmap;
  activeCanvas := bitmap.Canvas;
  bytesPerPixel := bytesPerPixels[ord(bitmap.PixelFormat)];
end;
//*********************************************************
// restore last drawing bitmap
//---------------------------------------------------------
procedure RestoreBitmap;
begin
{$ifdef test1}
  writeln (tst,'Graph.RestoreBitmap: ', savedBmName:10);
{$endif}
  SelectBitmap (savedBmName, savedBitmap);
end;
//*********************************************************
// select active drawing canvas
//---------------------------------------------------------
procedure SelectCanvas (canvas: tCanvas);
begin
  savedCanvas := activeCanvas;
  activeCanvas := canvas;
end;
//*********************************************************
procedure RestoreCanvas;
begin
  activeCanvas := savedCanvas;
end;
//*********************************************************
// select desktop as drawing device
//---------------------------------------------------------
procedure SelectDesktop;
begin
  activeCanvas := desktop;
  activeRect := desktop.ClipRect;
end;
//*********************************************************
// set bitmap size of active bitmap
//---------------------------------------------------------
procedure SetBitmapSize (xr,yr: integer);
begin
{$ifdef test2}
  writeln (tst,'Graph.SetBitmapSize: '
              , HexString(activeBitmap), xr:5,yr:5);
{$endif}
  activeBitmap.Width  := xr;
  activeBitmap.Height := yr;
  pageX := xr;
  pageY := yr;
  GetMaxX := pageX - 1;
  GetMaxY := pageY - 1;
  SetViewPort (0,0,GetMaxX, GetMaxY, false);
end;
//*********************************************************
// set bitmap size of given bitmap
//---------------------------------------------------------
procedure SetBitmapSize (bm: tBitmap; xr,yr: integer);
begin
{$ifdef test2}
  writeln (tst,'Graph.SetBitmapSize: '
              , HexString(bm), xr:5,yr:5);
{$endif}
  bm.Width  := xr;
  bm.Height := yr;
end;
//*********************************************************
// set bitmap size to actual window client rectangle
//---------------------------------------------------------
procedure CloneBitmapSize (bm: tBitmap);
begin
{$ifdef test2}
  writeln (tst,'Graph.CloneBitmapSize: '
              , HexString(bm), pageX:5,pageY:5);
{$endif}
  bm.Width  := activeBitmap.Width;
  bm.Height := activeBitmap.Height;
end;
//*********************************************************
function GetStartPointer (bm: tBitmap): pointer;
begin
  GetStartPointer := bm.ScanLine[bm.Height-1];
end;
//*********************************************************
function GetLinePointer (y: integer): pointer;
begin
{$ifopt D+}
  if (y < 0) or (y > GetMaxY)
  then RaiseGraphError (sScanLine);
{$endif}
   GetLinePointer := activeBitmap.ScanLine[y];
end;
//*********************************************************
// return pixel address of active bitmap
//---------------------------------------------------------
function GetPixelPointer (x,y: integer): pointer;
var      linePtr: pByteArray;
begin
{$ifopt D+}
  if (x < 0) or (x > activeBitmap.Width)  then RaiseGraphError (sLineIndex);
  if (y < 0) or (y > activeBitmap.Height) then RaiseGraphError (sScanLine);
{$endif}
  linePtr := activeBitmap.ScanLine[y];
  result := @linePtr^[x*bytesPerPixel];
end;
//*********************************************************
// return pixel address of given bitmap
//---------------------------------------------------------
function GetPixelPointer (bm: tBitmap;  x,y: integer): pointer;
var      linePtr: pByteArray;
begin
{$ifopt D+}
  if (x < 0) or (x > bm.Width)  then RaiseGraphError (sLineIndex);
  if (y < 0) or (y > bm.Height) then RaiseGraphError (sScanLine);
{$endif}
  linePtr := bm.ScanLine[y];
  case bm.PixelFormat of
    pf24bit, pf32bit: result := @linePtr^[x*3];
    pf15bit, pf16bit: result := @linePtr^[x*2];
    else              result := @linePtr^[x];
  end;
end;
//*********************************************************
// copy w pixels from position x,y to horizontal line buffer
//---------------------------------------------------------
procedure GetLine (hLine: pointer; x,y,w: integer);
begin
{$ifopt D+}
  if (longint(x)+w) > activeBitmap.Width  then RaiseGraphError (sLineIndex);
{$endif}
  Move (GetPixelPointer(x,y)^, hline^, w*bytesPerPixel);
end;
//*********************************************************
// copy w pixels of given horizontal line to position x,y
//---------------------------------------------------------
procedure PutLine (hLine: pointer; x,y,w: integer);
begin
{$ifopt D+}
  if (longint(x)+w) > pageX then RaiseGraphError (sLineIndex);
{$endif}
  Move (hline^, GetPixelPointer(x,y)^, w*bytesPerPixel);
end;
//*********************************************************
// copy w pixels of horizontal line from src position x1,y1
// to dest canvas position x2,y2
//---------------------------------------------------------
procedure CopyLine (src, dest: tCanvas; x1,y1, x2,y2, w: integer);
begin
{$ifopt D+}
  if (longint(x2)+w) > pageX    then RaiseGraphError (sDimension);
  if (y1 < 0) or (y1 > GetMaxY) then RaiseGraphError (sScanLine);
  if (y2 < 0) or (y2 > GetMaxY) then RaiseGraphError (sScanLine);
{$endif}
  BitBlt (dest.handle, x2,y2, w, 1
          ,src.handle, x1,y1, SrcCopy);
end;
//*********************************************************
// copy w pixels of horizontal line from given bitmap
// at position x1,y1 to active bitmap at position x2,y2
//---------------------------------------------------------
procedure CopyLine (src: tBitmap; x1,y1, x2,y2, w: integer);
begin
  CopyLine (src.Canvas, activeCanvas, x1,y1, x2,y2, w);
end;
//*********************************************************
// copy rectangle with size w*h from src at position x1,y1
// to dest Canvas at position x2,y2
//---------------------------------------------------------
procedure CopyRect (src, dest: tCanvas; x1,y1, x2,y2, w,h: integer);
begin
  if y1 > GetMaxY then exit;
  if y2 > GetMaxY then exit;
  if y1 < 0
  then begin
    h := h - y1;
    y2 := y2 - y1;
    y1 := 0;
  end;
  if y2 < 0
  then begin
    h := h - y2;
    y1 := y1 - y2;
    y2 := 0;
  end;
  BitBlt (dest.handle, x2,y2, w, h
          ,src.handle, x1,y1, SrcCopy);
end;
//*********************************************************
// copy rectangle with size w*h from given bitmap
// at position x1,y1 to active bitmap at position x2,y2
//---------------------------------------------------------
procedure CopyRect (bm: tBitmap; x1,y1, x2,y2, w,h: integer);
begin
  CopyRect (bm.Canvas, activeCanvas, x1,y1, x2,y2, w,h);
end;
//*********************************************************
// copy rectangle with size w*h from given bitmap
// at position x1,y1 to active bitmap at position x2,y2
//---------------------------------------------------------
procedure CopyLines (src, dest: tBitmap; y1,y2, h: integer);
begin
  CopyRect (src.Canvas, dest.Canvas, 0,y1, 0,y2, src.Width, h);
end;
//*********************************************************
// direct bitmap copy.
// note: only for same pixel format !!!
//       and destination must be >= source bitmap
//---------------------------------------------------------
procedure CloneBitmap (src, dest: tBitmap);
begin
  try
    dest.Height := 0;
    dest.Width  := 0;
    dest.PixelFormat := src.PixelFormat;
    dest.Height := src.Height;
    dest.Width  := src.Width;
    dest.Canvas.Draw (0,0, src);
  except
{$ifopt D+}
    writeln (tst,'>>> ERROR Graph.CloneBitmap: ',sBitmapOperation,#13#10
    ,'  src:  ',HexString(src)
    ,' w=',src.width:5,  ' h=',src.height:5, ord(src.PixelFormat):5, #13#10
    ,'  dest: ',HexString(dest)
    ,' w=',dest.width:5, ' h=',dest.height:5,ord(dest.PixelFormat):5);
    flush (tst);
{$endif}
  end;
end;
//*********************************************************
// direct bitmap copy.
// note: only for same pixel format !!!
//       destination must be >= source bitmap.
//       if pixel format is not the same use DrawBitmap.
//---------------------------------------------------------
procedure CopyBitmap (src, dest: tBitmap);
var   yi, bpp, w, h: word;
begin
  try
{$ifopt D+}
    if src.PixelFormat <> dest.PixelFormat
    then begin
      RaiseGraphError (sPixelFormat);
      dest.Canvas.Draw (0,0, src);
      exit;
    end;
{$endif}
    w := Min (src.Width, dest.Width);
    h := Min (src.Height, dest.Height);
{
    // color palette error ???
    BitBlt(dest.Canvas.handle, 0,0, src.Width, src.Height
         ,src.Canvas.handle, 0,0, SrcCopy);
}
    case src.pixelFormat of
      pf15bit, pf16bit: bpp := 2;
      pf24bit, pf32bit: bpp := 3;
      else              bpp := 1;
    end;

    bpp :=  bpp * w;
    for yi := 0 to h-1
    do Move ( src.Scanline[yi]^
            ,dest.Scanline[yi]^, bpp);
  except
{$ifopt D+}
    writeln (tst,'>>> ERROR Graph.CopyBitmap: ',sBitmapOperation,#13#10
    ,'  src:  ',HexString(src), ' w=',src.width:5,  ' h=',src.height:5, ord(src.PixelFormat):5, #13#10
    ,'  dest: ',HexString(dest),' w=',dest.width:5, ' h=',dest.height:5,ord(dest.PixelFormat):5);
    flush (tst);
{$endif}
  end;
end;
//*********************************************************
//  load bitmap and in case of an error create a bitmap
//  and write loading error to it
//---------------------------------------------------------
function LoadBitmap2 (var bm: tBitmap; filename: string): boolean;
begin
  result := LoadBitmap (bm, filename);
  if not result
  then begin          //draw loading error to new bitmap
    bm.Width  := 400;
    bm.Height := 240;
    SelectBitmap (filename, bm);
    SetBkColor (tColor(clBlack));
    ClearBitmap (bm);
    TextColor (clYellow);
    OutTextXY (55,55, 'error loading picture ' + filename);
    RestoreBitmap;
  end;
end;

//*********************************************************
// return pixel format of given bitmap
//---------------------------------------------------------
var pfText: array [tPixelFormat] of string[2]
    = ('sc', '1', '4', '8', '15', '16', '24', '32', 'cu');

function PixelFormatString (pixelFormat: tPixelFormat): string;
begin
  result := pfText[pixelFormat];
end;

//*********************************************************
// set pixel format of active bitmap
//---------------------------------------------------------
procedure SetGraphMode (graphMode: tPixelFormat);
begin
{$ifdef test1}
  writeln (tst,'Graph.SetGraphMode: ',ord(graphMode)
              ,' pf=', PixelFormatString (graphMode));
{$endif}

  if graphMode = pfDevice
  then graphMode := screenInfo.pixelFormat;
  GetGraphMode := graphMode;
//  bytesPerPixel := bytesPerPixels[ord(graphMode)];

  SetBkColor (clBlack);
  if activeBitmap.PixelFormat <> graphMode
  then begin
    ClearViewPort;
    activeBitmap.PixelFormat := graphMode;
  end;
  SelectBitmap ('bg', bgBitmap);
//  screen.xAspect := round (10000.0 * screen.Y * 4 / (screen.X * 3));
//  screen.yAspect := 10000;
end;
//*********************************************************
// draw horizontal line with clipping
//---------------------------------------------------------
procedure DrawHLine (x1,x2,y: integer);
begin
  if (y < 0) or (y > GetMaxY) then exit;
  Sort (x1,x2);       // if x1 > x2 then Swap (x1,x2);
{
  if x1 > GetMaxX then exit
  else if x1 < 0 then x1 := 0;
  if x2 < 0 then exit
  else if x2 > GetMaxX then x2 := GetMaxX;
}
  MoveTo (x1,y);
  LineTo (x2+1,y);
end;
//*********************************************************
// draw vertical line with clipping
//---------------------------------------------------------
procedure DrawVLine (x,y1,y2: integer);
begin
  if (x < 0) or (x > GetMaxX) then exit;
  Sort (y1,y2);       // if y1 > y2 then Swap (y1,y2);
{
  if y1 > GetMaxY then exit
  else if y1 < 0 then y1 := 0;
  if y2 < 0 then exit
  else if y2 > GetMaxY then y2 := GetMaxY;
}
  MoveTo (x,y1);
  LineTo (x,y2);
end;

//*********************************************************
// fast drawing of horizontal line only for pf8bit mode !!!
// attention:  no limit check !!!
//---------------------------------------------------------
procedure _DrawHLine256 (x1,x2,y: integer; col: byte);
begin
  FillChar (pByteArray(activeBitmap.ScanLine[y])^[x1], x2-x1+1, col);
end;
{********************************************************}
// fast drawing of vertical line only for pf8bit mode !!!
// attention:  no limit check !!!
//---------------------------------------------------------
procedure _DrawVLine256 (x,y1,y2: integer; col: byte);
var   yi, delta: longint;
      pixPtr: ^Byte;
begin
//  writeln (tst, '_DrawVLine256: y2=',y2);
  pixPtr := activeBitmap.ScanLine[y2];
  delta := longint(activeBitmap.ScanLine[y2-1]) - longint(pixPtr);
  inc(longint(pixPtr), x);
  for yi := y1 to y2
  do begin
    pixPtr^ := col;
    inc (longint(pixPtr),delta);
  end;
end;

//*********************************************************
(*
procedure PrintBitmap2 (bitmap: tBitmap);
var
    bitmapHeader : pBitmapInfo;
    bitmapImage  : pointer;
    headerSize   : DWORD;
    imageSize    : DWORD;
begin
  // get device independent bitmap sizes
  GetDIBSizes (Bitmap.Handle, HeaderSize, ImageSize);
  GetMem (bitmapHeader, headerSize);
  GetMem (bitmapImage,  imageSize);
  try
    // convert to device independent bitmap
    GetDIB (bitmap.Handle, bitmap.Palette, bitmapHeader^, bitmapImage^);
    // copies bitmap into a device-independent bitmap (DIB)
    printer.BeginDoc;
    with Printer
    do StretchDIBits(Printer.Canvas.Handle,
                  0,0,                 // destination origin
                  printer.pageWidth,   // destination width
                  printer.pageHeight,  // destination height
                  0, 0,                            // source origin
                  bitmap.Width, bitmap.Height,     // source size
                  BitmapImage,
                  TBitmapInfo(BitmapHeader^),
                  DIB_RGB_COLORS,
                  SRCCOPY)
  finally
    printer.EndDoc;
    FreeMem (bitmapHeader);
    FreeMem (bitmapImage)
  end;
end;
*)
//*********************************************************
// print given bitmap
//---------------------------------------------------------
procedure PrintBitmap (bitmap: tBitmap);
var   dr: tRect;
begin
  dr := Rect (0,0, printer.pageHeight, printer.pageWidth);
  if dr.bottom > dr.Right          // portrait ?
  then dr.bottom := dr.Bottom div 2;
  try
    printer.BeginDoc;
    printer.Canvas.StretchDraw (dr, bitmap);
  finally
    printer.EndDoc;
  end;
end;
//*********************************************************
// print actual window content
//---------------------------------------------------------
procedure Print;
begin
  PrintBitmap (activeBitmap);
end;
//*********************************************************
// initilize background bitmap
//---------------------------------------------------------
procedure InitBackgroundBitmap;
begin
  bgBitmap := TBitmap.Create;
  if bgBitmap = nil
  then RaiseGraphError (sBitmapCreation);
  bgBitmap.HandleType  := bmDIB;
  bgBitmap.PixelFormat := screenInfo.pixelFormat;
  SelectBitmap ('bg', bgBitmap);
  SetBitmapSize (pageX, pageY);
end;
//*********************************************************
var   pixPtr: ^word;

initialization

  {$ifdef InitTest}
    writeln (tst, '   Graph.initialization: ');
  {$endif}

  // get desktop
  desktop := TCanvas.Create;
  desktop.Handle := GetWindowDC (GetDesktopWindow);

  // create background image and set parameter
  InitBackgroundBitmap;
  SetGraphMode (bgBitmap.PixelFormat);

  // if Win-95 reports pf16bit pixel format,
  // check if pf15bit is used instead
  // and if it is -> correct device pixel format
  if screenInfo.pixelFormat = pf16bit
  then begin
    activeCanvas.Pixels[0,0] := $00808080;  // PutPixel
    pixPtr := GetLinePointer(0);
    if pixPtr^ and $8000 = 0           // bit 0 not set ?
    then begin
      screenInfo.pixelFormat := pf15bit;
      bgBitmap.PixelFormat := screenInfo.pixelFormat;
    end;
  end;
//---------------------------------------------------------
finalization
  {$ifdef ExitTest}
    writeln (tst, '   Graph.finalization: ');
  {$endif}

  bgBitmap.Free;
  desktop.Free;
end.
