//*******************************************************************
//
//  UNIT:     M a g n i f y                      08.09.99, 04.10.2002
//
//  interactive magnify animation, coded by Gerd Platl
//
//  animated sizeable lens which magnifies background picture
//
//*******************************************************************
{$I comp_opt.inc}

unit Magnify;

{$ifopt D+}
  {$define xtest1    main functions}
  {$define xtest2    sub functions}
{$endif}

interface

//*********************************************************
implementation

uses  SysUtils, Graphics,
      SysUtil2, Graph, Images, Values, GrForm;

procedure AnimResize; forward;

//=========================================================
// Background picture stretching
//=========================================================

var   origBm: tBitmap = nil;   // original background picture
      backBm: tBitmap = nil;   // transformed background picture

//--- file handling ---

var   picFileNo    :word = 0;          // actual picture number
      picName      :string[40] = '';   // actual file name

//*********************************************************
// load next picture from file
//---------------------------------------------------------
procedure GetNextPicture;
begin
  picFileNo := LoadDemoPicture (origBm);
  if picFileNo = 0
  then MainForm.FlashInfo ('no ripple files found');
  AnimResize;
  MarkTime;
end;

//=========================================================
// Magnify
//=========================================================

var   moveSpeed:   integer =  15;      // moving speed
      moveRadius:  integer =  80;      // moving radius

      lensRadius:  integer =  70;      // lens radius
      sphereFac:   integer = 120;      // percent of lens radius
      focDist:     integer =  50;      // distance from plane to focus point

      lx:          longint = 0;        // lens left top position
      ly:          longint = 0;

const maxMMI  = 255;                   // max. magnify map index (2^n)
      magSize = maxMMI+1;              // magnify map size
      maxLens = maxMMI div 2;          // max. lens radius

var   magXMap: tBitmap = nil;          // magnify X deviation map
      magYMap: tBitmap = nil;          // magnify Y deviation map
      sphRadius: integer =  100;       // lens sphere radius

{********************************************************}
procedure CreateValues;
begin
  with valuePtr^
  do if ValueCount = 0
  then begin
    AppendValue ('moving speed',    1,  30);
    AppendValue ('moving radius',   0, 160);
    AppendValue ('focus distance',-50, 150);
    AppendValue ('lens radius',    12, maxLens);
    AppendValue ('sphere radius', 100, 140);
  end;
end;
//*********************************************************
// create magnify maps and background bitmap
//---------------------------------------------------------
procedure InitMagnify;

begin
  // create magnify X map
  magXMap := tBitmap.Create;
  magXMap.PixelFormat := pf16Bit;
  magXMap.Width  := magSize;
  magXMap.Height := magSize;

  // create magnify Y map
  magYMap := tBitmap.Create;
  magYMap.PixelFormat := pf16Bit;
  magYMap.Width  := magSize;
  magYMap.Height := magSize;

  // create background bitmap
  backBm := tBitmap.Create;
  backBm.PixelFormat := pf16Bit;
end;
//*********************************************************
// start new animation
//---------------------------------------------------------
procedure AnimStart;
begin
  {$ifdef test1}
    writeln (tst, #13#10'Magnify.AnimStart: ', modeNo);
  {$endif}

  CreateValues;
  SetGraphMode (pf16bit);
  GetNextPicture;
end;
//*********************************************************
// resize animation window
//---------------------------------------------------------
procedure AnimResize;
begin
  {$ifdef test1}
    writeln (tst, 'Magnify.AnimResize: ',pageX,'*',pageY);
  {$endif}

  if origBm <> nil
  then MainForm.SetBackground (origBm);

  // copy transformed background picture
  backBm.Width  := bgBitmap.Width;
  backBm.Height := bgBitmap.Height;
  CopyBitmap (bgBitmap, backBm);
end;
//*********************************************************
// calculate lens magnification values
//---------------------------------------------------------
procedure CalculateMagnify;
var   xi,yi,z: Integer;
      ux,uy  : Integer;      // distance to center point
      xq,yq  : longint;      // x/y square values
      rq,sq  : longint;      // radius square values
      xp,yp  : ^smallInt;    // pointer to correction value
begin
  // check parameter
  if lensRadius > maxLens
  then begin
    lensRadius := maxLens;
    valuePtr^.SetValue (4, lensRadius);
  end;
  sphRadius := lensRadius * sphereFac div 100;

  // calculate refrection map
  rq := lensRadius * lensRadius;
  sq := sphRadius * sphRadius;
  for yi := 0 to maxMMI
  do begin
    xp := magXMap.scanline[yi];
    yp := magYMap.scanline[yi];
    uy := yi - magSize div 2;
    yq := uy * uy;
    for xi := 0 to maxMMI
    do begin
      ux := xi - magSize div 2;
      xq := ux * ux;
      if xq + yq >= rq - 10
      then begin
        xp^ := 0;
        yp^ := 0;
      end
      else begin
        z   := round (sqrt (sq - xq - yq));
        xp^ := round ((focDist-z) * ux/z);
        yp^ := round ((focDist-z) * uy/z);
      end;
      inc (xp);
      inc (yp);
    end;
  end;
end;
//*********************************************************
// draw one animation picture
//---------------------------------------------------------
procedure AnimDraw;

var   mx,my,                 // background picture dimension
      dx,dy,                 // destination position
      xi,yi,                 // pixel position
      xn,yn: longint;        // diversion values
      sp: pWordArray;        // source line pointer
      xp,yp: ^smallInt;      // magnify values
      dp: ^word;             // destination pointer

begin
  {$ifdef test2}
    writeln (tst, 'Magnify.AnimDraw: ');
  {$endif}

  // restore lens rectangle
  SelectBitmap ('bg', bgBitmap);
  CopyRect (backBm, lx,ly, lx,ly, magSize,magSize);

  // calculate light position
  lx := pageX div 2 + round(cos(picNo*moveSpeed / 256) * moveRadius);
  ly := pageY div 2 + round(sin(picNo*moveSpeed / 200) * moveRadius);
  lx := lx - magSize div 2;
  ly := ly - magSize div 2;

  mx := backBm.Width;
  my := backBm.Height;

  dy := ly;
  for yi := 0 to maxMMI
  do begin
    if dy >= my then break;            // destination line outside ?
    if dy >= 0                         // destination line inside ?
    then begin
      dx := lx;
      dp := bgBitmap.scanline[dy];     // destination pixel pointer
      inc(dp,dx);
      xp := magXMap.scanline[yi];      // deviation value pointer
      yp := magYMap.scanline[yi];
      for xi := 0 to maxMMI
      do begin
        if (dx > 0) and (dx < mx)      // drawing position inside ?
        then if (xp^ <> 0) or (yp^ <> 0)
        then begin
          yn := dy + yp^;              // get refracted y position
          if yn < 0 then yn := 0
          else if yn >= my then yn := my -1;

          xn := dx + xp^;              // get refracted x position
          if xn < 0 then xn := 0
          else if xn >= mx then xn := mx - 1;

          sp := backBm.scanline[yn];   // get source line pointer
          dp^ := sp^[xn];              // draw pixel
        end;
        inc (xp);
        inc (yp);
        inc (dp);
        inc (dx);
      end;
    end;
    inc (dy);
  end;
end;
//*********************************************************
// handle animation input
//---------------------------------------------------------
procedure AnimInput;

var recalc: boolean;

begin
  {$ifdef test2}
    writeln (tst, 'Magnify.AnimInput: ', ord(cb));
  {$endif}

  with valuePtr^ do begin
    recalc := (focDist    <> IntValue (3))
           or (lensRadius <> IntValue (4))
           or (sphereFac  <> IntValue (5));
  end;

  case cb of
    CR : begin valuePtr^.RandomValues;   recalc := true;  end;
    'r': begin valuePtr^.Reset;          recalc := true;  end;
    '#': GetNextPicture;
    '': begin PrevDemoPicture;          GetNextPicture;  end;
  end;

  with valuePtr^ do begin
    moveSpeed  := IntValue (1);
    moveRadius := IntValue (2);
    focDist    := IntValue (3);
    lensRadius := IntValue (4);
    sphereFac  := IntValue (5);
  end;
  if recalc then CalculateMagnify;
end;

//*********************************************************
// display animation specific help
//---------------------------------------------------------
procedure AnimHelp;
begin
  HelpText (1, '<ret> random values');
  HelpText (2, '  #   picture ' + IntToStr(picFileNo));
end;
//*********************************************************
initialization

  {$ifdef InitTest}
    writeln (tst, 'Magnify.initialization: ');
    flush (tst);
  {$endif}

  InitMagnify;
  CalculateMagnify;

  RegisterAnimation (' Sizeable Magnification ',
    AnimStart, AnimResize, AnimDraw, AnimInput, AnimHelp);

//----------------------------------------------------------
finalization

  {$ifdef ExitTest}
    writeln (tst, 'Magnify.finalization: ');
    flush (tst);
  {$endif}

  backBm.Free;
end.

