//*******************************************************************
//
//  UNIT:     P l a s m a                        02.06.99, 05.04.2002
//
//  true color plasma animation, coded by Gerd Platl
//
//*******************************************************************
{$I 'comp_opt.inc'}

unit Plasma;

{$ifopt D+}
  {$define test1    main functions}
  {$define xtest2    sub functions}
{$endif}

interface

//*********************************************************
implementation

uses  SysUtils, SysUtil2, Graphics,
      Graph, Palette, Images, Values, GrForm;

const modes = 3;                  // # of effects
var   modeNo: word = 0;           // actual demo mode

var   sinTab, cosTab: array[0..255] of byte;

var   d1: byte =  3;              // values
      d2: byte =  3;
      d3: byte =  3;
      d4: byte =  3;
      d5: byte =  3;
      d6: byte =  3;

var   i1: longint = 11;
      i2: longint = 22;
      i3: longint = 33;
      i4: longint = 44;
      i5: longint = 55;
      i6: longint = 66;

var   heightBm:    tBitmap = nil;
      palIntBm:    tBitmap = nil;
      showMap:     boolean = false;
      cPal: tPalette;

//*********************************************************
procedure CreateValues;
begin
  with valuePtr^
  do if ValueCount = 0
  then begin
    AppendValue ('d1  ' , -10, 11, d1);
    AppendValue ('d2  ' , -10, 11, d2);
    AppendValue ('d3  ' , -10, 11, d3);
    AppendValue ('d4  ' , -10, 11, d4);
    AppendValue ('d5  ' , -10, 11, d5);

    AppendValue ('d6  ' , -10, 11, d6);
//    AppendValue ( 0,  255, 'xr');
  end;
end;
//*********************************************************
procedure CalcSinusTab;
var   xi: word;
begin
  for xi := 0 to 255
  do begin
    sinTab[xi] := round (sin(pi*xi/128)*127)+127;
    cosTab[xi] := round (cos(pi*xi/128)*127)+127;
{    writeln (tst, xi:10, sinTab[xi]:10, cosTab[xi]:10);}
  end;
end;

//*********************************************************
// start new animation
//---------------------------------------------------------
procedure InitAnim;
begin
  SetSineWavePalette (cPal, 65000, 5+Random(250)
                                 , 5+Random(250)
                                 , 5+Random(250));
  MakePalIntensityMap (palIntBm, cPal);
end;

//*********************************************************
// start new animation
//---------------------------------------------------------
procedure AnimStart;
begin
  {$ifdef test1}
    writeln (tst, #13#10'Plasma.AnimStart: ', modeNo);
  {$endif}

  CreateValues;
  InitAnim;
  SetGraphMode (pf32Bit);
  modeNo := 1 + modeNo mod modes;
end;
//*********************************************************
// resize animation window
//---------------------------------------------------------
procedure AnimResize;
begin
  {$ifdef test1}
    writeln (tst, 'Plasma.AnimResize: ',pageX,'*',pageY);
  {$endif}

  heightBm := tBitmap.Create;
  if heightBm = nil
  then RaiseGraphError (sBitmapCreation);
  heightBm.PixelFormat := pf8Bit;
  heightBm.Width  := pageX + 2;    // 2 pixel overscan
  heightBm.Height := pageY + 2;
  SetBitmapPalette (heightBm, cPal);
end;
//*********************************************************
procedure RandomColors;
begin
  InitAnim;
  SetBitmapPalette (heightBm, cPal);
end;

//*********************************************************
procedure AnimatePlasma1;

var   xi,yi, xr2, xg2, xb2, yr2, yg2, yb2: longint;
      rgbPtr: ^byte;
begin
  for yi := 0 to GetMaxY
  do begin
    yr2 := i1 + sinTab[(yi+i2) and $FF];
    yg2 := sinTab[(yi+i4) and $FF];
    yb2 := sinTab[(yi+i6) and $FF];
    rgbPtr := GetPixelPointer (0,yi);
    for xi := 0 to GetMaxX
    do begin
      xr2 := cosTab[(sinTab[byte(xi+yr2)]) and $FF];
      xg2 := sinTab[(cosTab[byte((xi+yg2+xr2)div 4)]) and $FF];
      xb2 := cosTab[(cosTab[byte((xi+yb2+xg2)div 8)]) and $FF];
      rgbPtr^ := sinTab[byte(xi+xr2)];  inc (rgbPtr);
      rgbPtr^ := sinTab[byte(xi+xg2)];  inc (rgbPtr);
      rgbPtr^ := sinTab[byte(xi+xb2 + yb2)];  inc (rgbPtr);
      inc (rgbPtr);
    end;
  end;
end;
//*********************************************************
procedure AnimatePlasma2;

var   xi, yi: integer;
      xr2, xg2, xb2: byte;
      yr2, yg2, yb2: byte;
      rgbPtr: ^byte;

begin
  for yi := 0 to GetMaxY
  do begin
    xr2 := sinTab[(yi+i1)and $FF];   yr2 := cosTab[(xr2+i2) and $ff];
    xg2 := sinTab[(yi+i3)and $FF];   yg2 := cosTab[(xg2+i4) and $FF];
    xb2 := sinTab[(yi+i5)and $FF];   yb2 := cosTab[(xb2+i6) and $FF];
    rgbPtr := GetLinePointer (yi);
    for xi := 0 to GetMaxX
    do begin
      rgbPtr^ := cosTab[(xr2 + yr2 + xi) and $FF];  inc (rgbPtr);
      rgbPtr^ := cosTab[(xg2 + yg2 + xi) and $FF];  inc (rgbPtr);
      rgbPtr^ := cosTab[(xb2 + yb2 + xi) and $FF];  inc (rgbPtr);
      inc (rgbPtr);
    end;
  end;
end;
//*********************************************************
procedure AnimatePlasma3;

var   xi,yi, x1,x2,x3, z1,z2,z3: longint;
      linePtr: ^byte;
begin
  // create height field with 256 heights
  for yi := 0 to GetMaxY
  do begin
    x1 := i1 + sinTab[(yi+i2) and $FF];
    x2 := i5 + sinTab[(yi+i4) and $FF];
    linePtr := heightBm.ScanLine[yi];
    for xi := 0 to GetMaxX
    do begin
      x3 := sinTab[(xi+i3) and $FF];
      z1 := sinTab[byte(xi+x1)];
      z2 := sinTab[byte(xi+x3)];
      z3 := sinTab[byte(x3+x2)];
      linePtr^ := ((z1 + z2) shr 1 + z3) shr 1;
      inc (linePtr);
    end;
  end;
  bgBitmap.Canvas.Draw (0,0, heightBm);
end;

//*********************************************************
procedure AnimatePlasma4;

var   xi,yi, x1,x2,x3, z1,z2,z3: longint;
      linePtr: ^byte;
begin
  // create height field with 256 heights
  for yi := 0 to GetMaxY
  do begin
    x1 := i1 + sinTab[(yi+i2) and $FF];
    x2 := i5 + sinTab[(yi+i4) and $FF];
    linePtr := heightBm.ScanLine[yi];
    for xi := 0 to GetMaxX
    do begin
      x3 := sinTab[(xi+i3) and $FF];
      z1 := sinTab[byte(xi+x1)];
      z2 := sinTab[byte(xi+x3)];
      z3 := sinTab[byte(x3+x2)];
      linePtr^ := ((z1 + z2) shr 1 + z3) shr 1;
      inc (linePtr);
    end;
  end;

  bgBitmap.Canvas.Draw (0,0, heightBm);
//  test shading routine
  Convert256To3d (heightBm, palIntBm, bgBitmap, 100,pageY-100);
end;

//*********************************************************
// draw one animation picture
//---------------------------------------------------------
procedure AnimDraw;
begin
  {$ifdef test2}
    writeln (tst, 'Plasma.AnimDraw: ', ord(cb));
  {$endif}

  i1 := i1 + d1;   i2 := i2 + d2;
  i3 := i3 + d3;   i4 := i4 + d4;
  i5 := i5 + d5;   i6 := i6 + d6;
  valuePtr^.SetValue (7, byte(i1));
  case modeNo of
    1: AnimatePlasma1;
    2: AnimatePlasma2;
    3: AnimatePlasma3;
    4: AnimatePlasma4;
  end;
  if showMap then DrawBitmap (0,0, palIntBm, bgBitmap);
end;

//*********************************************************
procedure NextMode;
begin
  modeNo := 1 + modeNo mod modes;
  MarkTime;
end;

//*********************************************************
// handle animation input
//---------------------------------------------------------
procedure AnimInput;
begin
  {$ifdef test2}
    writeln (tst, 'Plasma.AnimInput: ', ord(cb));
  {$endif}

  with valuePtr^ do begin
    d1 := ByteValue (1);
    d2 := ByteValue (2);
    d3 := ByteValue (3);
    d4 := ByteValue (4);
    d5 := ByteValue (5);
    d6 := ByteValue (6);
  end;

  if animValues then valuePtr^.Animate;

  case cb of
    CR : NextMode;
    'r': valuePtr^.Reset;
    'c': RandomColors;
    'n': begin
           valuePtr^.RandomValues;
           RandomColors;
         end;
    'a': animValues := not animValues;
    'M': showMap := not showMap;
  end;
end;
//*********************************************************
// display animation specific help
//---------------------------------------------------------
procedure AnimHelp;
begin
  HelpText (1, '<ret> effect ' + IntToStr(modeNo));
  HelpText (2, '  a   animate values');
  HelpText (3, '  c   random colors'  );
  HelpText (4, '  n   random values'  );
end;

//*********************************************************
initialization
  {$ifdef InitTest}
    writeln (tst, 'Plasma.initialization: ');
    flush (tst);
  {$endif}

  CalcSinusTab;

  RegisterAnimation (' RGB Plasma ',
    AnimStart, AnimResize, AnimDraw, AnimInput, AnimHelp);

end.

