//*******************************************************************
//
//  UNIT:     R i p p l e                      16.03.2000, 05.04.2002
//
//  animated ripple effects, coded by Gerd Platl
//
//*******************************************************************
{$I comp_opt.inc}

unit Ripple;

{$ifopt D+}
  {$define xtest1    main functions}
  {$define xtest2    sub functions}
{$endif}

interface

//*********************************************************
implementation

uses  SysUtils, Graphics,
      SysUtil2, Values, Graph, Palette, Images, GrForm;

const modes = 2;                       // moving modes
      frame = 40;                      // picture frame pixels

var   modeNo       :word = 2;          // actual demo mode
      origBm       :tBitmap = nil;     // original background picture
      backBm       :tBitmap = nil;     // transformed background picture
      backX, backY :word;              // background picture dimensions
      picFileNo    :word = 0;          // actual picture number

var   scanline: array [0..2048] of pointer;

//*********************************************************
// calculate all scanline pointers
//---------------------------------------------------------
procedure CreateScanlinePointerArray (bm: tBitmap);
var   yi: integer;
begin
  for yi := 0 to bm.height - 1
  do scanline[yi] := bm.ScanLine[yi];
end;

//*********************************************************
// stretch loaded picture to oversized background picture
//---------------------------------------------------------
procedure StretchPicture;
begin
  backX := bgBitmap.Width  + 2 * frame;
  backY := bgBitmap.Height + 2 * frame;
  backBm.Width  := backX;
  backBm.Height := backY;
  StretchBitmap (origBm, backBm);
  CreateScanlinePointerArray (backBm);
end;
//*********************************************************
// load next picture from file
//---------------------------------------------------------
procedure GetNextPicture;
begin
  picFileNo := LoadDemoPicture (origBm);
  if picFileNo = 0
  then MainForm.FlashInfo ('no ripple files found')
  else StretchPicture;
  MarkTime;
end;

//=========================================================
// Ripple
//=========================================================

const xm = 0.0001;                     // scaling
      ym = 0.0001;

var   moveSpeed:   integer = 10;       // moving speed
      xFreq:       integer = 64;       // x frequence
      xSize:       integer = 29;       // x step size
      yFreq:       integer = 59;       // y frequence
      ySize:       integer = 32;       // y step size

{********************************************************}
procedure CreateValues;
begin
  with valuePtr^
  do if ValueCount = 0
  then begin
    AppendValue ('moving speed', 1,  19, moveSpeed);
    AppendValue ('x frequence',  5, 115, xFreq);
    AppendValue ('x size',       5, 115, xSize);
    AppendValue ('y frequence',  5, 115, yFreq);
    AppendValue ('y size',       5, 115, ySize);
  end;
end;

//*********************************************************
// start new animation
//---------------------------------------------------------
procedure AnimStart;
begin
  {$ifdef test1}
    writeln (tst, #13#10'Ripple.StartAnim: ', modeNo);
  {$endif}

  CreateValues;
  SetGraphMode (pf16bit);
  GetNextPicture;
end;
//*********************************************************
// resize animation window
//---------------------------------------------------------
procedure AnimResize;
begin
  {$ifdef test1}
    writeln (tst, 'Ripple.AnimResize: ',pageX,'*',pageY);
  {$endif}

  StretchPicture;       // resize background bitmap
end;
//*********************************************************
// switch to next mode
//---------------------------------------------------------
procedure NextMode;
begin
  modeNo := 1 + modeNo mod modes;
  AnimResize;
end;
//*********************************************************
// change only vertical ripples
//---------------------------------------------------------
procedure DrawRipplePicture1;

var   yi, dy, lx,ly: longint;
      srcPtr: pByteArray;

begin
  for yi := 0 to GetMaxY
  do begin
    lx := round (xSize * (1.0 + sin ((picNo+yi) * 2*xFreq * xm)) / 4) + frame;
    dy := round (ySize * 0.4 *  sin ((picNo+yi) * yFreq * ym));
    ly := frame + yi + dy;

//    if      lx < 0     then lx := 0
//    else if lx >= backX then lx := backX - 1;

    if      ly < 0     then ly := 0
    else if ly >= backY then ly := backY - 1;

    srcPtr := GetPixelPointer (backBm, lx,ly);
    PutLine (srcPtr, 0,yi, pageX);

    if demoHelp
    then begin
      PutPixel (lx, yi, clLightGreen);
      PutPixel ((pageX + dy) div 2, yi, clYellow);
    end;
  end;
end;
//*********************************************************
// change horizontal and vertical ripples
//---------------------------------------------------------
procedure DrawRipplePicture2;

var   xi,yi, dx,dy, lx,ly: longint;
      xv: array[0..1600] of integer;
      sp: pWordArray;       // source line pointer
      dp: ^word;            // destination pointer
begin
  for xi := 0 to backX-1
  do begin
    //ToDo: change formula
    dx := round ((xSize * sin ((picNo+xi)     * xFreq * xm)
                 +xSize * sin ((picNo+xi) * 3 * xFreq * ym)) / 4);
    lx := xi+dx;
    if      lx < 0 then lx := 0
    else if lx >= backX then lx := backX - 1;
    xv[xi] := lx;
    if demoHelp
    then if xi < pageX then PutPixel (xi, lx, clYellow);
  end;

  for yi := 0 to GetMaxY
  do begin
    dy := round ((ySize * sin ((picNo+yi)     * yFreq * xm)
                 +ySize * sin ((picNo+yi) * 3 * yFreq * ym)) / 4);
    ly := frame + yi + dy;
    if      ly < 0 then ly := 0
    else if ly >= backY then ly := backY -1;

    sp := scanLine[ly];           // sp := backBm.ScanLine[ly];
    dp := GetLinePointer (yi);
    for xi := 0 to GetMaxX
    do begin
      dp^ := sp^[xv[xi+frame]];   // draw pixel
      inc(dp);                    // x+1
    end;

    if demoHelp
    then PutPixel ((pageX + dy) div 2, yi, clYellow);
  end;

  if demoHelp
  then for xi := 0 to GetMaxX
  do PutPixel (xi, GetMaxY div 2 + xv[xi]-xi, clLightGreen);
end;

//*********************************************************
// draw one animation picture
//---------------------------------------------------------
procedure AnimDraw;
begin
  case modeNo of
    1: DrawRipplePicture1;
    2: DrawRipplePicture2;
  end;
  inc (picNo, moveSpeed);
end;
//*********************************************************
// handle animation input
//---------------------------------------------------------
procedure AnimInput;
begin
  {$ifdef test2}
    writeln (tst, 'Ripple.AnimInput: ', ord(cb));
  {$endif}

  with valuePtr^ do begin
    moveSpeed := IntValue (1);
    xFreq     := IntValue (2);
    xSize     := IntValue (3);
    yFreq     := IntValue (4);
    ySize     := IntValue (5);
  end;

  case cb of
    'r': valuePtr^.Reset;
    CR : valuePtr^.RandomValues;
    'e': NextMode;
    '#': GetNextPicture;
    '': begin   PrevDemoPicture;   GetNextPicture;   end;
  end;
end;
//*********************************************************
// display animation specific help
//---------------------------------------------------------
procedure AnimHelp;
begin
  HelpText (1, ' <ret>   random values ');
  HelpText (2, '  e   effect '  + IntToStr(modeNo));
  HelpText (3, '  #   picture ' + IntToStr(picFileNo));
end;

//*********************************************************
initialization

  {$ifdef InitTest}
    writeln (tst, 'Ripple.initialization: ');
    flush (tst);
  {$endif}

  backBm := tBitmap.Create;
  backBm.PixelFormat := pf16Bit;

  RegisterAnimation (' Ripple Effect ',
    AnimStart, AnimResize, AnimDraw, AnimInput, AnimHelp);

//----------------------------------------------------------
finalization

  {$ifdef ExitTest}
    writeln (tst, 'Ripple.finalization: ');
    flush (tst);
  {$endif}

  backBm.Free;
end.

