
//*********************************************************
//
//  UNIT:  R o t 3 d                   08.10.99, 05.04.2002
//
//  rotate 3d object with different dots
//  coded by Gerd Platl
//
//*********************************************************
{$I comp_opt.inc}

unit Rot3d;

interface

procedure ResizeRot3d;
procedure Rotate3d;
procedure SetBaseColor (baseCol: byte);
procedure ChangeEffect (effectNo: integer);

//*********************************************************
implementation

uses  SysUtils, Graphics,
      SysUtil2, Math2, ColorRGB, Graph, Palette, Blur, Values, GrForm;

{$I float.inc}

//-- animation
const modes = 3;                       // # of effects
      figures = 4;                     // # of figures

var   modeNo:      integer =    1;     // mode number
      figureNo:    integer =    1;     // figure number
      moveSpeed:   integer =    2;     // moving speed
      turnSpeed:   integer =    4;     // rotation speed
      objRadius:   integer =  100;     // object radius
      objPoints:   integer =  300;     // object points
      distance:    integer =  300;     // viewing distance
      phi:         integer =    0;     // rotation angle
      rho:         integer =    0;     // moving angle

//-- sinus, cosinus
const tabLen = 720;

var   sinTab: Array[0..tabLen] Of float;
      cosTab: Array[0..tabLen] Of float;

//-- dots
const dotPixels = 24;

type  tDot = Array[0..dotPixels] of Byte;
      pDot = ^byte;

const dot2: tDot
      =(  0,  8, 16,  8,  0
       ,  8, 32, 48, 32,  8
       , 16, 48, 63, 48, 16
       ,  8, 32, 48, 32,  8
       ,  0,  8, 16,  8,  0);

const c = 5;         // intensity
      dot3: tDot
      =(  1,2+c,3+c,2+c,  1
       ,2+c,3+c,4+c,3+c,2+c
       ,3+c,4+c,5+c,4+c,3+c
       ,2+c,3+c,4+c,3+c,2+c
       ,  1,2+c,3+c,2+c,  1);

var   dotObj: tDot;

//-- points
type  v3d = record x,y,z: float; end;

const zeroVector3d: v3d = (x:0.0; y:0.0; z:0.0);

const minPoints =   16;
      maxPoints = 1000;

type  tPointArray = array[0..maxPoints] of v3d;
      tColorArray = array[0..maxPoints] of tColor;

var   pPoint: ^tPointArray;
      pColor: ^tColorArray;
      points: word = 0;

// fixed sphere point figure
const iPoints = 339;
      iPoint: array[0..iPoints,0..2] of smallInt =
((0,74,0),    (-70,-25,2),  (37,-25,60),  (33,-25,-62),
(-10,74,0),   (-21,71,0),   (-31,67,1),   (-41,62,1),
(-50,55,1),   (-57,47,1),   (-64,38,2),   (-69,28,2),
(-72,18,2),   (-74,7,2),    (-75,-3,2),   (-73,-14,2),
(6,74,9),     (11,71,18),   (17,67,27),   (22,62,35),
(26,55,42),   (31,47,49),   (34,38,54),   (37,28,58),
(38,18,61),   (39,7,63),    (40,-3,63),   (39,-14,62),
(5,74,-9),    (10,71,-19),  (15,67,-28),  (19,62,-37),
(23,55,-44),  (27,47,-51),  (30,38,-57),  (32,28,-61),
(34,18,-64),  (35,7,-66),   (35,-3,-66),  (34,-14,-65),
(-67,-30,11), (-63,-34,21), (-57,-38,29), (-50,-40,37),
(-42,-42,45), (-33,-43,51), (-23,-43,56), (-13,-42,60),
(-2,-40,63),  (7,-38,64),   (18,-34,64),  (28,-30,62),
(44,-30,52),  (50,-34,44),  (54,-38,35),  (58,-40,24),
(60,-42,14),  (61,-43,3),   (61,-43,-7),  (59,-42,-18),
(56,-40,-29), (52,-38,-39), (47,-34,-48), (40,-30,-56),
(23,-30,-65), (13,-34,-65), (3,-38,-65),  (-7,-40,-63),
(-17,-42,-59),(-27,-43,-55),(-37,-43,-49),(-45,-42,-42),
(-53,-40,-34),(-59,-38,-25),(-64,-34,-16),(-68,-30,-7),
(-5,74,9),    (-16,72,10),  (0,72,19),    (-27,68,11),
(-11,71,20),  (5,68,29),    (-37,63,12),  (-22,67,22),
(-6,67,31),   (10,63,38),   (-47,56,13),  (-33,62,24),
(-18,64,34),  (-1,62,41),   (15,56,46),   (-55,47,15),
(-44,53,27),  (-30,57,37),  (-14,57,46),  (2,53,52),
(18,47,54),   (-63,36,16),  (-53,42,29),  (-41,46,41),
(-27,47,50),  (-11,46,57),  (5,42,61),    (21,36,61),
(-68,23,17),  (-61,28,31),  (-51,31,44),  (-38,33,54),
(-23,33,62),  (-8,31,67),   (8,28,68),    (24,23,67),
(-72,9,17),   (-66,11,32),  (-58,13,45),  (-47,14,56),
(-34,14,65),  (-20,14,70),  (-4,13,73),   (10,11,73),
(25,9,69),    (-73,-4,16),  (-68,-5,30),  (-61,-6,42),
(-52,-7,53),  (-41,-7,62),  (-28,-7,69),  (-14,-7,73),
(0,-6,74),    (13,-5,73),   (27,-4,70),   (-71,-18,14),
(-67,-21,26), (-60,-24,36), (-53,-26,46), (-43,-27,55),
(-32,-27,61), (-21,-27,66), (-8,-26,70),  (3,-24,71),
(16,-21,70),  (28,-18,67),  (11,74,0),    (17,72,8),
(17,72,-10),  (23,68,17),   (24,71,-1),   (22,68,-19),
(30,63,26),   (31,67,8),    (30,67,-10),  (28,63,-28),
(35,56,34),   (38,62,16),   (39,64,-1),   (37,62,-19),
(33,56,-36),  (41,47,40),   (46,53,24),   (48,57,7),
(47,57,-10),  (44,53,-28),  (38,47,-43),  (46,36,46),
(53,42,31),   (57,46,15),   (57,47,-2),   (55,46,-19),
(50,42,-35),  (42,36,-49),  (50,23,50),   (58,28,37),
(64,31,22),   (67,33,5),    (66,33,-10),  (62,31,-26),
(55,28,-41),  (46,23,-54),  (51,9,53),    (61,11,41),
(68,13,27),   (72,14,12),   (73,14,-2),   (71,14,-18),
(66,13,-32),  (58,11,-46),  (47,9,-57),   (51,-4,54),
(60,-5,44),   (68,-6,32),   (72,-7,18),   (75,-7,4),
(74,-7,-9),   (71,-7,-23),  (65,-6,-36),  (57,-5,-48),
(47,-4,-58),  (48,-18,54),  (56,-21,45),  (62,-24,34),
(67,-26,22),  (69,-27,10),  (70,-27,-2),  (68,-27,-15),
(65,-26,-27), (60,-24,-38), (53,-21,-49), (44,-18,-58),
(-5,74,-9),   (0,72,-19),   (-17,72,-9),  (3,68,-29),
(-12,71,-20), (-27,68,-9),  (8,63,-39),   (-8,67,-31),
(-24,67,-21), (-38,63,-10), (11,56,-48),  (-4,62,-41),
(-20,64,-33), (-35,62,-22), (-48,56,-10), (15,47,-56),
(-1,53,-52),  (-17,57,-45), (-32,57,-35), (-46,53,-24),
(-56,47,-11), (17,36,-63),  (1,42,-61),   (-15,46,-57),
(-30,47,-49), (-44,46,-38), (-55,42,-26), (-64,36,-12),
(19,23,-68),  (3,28,-69),   (-12,31,-67), (-28,33,-61),
(-42,33,-52), (-54,31,-41), (-63,28,-27), (-69,23,-12),
(20,9,-71),   (5,11,-74),   (-9,13,-73),  (-24,14,-69),
(-38,14,-62), (-51,14,-53), (-61,13,-41), (-68,11,-27),
(-73,9,-12),  (22,-4,-71),  (8,-5,-74),   (-6,-6,-74),
(-20,-7,-72), (-33,-7,-67), (-45,-7,-59), (-55,-7,-49),
(-64,-6,-38), (-70,-5,-25), (-74,-4,-11), (23,-18,-69),
(11,-21,-71), (-1,-24,-71), (-13,-26,-69),(-25,-27,-65),
(-37,-27,-59),(-47,-27,-52),(-56,-26,-43),(-63,-24,-32),
(-68,-21,-21),(-72,-18,-9), (-66,-35,2),  (-62,-40,-7),
(-62,-40,11), (-58,-45,-16),(-59,-46,2),  (-56,-45,20),
(-51,-48,-25),(-54,-52,-7), (-53,-52,11), (-49,-48,29),
(-43,-51,-33),(-46,-56,-16),(-47,-58,1),  (-45,-56,19),
(-40,-51,36), (-33,-53,-41),(-36,-60,-25),(-38,-64,-7),
(-37,-64,10), (-35,-60,27), (-30,-53,43), (-21,-54,-47),
(-24,-63,-32),(-25,-69,-16),(-26,-71,0),  (-24,-69,18),
(-22,-63,34), (-18,-54,48), (-8,-53,-52), (-9,-63,-39),
(-9,-70,-24), (-9,-74,-8),  (-9,-74,8),   (-8,-70,24),
(-6,-63,39),  (-4,-53,52),  (5,-50,-55),  (7,-60,-44),
(8,-68,-30),  (9,-73,-16),  (10,-74,0),   (11,-73,15),
(10,-68,29),  (10,-60,43),  (9,-50,55),   (18,-45,-57),
(23,-54,-47), (26,-61,-35), (29,-66,-22), (31,-68,-8),
(31,-68,5),   (31,-66,20),  (29,-61,33),  (26,-54,45),
(22,-45,55),  (30,-38,-57), (36,-44,-48), (41,-50,-38),
(45,-54,-26), (48,-56,-14), (49,-57,-1),  (49,-56,10),
(47,-54,23),  (44,-50,35),  (40,-44,45),  (34,-38,55));

procedure Generate3dPoints;                                Forward;

type  tDrawDot = procedure (x,y: Integer);

var   DrawDot: tDrawDot;

//*********************************************************
procedure CreateValues;
begin
  with valuePtr^
  do if ValueCount = 0
  then begin
    AppendValue ('moveSpeed',  -10,   14);
    AppendValue ('turnSpeed',  -10,   12);
    AppendValue ('distance',   200,  600);
    AppendValue ('radius',      50,  150);
    AppendValue ('points',     100, 1000);
    Reset;
  end;
end;

//*********************************************************
// generate palette part  ci: 0..3, int: intensity 155..255
//---------------------------------------------------------
var   colorPalette: tPalette;

procedure PalettePart (ci, int: byte);

var   spreadValues: array[0..2] of tColorRGBI;
begin
  ci := ci * 64;
  spreadValues[0].rb := 0;
  spreadValues[0].gb := 0;
  spreadValues[0].bb := 0;
  spreadValues[0].ii := ci;

  spreadValues[1].rb := byte ((int + sinB[random(256)]) div 2);
  spreadValues[1].gb := byte ((int + sinB[random(256)]) div 2);
  spreadValues[1].bb := byte ((int + sinB[random(256)]) div 2);
  spreadValues[1].ii := ci+58;

  spreadValues[2].rb := int;
  spreadValues[2].gb := 255;
  spreadValues[2].bb := 255;
  spreadValues[2].ii := ci+63;
  SetSpreadPalette (colorPalette, spreadValues, 3);
end;
//*********************************************************
// generate palette
//---------------------------------------------------------
procedure AnimateColorMap;
begin
  PalettePart (0, 155 + random(100));
  PalettePart (1, 155 + random(100));
  PalettePart (2, 155 + random(100));
  PalettePart (3, 155 + random(100));
  SetAllPalette (colorPalette);
end;

//*********************************************************
procedure InitSinCos;
var   ni: Integer;
begin
  for ni := 0 to tabLen
  do begin
    sinTab[ni] := sin (ni*2*Pi / tabLen);
    cosTab[ni] := cos (ni*2*Pi / tabLen);
  end;
end;
//*********************************************************
procedure SetVPoints;
var   ni: Integer;
begin
  for ni := 0 to iPoints
  do with pPoint^[ni]
  do begin
    x := iPoint[ni,0] * objRadius div 48;
    y := iPoint[ni,1] * objRadius div 48;
    z := iPoint[ni,2] * objRadius div 48;
  end;
  points := iPoints + 1;
end;

//*********************************************************
procedure RotXYZ (var p: v3d; phi: Word);
var   s,c, x,y,z: float;
begin
  c := cosTab[phi];
  s := sinTab[phi];
  {z}
    x := p.x * c - p.y * s;
    y := p.x * s + p.y * c;
  {y}
  p.x := p.z * s + x * c;
    z := p.z * c - x * s;
  {x}
  p.y := y * c - z * s;
  p.z := y * s + z * c;
end;

//*********************************************************
procedure SetBaseColor (baseCol: byte);
var   ni: integer;
begin
  for ni := 0 to dotPixels
  do dotObj[ni] := byte(baseCol + Dot2[ni]);
end;
//*********************************************************
// draw dot as pixel
//---------------------------------------------------------
procedure DrawDot1 (x,y: Integer);
begin
  PutPixel (x,y, 250);
end;
//*********************************************************
// add transparent dot
//---------------------------------------------------------
procedure DrawDot2 (x,y: Integer);
var   xi,yi, xe,ye, yp: word;
      pixelPtr, dotPtr: ^byte;
begin
  if x > GetMaxX then exit;
  if y > GetMaxY then exit;
  if x < 0 then exit;
  if y < 0 then exit;
  if x+4 < pageX then xe := 5 else xe := pageX - x;
  if y+4 < pageY then ye := 5 else ye := pageY - y;
  yp := y;
  dotPtr := @Dot3;
  for yi := 1 to ye
  do begin
    pixelPtr := GetPixelPointer (x,yp);
    for xi := 1 to xe
    do begin
      pixelPtr^ := byte(pixelPtr^ + dotPtr^);
      if pixelptr^ > 63 then pixelptr^ := 63;
      inc (pixelPtr);
      inc (dotPtr);
    end;
    inc (yp);
  end;
end;
//*********************************************************
// draw solid dot
//---------------------------------------------------------
procedure DrawDot3 (x,y: Integer);
var   xi,yi, xe,ye, yp: word;
      pix: byte;
      pixelPtr, dotPtr: ^byte;
begin
  if x > GetMaxX then exit;
  if y > GetMaxY then exit;
  if x < 0 then exit;
  if y < 0 then exit;
  if x+4 < pageX then xe := 5 else xe := pageX - x;
  if y+4 < pageY then ye := 5 else ye := pageY - y;
  yp := y;
  dotPtr := @dotObj;
  for yi := 1 to ye
  do begin
    pixelPtr := GetPixelPointer (x,yp);
    for xi := 1 to xe
    do begin
      pix := dotPtr^;
      if pix <> 0 then pixelPtr^ := pix;
      inc (pixelPtr);
      inc (dotPtr);
    end;
    inc (yp);
  end;
end;

//*********************************************************
// convert vector to color with length = 255
//                            _    _
// Formel:      color = 255 * v / |v|
//---------------------------------------------------------
function VectorColor (var v: v3d): tColor;
var   lv: float;
begin
  with v do lv := sqrt (x*x + y*y + z*z);
  if lv < eps
  then VectorColor := iWhite
  else begin
    lv := lv * 255;
    VectorColor := WinColor(round (v.x / lv)
                           ,round (v.y / lv)
                           ,round (v.z / lv));
  end;
end;

//=========================================================
// animation
//=========================================================

//*********************************************************
// do rotation and draw dots
//---------------------------------------------------------
procedure Rotate3d;
var ni, mx,my, px,py: integer;   p: v3d;
begin
  mx := pageX div 2 + round (sinTab[rho] * 67);
  my := pageY div 2 + round (cosTab[rho] * 30);
  for ni := 0 to points - 1
  do begin
    p := pPoint^[ni];
    RotXYZ (p,phi);
    p.z := p.z + distance;
    px := mx + Round (p.x * 300 / p.z);
    py := my + Round (p.y * 300 / p.z);

    DrawDot (px,py);
  end;
  phi := (phi + turnSpeed + tabLen) mod tabLen;
  rho := (rho + moveSpeed + tabLen) mod tabLen;
end;

//*********************************************************
function SqrtPoints (cPoints: integer): integer;
begin
  if cPoints > maxPoints then cPoints := maxPoints;
  SqrtPoints := trunc (sqrt (cPoints));
end;
//*********************************************************
// generate cylinder points
//---------------------------------------------------------
procedure GenCylinder (radius, cPoints: integer);

var   segments, planes: word;
      ri, si: integer;
      d, r: float;
begin
  planes := SqrtPoints (cPoints);
  segments := cPoints div planes;
  d := 2 * radius / (planes - 1);
  points := 0;
  for ri := 0 to planes - 1
  do for si := 0 to segments - 1
  do with pPoint^[points]
  do begin
    r := ((180.0 / segments) * (ri and 1)
         +(360.0 / segments * si)) * deg;
    x := round (sin(r) * radius);
    y := round (cos(r) * radius);
    z := round (ri * d - radius);
    pColor[points] := VectorColor (pPoint^[points]);
    inc (points);
  end;
end;
//*********************************************************
// generate galaxis points
//---------------------------------------------------------
procedure GenGalaxis (radius, cPoints: integer);

var   segments, rings: word;
      minRadius, maxRadius, ri, si: integer;
      d, r, w: float;

begin
  rings := SqrtPoints(cPoints);
  minRadius := radius div 3;
  maxRadius := radius;
  segments := cPoints div rings;
  d := (maxRadius - minRadius) / rings;
  points := 0;
  for ri := 0 to rings - 1
  do begin
    r := minRadius + ri * d;
    for si := 0 to segments - 1
    do with pPoint^[points]
    do begin
      w := ((90.0 / segments) * ri + (360.0 / segments * si)) * deg;
      x := round (sin(w) * r);
      y := round (cos(w) * r);
      z := 0;
//      pixColor[points] := 50 - ri;
      pColor[points] := VectorColor (pPoint^[points]);
      inc (points);
    end;
  end;
end;
//*********************************************************
// http://www.swin.edu.au/astronomy/pbourke/geometry/owl/
//
// formula:
//   x =   v * cos(u) - 0.5 * v^2 * cos(2 u)
//
//   y = - v * sin(u) - 0.5 * v^2 * sin(2 u)
//
//   z = 4 / 3 * v^1.5 * cos(1.5 u)
//
// range:  0 <= u <= 4  and 0 <= v <= 1
//---------------------------------------------------------
procedure GenMaedersOwl (radius, cPoints: integer);

const nu = 100;
      nv = 61;
      rings = 12;

var   ri, si: integer;
      u,v,w: float;

begin
  // seg := cPoints div sp;  // segments
  pPoint^[0] := zeroVector3d;
  points := 1;
  ri := 0;
  repeat
    for si := 3 to rings + 2
    do with pPoint^[points]
    do begin
      u := ri * 4 * PI / nu;
      v := si * 0.8 * radius / nv;

      w := - 0.5 * v * v;
      x := w * cos(2 * u) + v * cos(u);
      y := w * sin(2 * u) - v * sin(u);
      z := 1.3333 * XPowerY(v,1.5) * cos(1.5 * u);
//         colour = GetColour(u, 0, 4*PI, 4);
      pColor[points] := VectorColor (pPoint^[points]);

      inc (points);
      if points > objPoints then exit;
    end;
    inc (ri);
  until false;
end;
//*********************************************************
// generate 3d points
//---------------------------------------------------------
procedure Generate3dPoints;
begin
  case figureNo of
    1: GenCylinder   (objRadius, objPoints);
    2: GenGalaxis    (2 * objRadius, objPoints);
    3: GenMaedersOwl (objRadius, objPoints);
    4: SetVPoints;
  end;
end;
//*********************************************************
procedure NextFigure;
begin
  figureNo := 1 + figureNo mod figures;
  Generate3dPoints;
  AnimateColorMap;
end;
//*********************************************************
procedure ChangeEffect (effectNo: integer);
begin
  modeNo := 1 + (effectNo - 1) mod modes;
  case modeNo of
    1: DrawDot := DrawDot1;
    2: DrawDot := DrawDot2;
    3: DrawDot := DrawDot3;
  end;
  MarkTime;
end;
//*********************************************************
procedure ChangeValues;
begin
  valuePtr^.RandomValues;
  Generate3dPoints;
end;
//*********************************************************
procedure ChangePoints (delta: longint);
begin
  objPoints := SysUtil2.Clamp (objPoints + delta, minPoints, maxPoints);
  valuePtr^.SetValue (5, objPoints);
  Generate3dPoints;
end;
//*********************************************************
// start new animation
//---------------------------------------------------------
procedure AnimStart;
begin
  {$ifdef test1}
    writeln (tst, #13#10'Rot3d.AnimStart: ');
  {$endif}

  CreateValues;
  SetGraphMode (pf8bit);
  NextFigure;
  ChangeEffect (modeNo + 1);
end;
//*********************************************************
// create / resize drawing map
//---------------------------------------------------------
procedure AnimResize;
begin
  Generate3dPoints;
  SetBaseColor (0);
end;
//---------------------------------------------------------
procedure ResizeRot3d;
begin
  AnimResize;
end;
//*********************************************************
// draw one animation picture
//---------------------------------------------------------
procedure AnimDraw;

begin
  {$ifdef test2}
    writeln (tst, 'Rot3d.AnimDraw: ');
  {$endif}

  if modeNo = 3
  then ClearViewport
  else Blur4L (bgBitmap, bgBitmap, 0, GetMaxY);
  Rotate3d;
end;
//*********************************************************
// handle animation input
//---------------------------------------------------------
procedure AnimInput;
var oldRadius, oldPoints: integer;
begin
  {$ifdef test2}
    writeln (tst, 'Rot3d.Input: ', ord(cb));
  {$endif}

  if animValues then valuePtr^.Animate;

  if animColors then AnimateColorMap;
  case cb of
    CR : NextFigure;
    'e': ChangeEffect (modeNo + 1);
    'r': valuePtr^.Reset;
    'n': ChangeValues;
    'a': animValues := not animValues;
    'c': AnimateColorMap;
    '+': ChangePoints (+5);
    '-': ChangePoints (-5);
  end;

  oldRadius := objRadius;
  oldPoints := objPoints;
  with valuePtr^ do begin
    moveSpeed :=  IntValue (1);
    turnSpeed :=  IntValue (2);
    distance  :=  IntValue (3);
    objRadius :=  IntValue (4);
    objPoints :=  IntValue (5);
  end;
  if objRadius <> oldRadius then Generate3dPoints;
  if objPoints <> oldPoints then Generate3dPoints;

end;
//*********************************************************
// display animation specific help
//---------------------------------------------------------
procedure AnimHelp;
begin
  HelpText (1, '<ret>  next figure');
  HelpText (2, '   e   effect '  + IntToStr(modeNo));
  HelpText (3, '   a   animate values');
  HelpText (4, '   c   random colors');
  HelpText (5, '   n   random values');
  HelpText (6, '  +/-  points ' + IntToStr(objPoints));
end;
//*********************************************************
initialization

  {$ifdef InitTest}
    writeln (tst, 'Rot3d.initialization: ');
    flush (tst);
  {$endif}

  GetMem (pPoint, SizeOf(tPointArray));
  GetMem (pColor, SizeOf(tColorArray));
  InitSinCos;
  DrawDot := DrawDot1;

  RegisterAnimation (' Rotate 3d ',
    AnimStart, AnimResize, AnimDraw, AnimInput, AnimHelp);

//----------------------------------------------------------
finalization

  {$ifdef ExitTest}
    writeln (tst, 'Rot3d.finalization: ');
    flush (tst);
  {$endif}

end.

