//*******************************************************************
//
//  UNIT:     R o t o Z o o m                    02.06.99, 04.10.2002
//
//  picture rotation and zooming with integer arithmetic
//  coded by Gerd Platl
//
//  Ref.  http://www.flipcode.com/demomaking/issue10.shtml
//
//*******************************************************************
{$I comp_opt.inc}

unit RotoZoom;

{Q+}   // must be set for correct integer arithmetic, don't know why

{$ifopt D+}
  {$define xtest1    main functions}
  {$define xtest2    sub functions}
{$endif}

interface

//=========================================================
implementation

uses  SysUtils, Classes, Windows, Graphics, Math, Values,
      SysUtil2, Graph, Palette, Images, GrForm;

//=========================================================
// Picture Rotation Demo
//=========================================================

const modes = 4;                       // # of moving modes

var   modeNo       :word    =   0;     // actual demo mode
      rCnt         :longint =   0;     // speed counter
      moveSpeed    :integer =  64;     // moving speed
      baseSize     :integer = 128;     // base zoom size
      rotBm        :tBitmap = nil;     // rotation bitmap
      xMask, yMask, ym: longint;       // texture masking
      xShift,yShift:integer;           // texture shifting

//--- file handling ---

const picFileName  = 'Images\Rot';     // name of picture files
      picFileExt   = '.BMP';           // file extension

var   picFiles     :word = 0;          // # of picture files
      picFileNo    :word = 0;          // actual picture number
      picName      :string[40] = '';   // actual file name

{********************************************************}
procedure CreateValues;
begin
  with valuePtr^
  do if ValueCount = 0
  then begin
    AppendValue ('move speed',  1,  121);
    AppendValue ('base size',   1,  250);
  end;
end;

//---------------------------------------------------------
procedure ResetParameter;
begin
  rCnt := 0;
  valuePtr^.Reset;
end;
//---------------------------------------------------------
// sinD(x) = sin(pi*x/180)*size       in: x[degree]
//---------------------------------------------------------
function sinD (x, size: extended): extended;
begin
  sinD := sin (pi*x/180) * size;
end;
//---------------------------------------------------------
// cosD(x) = cos(pi*x/180)*size
//---------------------------------------------------------
function cosD (x, size: extended): extended;
begin
  cosD := cos (pi*x/180) * size;
end;
//---------------------------------------------------------
// return shift and mask value of given pix size
//---------------------------------------------------------
function TextureMask(pix: integer; var shift: integer):integer;
begin
  shift := 0;
  pix := pix shr 1;
  while pix > 0
  do begin
    inc (shift);
    pix := pix shr 1;
  end;
  result := (1 shl shift) - 1;
end;
//---------------------------------------------------------
procedure PrepareTexture;
begin
  xMask := TextureMask (rotBm.width, xShift);
  yMask := TextureMask (rotBm.height, yShift);
  if (xMask+1 <> rotBm.width)
  or (yMask+1 <> rotBm.height)
  then begin                      // resize to 2^n size
    rotBm.width  := xMask + 1;
    rotBm.height := yMask + 1;
  end;
  ym := 1 shl yShift;
  yMask := yMask shl xShift;
  xShift := 16 - xShift;
end;
//---------------------------------------------------------
procedure LoadNextPicture;
begin
  if picFiles = 0
  then MainForm.FlashInfo ('no picture files found')
  else begin
    // load picture with 2^n * 2^n / 256 for roto zoomer
    picFileNo := 1 + picFileNo mod picFiles;
    picName := picFileName + IntToStr(picFileNo) + picFileExt;
    if LoadBitmap2 (rotBm, picName)
    then if rotBM.PixelFormat <> pf8Bit
    then rotBM.PixelFormat := pf8Bit;
  end;
end;
//*********************************************************
procedure NextMode;
begin
  modeNo := 1 + modeNo mod modes;
{
  MainForm.GetRotatePicture (modeNo, rotBm);
}
  LoadNextPicture;
  UseBitmapPalette (rotBm);
  PrepareTexture;
  MarkTime;
end;
//*********************************************************
// start new animation
//---------------------------------------------------------
procedure AnimStart;
begin
  {$ifdef test1}
    writeln (tst, #13#10'RotoZoom.AnimStart: ', modeNo);
  {$endif}

  CreateValues;
  ClearViewport;
  SetGraphMode (pf8Bit);
  NextMode;
end;

//---------------------------------------------------------
// do picture rotation and zooming with integer arithmetic
// input:      x0,y0:   texture origin
//             du,dv:   direction step 16.16 fixpoint
//             x1,y1,x2,y2: destination drawing rectangle
//   global:   rotBm: rotation bitmap
//---------------------------------------------------------
// ToDo: correct texture origin calculation
// ToDo: eliminate this compiler switch
{$R-}   // because byte array may be greater than 65536 !!!
procedure RotatePicture (src: pByteArray;
                         x0,y0, du,dv: longint;
                         x1,y1, x2,y2: integer);

var   xi,yi,            // destination pixel indizes
      u1,v1,            // texture start position
      u,v: longint;     // texture pixel position
      dst: pByte;       // pixel destination pointer
begin
//  u1 := 0;
  v1 := 0;
  u1 := (y0 * dv - x0 * du);
//  v1 := (y0 * du + x0 * dv) div Ym * 256;
  for yi := y1 to y2
  do begin
    inc (u1,dv);            // vert. texture step
    dec (v1,du);
    u := u1;                // set start position
    v := v1;
    dst := GetPixelPointer (x1,yi);
    for xi := x1 to x2
    do begin
      dst^ := src^[(u shr 16) and xMask + (v shr xShift) and yMask];
      inc (dst);
      inc (u,du);           // hor. texture step
      inc (v,dv);
    end;
  end;
end;
//*********************************************************
// draw one animation picture
//---------------------------------------------------------
var   du,dv: longint;   // texture steps

procedure AnimDraw;

var   ra,sd: extended;
      sa,ca: extended;  // sinus, cosinus values
      px,py: longint;   // texture origin
begin
  {$ifdef test2}
    writeln (tst, 'RotoZoom.AnimDraw: ', modeNo);
  {$endif}

  inc (rCnt, moveSpeed);
  px := 0;
  py := 0;
  du := 0;
  dv := 0;
  case modeNo of
    1: // rotozoom
       begin
         sd := rCnt / 64.0;
         ra := sinD(sd,128);
         du := 256 * baseSize + round (sinD(ra+sd    ,512) * sinD(sd+22,128)-160);
         dv := 256 * baseSize + round (sinD(ra+sd+555,512) * sinD(sd+44,128)-160);
       end;
    2: // wobble
       begin
         ra := sinD (rCnt / 16.0, 30);
         px := rotBm.width  div 2;
         py := rotBm.height div 2;
         SinCos (degToRad(ra), sa,ca);
         du := round (sa * 1024 * baseSize);
         dv := round (ca * 1024 * baseSize);
       end;
    3: // rotate
       begin
         px := rotBm.width  div 2;
         py := rotBm.height div 2;
         SinCos (degToRad(rCnt / 64.0), sa,ca);
         du := round (sa * 1024 * baseSize);
         dv := round (ca * 1024 * baseSize);
       end;
    4: // sizing
       begin
         sd := rCnt / 16.0;     // zoom
         px := rotBm.width  div 2;
         py := rotBm.height div 2;
         du := 1024 * baseSize + round (sinD(sd, 32000));
       end;
  end;
  RotatePicture (rotBm.scanline[rotBm.height-1]
                ,px,py, du,dv, 0,0, GetMaxX,GetMaxY);
end;

//*********************************************************
// handle animation input
//---------------------------------------------------------
procedure AnimInput;
begin
  {$ifdef test2}
    writeln (tst, 'Title.AnimInput: ', ord(cb));
  {$endif}

  with valuePtr^ do begin
    moveSpeed  := IntValue (1);
    baseSize   := IntValue (2);
  end;

  case cb of
    'r': ResetParameter;
    CR : modeNo := 1 + modeNo mod modes;
    'n': valuePtr^.RandomValues;
    '#': NextMode;
  end;
end;
//*********************************************************
// display animation specific help
//---------------------------------------------------------
procedure AnimHelp;
{$ifopt D+}
  var   an, zf: longint;
{$endif}
begin
  HelpText (1, '<ret> movement ' + IntToStr(modeNo));
  HelpText (2, '  #   picture ' + IntToStr(picFileNo));
  HelpText (3, '  n   random values');
{$ifopt D+}
  an := round (math.arcTan2(dv/256, du/256) * -180.0 / pi);
  zf := round (Sqrt (sqr(du/256) + sqr(dv/256)));
  HelpText (4, ' angle = ' + IntToStr(an));
  HelpText (5, ' zoom = '  + IntToStr(zf));
//  HelpText (6, ' du/dv = '  + IntToStr(du div 256)
//                      + ' ' + IntToStr(dv div 256));
{$endif}
end;
//*********************************************************
initialization
  {$ifdef InitTest}
    writeln (tst, 'RotoZoom.initialization: ');
    flush (tst);
  {$endif}

  rotBm := tBitmap.Create;
  rotBm.PixelFormat := pf8Bit;
  picFiles := GetFileCount (picFileName, picFileExt);

  RegisterAnimation (' Fast Zoom Rotation ',
    AnimStart, nil, AnimDraw, AnimInput, AnimHelp);

//---------------------------------------------------------
finalization
  {$ifdef ExitTest}
    writeln (tst, 'RotoZoom.finalization: ');
    flush (tst);
  {$endif}

  rotBm.Free;
end.

