//*******************************************************************
//
//  UNIT:     S h e l l y                        29.06.98, 04.10.2002
//
//  simulation of shell pattern generation
//  coded by Gerd Platl
//
//  feature: creating new rules on floating point exception
//
//  References:
//
//   The Algorithmic Beauty of Sea Shells, Hans Meinhard
//   Springer 95, Basic
//
//   Wie Schnecken sich in Schale werfen, Hans Meinhard
//   Springer 97, Basic
//
//*******************************************************************
{$I 'comp_opt.inc'}

unit Shelly;

{$ifopt D+}
  {$define xtest1    main functions}
  {$define xtest2    sub functions}
{$endif}

interface

//*********************************************************
implementation

uses SysUtils, Graphics,
     SysUtil2, Graph, Palette, Values, GrForm;

{=========================================================}
{  P a l e t t e
{=========================================================}

const baseColor = 20;
      colNo: longint = 0;

var   f1,f2,f3: integer;
      cPal: tPalette;

{*********************************************************}
procedure InitColors;
begin
  colNo := random(5);
  f1 := 21+random(15);
  f2 := 22+random(14);
  f3 := 23+random(13);
end;
{*********************************************************}
procedure NextColors;
var   ci: byte;
begin
  InitPalette;
  GetAllPalette (cPal);
  for ci := baseColor to 255
  do with cPal[ci]
  do begin
    rb := 128+trunc((cos((colNo+ci)/f1))*127);
    gb := 128+trunc((cos((colNo+ci)/f2))*127);
    bb := 128+trunc((cos((colNo+ci)/f3))*127);
  end;
  SetAllPalette (cPal);
  inc(colNo,7);
end;
{*********************************************************}
procedure NewColors;
begin
  InitColors;
  NextColors;
end;

{=========================================================}
{  S h e l l y
{=========================================================}

const yl = 42;
      maxL = 1600;
      viewChart:  boolean = false;

type  float = double; {extended;}
      fArray = array[0..maxL] of float;

var   fa,fb,fz: fArray;

type  tShelly = record
        col: tColor;    // chart color
        d: float;       // diffusion coefficient < 0.4
        r: float;       // Abbaurate
        b: float;       // Basale
        s: float;       // Sttigung
        c: float;       // Kopplung
        a: float;       // Anfangskonzentration
        g: float;       // Generelle Anfangskonzentration
      end;

var   da,ra,sa,dac,          // activator
      db,rb,sb,dbc,          // inhibitor
      rc,cc: float;
      kp: longint = 5;       // iterations

      wdl: integer = 2000;   // wanted display line
      adl: integer = 0;      // actual display line

{********************************************************}
procedure CreateValues;
begin
  with valuePtr^
  do if ValueCount = 0
  then begin
    AppendValue ('A.d',  0.07,  0.20, 0, 2);
    AppendValue ('A.r',  0.08,  0.14, 0, 2);
    AppendValue ('A.s',  0.25,  0.35, 0, 2);

    AppendValue ('B.d',  0.01,  0.26, 0, 2);
    AppendValue ('B.r',  0.02,  0.04, 0, 2);
    AppendValue ('B.s',  0.10,  0.15, 0, 2);

    AppendValue ('rc',   0.15,  0.25, 0, 2);
    AppendValue ('cc',   0.20,  0.60, 0, 2);
    AppendValue ('iterations', 1, 10);
  end;
end;
{*********************************************************}
function RandomValue (minV, range: float): float;
begin
  result := minV + range * random;
end;
{*********************************************************}
procedure RestartValues;
var   xi,ci: integer;
begin
  FillChar (fa,sizeof(fArray),0);
  FillChar (fb,sizeof(fArray),0);
  FillChar (fz,sizeof(fArray),0);
  for xi := 0 to maxL
  do begin
    fb[xi] := RandomValue(0.01, 0.1);
    fz[xi] := ra * RandomValue(0.96, 0.08);
  end;
  ci := 0;
  for xi := 1 to maxL
  do begin
    ci := ci + 5 + random(40);
    if ci > maxL then break;
    fa[ci] := RandomValue(0.1, 9.9);
  end;
end;
{*********************************************************}
procedure InitShelly;
begin
  InitColors;
  NextColors;
  RestartValues;
  wdl := 2000;     // wanted display line
  adl := 0;        // active display line
end;
//*********************************************************
// start new animation
//---------------------------------------------------------
procedure AnimStart;
begin
  {$ifdef test1}
    writeln (tst, #13#10'Shelly.AnimStart: ');
  {$endif}

  CreateValues;
  SetGraphMode (pf8Bit);
  InitShelly;
end;
//*********************************************************
// resize animation window
//---------------------------------------------------------
procedure AnimResize;
begin
  {$ifdef test1}
    writeln (tst, 'Shelly.AnimResize: ',pageX,'*',pageY);
  {$endif}

  if viewChart
  then begin
    wdl := GetMaxY - yl;
    if wdl < 100 then wdl := 100;
  end
  else wdl := GetMaxY;
  if adl > GetMaxY then adl := GetMaxY;
  if wdl > GetMaxY then wdl := GetMaxY;
  SetColor (iBlack);
  Bar (0,adl, GetMaxX,GetMaxY);
end;
{*********************************************************}
procedure RandomPattern;
var   xi,ci: integer;
begin
  ci := 0;
  for xi := 1 to maxL
  do begin
    ci := ci+random(10);
    if ci > maxL then break;
    fa[ci] := RandomValue (0.1, 0.3);
  end;
end;
{*********************************************************}
procedure CheckValues;
var   xi: integer;
begin
  for xi := 1 to GetMaxX
  do if fa[xi] > 0.1 then exit;
  RestartValues;
end;
{*********************************************************}
procedure DrawChart;
var   xi, yi: integer;
begin
  yi := adl + 1;
  SetColor (iBlack);
  Bar (0,yi, GetMaxX,yi+yl);

  SetColor (iRed);
  MoveTo (0, yi + trunc(30*ln(1+fa[0])));
  for xi := 1 to pageX
  do LineTo (xi, yi + trunc(30*ln(1+fa[xi])));

  SetColor (iGreen);
  MoveTo (0, yi + trunc(30*ln(1+fb[0])));
  for xi := 1 to pageX
  do LineTo (xi, yi + trunc(30*ln(1+fb[xi])));
end;
{*********************************************************}
procedure RestartShelly;
begin
  valuePtr^.RandomValues;
  RestartValues;
  Beep;
end;
//*********************************************************
// draw one animation picture
//---------------------------------------------------------
procedure AnimDraw;

var   ii, xi, kx: integer;
      linePtr: pByteArray;
      af,bf, a2, aq, bsa, a1,b1: float;
begin
  {$ifdef test2}
    writeln (tst, 'Shelly.AnimDraw: ');
  {$endif}

  try
    {calculate new line}
    kx := GetMaxX-1;
    for ii := 1 to kp
    do begin
      a1 := fa[0];
      b1 := fb[0];
      bsa := 0;
      for xi := 1 to kx
      do begin
        af := fa[xi];   a2 := af*af;
        bf := fb[xi];
        aq := fz[xi]*a2/(1+sa*a2);
        fa[xi] := af * dac + da * (a1 + fa[xi+1]) + aq/(sb+bf);
        fb[xi] := bf * dbc + db * (b1 + fb[xi+1]) + aq;
        bsa := bsa + rc * af;
        a1 := af;
        b1 := bf;
        // prevent floation point exceptions
        if fa[xi] < 0.00001 then fa[xi] := 0.00001;
        if fb[xi] < 0.00001 then fb[xi] := 0.00001;
      end;
      cc := cc*(1-rc) + bsa/kx + 0.01;
      dbc := 1 - 2*db - rb/cc;
      // make endless pattern
      fa[GetMaxX] := fa[1];
      fb[GetMaxX] := fb[1];
      fa[0] := fa[kx];
      fb[0] := fb[kx];
    end;

    {draw new line}
    linePtr := GetLinePointer(adl);
    for xi := 0 to GetMaxX
    do linePtr^[xi] := byte(baseColor+trunc(200*ln(1+fa[xi])));

    if (adl < GetMaxY) and viewChart
    then DrawChart;

    if adl < wdl
    then inc (adl)
    else begin
      if adl > wdl
      then begin
        CopyRect (bgBitmap, 0,1, 0,0, pageX,adl);  // move picture 1 row up
        dec (adl);
      end;
      CopyRect (bgBitmap, 0,1, 0,0, pageX,adl);    // move picture 1 row up
    end;

    if picNo and $3F = 0 then CheckValues;

  except
    on EMathError     do RestartShelly;
    on EIntError      do RestartShelly;
  end;
end;

{*********************************************************}
procedure ToggleChart;
begin
  viewChart := not viewChart;
  AnimResize;
  MarkTime;
end;
//*********************************************************
// handle animation input
//---------------------------------------------------------
procedure AnimInput;
begin
  {$ifdef test2}
    writeln (tst, 'Shelly.AnimInput: ', ord(cb));
  {$endif}

  with valuePtr^ do begin
    da :=  FloatValue (1);
    ra :=  FloatValue (2);
    sa :=  FloatValue (3);
    db :=  FloatValue (4);
    rb :=  FloatValue (5);
    sb :=  FloatValue (6);
    rc :=  FloatValue (7);
    cc :=  FloatValue (8);
    kp :=  IntValue (9);
  end;
  dac := 1 - ra - 2*da;
  dbc := 1 - rb - 2*db;

  if animValues then valuePtr^.Animate;

  case cb of
    'i': InitShelly;
    'r': valuePtr^.Reset;
    CR : valuePtr^.RandomValues;
    'a': animValues := not animValues;
    'o': ToggleChart;
    'c': NewColors;
    'y': NextColors;      //ToDo: correct color cycling
    'z': RandomPattern;
  end;
end;
//*********************************************************
// display animation specific help
//---------------------------------------------------------
procedure AnimHelp;
begin
  HelpText (1, '  i    restart');
  HelpText (2, '<ret>  random values');
  HelpText (3, '  o    chart on/off');
  HelpText (4, '  c    random colors');
  HelpText (5, '  y    cycle colors');
  HelpText (6, '  z    random pattern');
end;
//*********************************************************
initialization

  {$ifdef InitTest}
    writeln (tst, 'Shelly.initialization: ');
    flush (tst);
  {$endif}

  RegisterAnimation (' Shell Pattern Generator ',
    AnimStart, AnimResize, AnimDraw, AnimInput, AnimHelp);

//----------------------------------------------------------
finalization

  {$ifdef ExitTest}
    writeln (tst, 'Shelly.finalization: ');
    flush (tst);
  {$endif}
end.

