
//*******************************************************************
//
//  UNIT:     S y s U t i l 2               14.01.2000,  24.06.2003
//
//  This Unit consists of additional system routines for Delphi-5/6
//
//-------------------------------------------------------------------
//  Min                    minimum value
//  Max                    maximum value
//  Clamp                  return value between given limits
//  ClampByte              return value between 0 and 255
//  ClampWord              return value between 0 and 65535
//  Swap                   swap values
//  SwapMem                swap memory values
//  Sort                   sort values
//
//  HexString              return value as hexadecimal string
//  HexToLong              convert hexadecimal string to long integer
//  OnOff                  return 'on' or 'off' string
//
//  Sound                  play sound with given frequency
//  NoSound                turn off sound
//  PlayWaveSound          play wave sound given by name without waiting
//
//  GetTime                return time values
//  GetDate                return date values
//  SystemTimeString       return actual system time as string [hh:mm:ss]
//  SystemDateString       return actual system date as string [dd.mm.yyyy]
//  MarkTime               set time marker
//  RunTimeMS              return runtime in milliseconds
//  FPS                    frames per second
//
//  WindowsVersion         return windows version string
//  CPUSpeed               return CPU speed [MHz]
//  CPUSpeed2              return CPU speed [MHz] (with priority handling)
//  ProcessorName          return name of processor
//
//  MemoryStatus             return memory values
//  AvailableVirtualMemory   return available virtual memory in bytes
//  AvailablePhysicalMemory  return available physical memory in bytes
//  TotalPhysicalMemory      return whole physical memory in bytes
//  TotalVirtualMemory       return whole virtual memory in bytes
//
//  GetDisplayInfo         return display information
//  DisplayInfoString      return display information string
//
//  ShowTaskBar            show or hide system task bar
//
//  GetDriveList           return list of existing drives
//  GetVolumeName          return volume name of given drive
//  GetCDVolumeName        return volume name of first CD drive
//  GetFileList            return file list
//  SubstituteFilename     return substitute filename
//
//  RaiseObjectNilError    Objektfehler Nil-Pointer anzeigen
//  RaiseObjectInitError   Objektfehler Initialisierung anzeigen
//  ShowMessageBox         Meldung anzeigen
//  ShowWarning            Warnung anzeigen
//  ShowError              display error
//  ShowFileError          Dateifehler anzeigen
//
//  --- global Variables ---
//  tst: text;             debug file, if compiler option is $D+
//  ioError: Integer       last file error
//  picCnt: longWord       picture counter
//  screenData             screen device data
//
//*******************************************************************
{$I comp_opt.inc}

unit SysUtil2;

{$define useAsm          definition to use asm-code}

interface

uses  SysUtils, Forms, Windows, Classes, Graphics, Types;

const No = False;
      Yes = True;

const NUL = #0;
      BEL = #7;
      DEL = #8;
      TAB = #9;
      LF  = #10;
      FF  = #12;
      CR  = #13;
      ESC = #27;
      BLANC = #32;
      CRLF = #13#10;

const VK_PAGE_UP = 33;    // VK_PRIOR = 33;
      VK_PAGE_DOWN = 34;  // VK_NEXT = 34;

const errZeroDivision      = 200;
      errOutOfRange        = 201;
      errStackOverflow     = 202;
      errHeapOverflow      = 203;
      errNoObjectInit      = 210;     // e.g. wrong object id
      errOutOfMemory       = 217;     // e.g. GetMem
      errNilPointer        = 221;
      errNilObject         = 222;     // e.g. New
      errObjectType        = 223;
      errNoTypeDefined     = 224;
      errFatalError        = 225;
      errNoGraphicActive   = 230;
      errFatalGraphicError = 231;
      errWrongGraphicMode  = 232;
      errNotImplemented    = 233;
      errUserBreak         = 255;

const maxByte     = $FF;
      maxWord     = $FFFF;
      maxLongWord = $FFFFFFFF;   //4294967295;
      maxInt64    = $7FFFFFFFFFFFFFFF;

type  PShortArray = ^TShortArray;
      TShortArray = array[0..maxWord] of shortInt;  // 8 Bit

      PSmallIntArray = ^TSmallIntArray;
      TSmallIntArray = array[0..maxWord] of smallInt;  // 16 Bit

      PLongIntArray = ^TLongIntArray;
      TLongIntArray = array[0..maxWord] of longInt;  // 32 Bit

      // dynamic array types additional to unit Types
      pShortIntArray   = ^TShortIntDynArray;

      pIntegerArray    = ^TIntegerDynArray;

      pPointerArray    = ^TPointerArray;
      tPointerArray    = array of pointer;

      pPointArray      = ^tPointArray;      // longint point(x,y)
      tPointArray      = array of Types.TPoint;

      pSmallPointArray = ^tSmallPointArray; // smallint point(x,y)
      tSmallPointArray = array of Types.tSmallPoint;

    tDeviceInfo = record          //--- GetDevicesCaps - Info ----
      dc             : HDC;       //
      x, y           : word;      // screen resolution [pixel]
      bitsPerPixel   : word;      // bits per pixel
      pixelFormat    : tPixelFormat;
      pPlanes        : word;      // pixel planes
      xSize, ySize   : word;      // screen size [mm]
      xAspect        : word;      // aspect ratio
      yAspect        : word;
      colRes         : word;      // colors = color resolution
      dVersion       : word;      // driver version
      // for Win-NT only
      refreshRate    : word;      // vertical refresh rate [hz]
      xDesktop, yDesktop: word;   // desktop size
    end;

// get minima or maxima (more overloaded functions in Unit Math2)
function Min (const a,b: longWord): longWord;              overload;
function Min (const a,b: longInt): longInt;                overload;
function Max (const a,b: longWord): longWord;              overload;
function Max (const a,b: longInt): longInt;                overload;

// return value between min and max value
function Clamp (const value, min, max: byte): byte;        overload;
function Clamp (const value, min, max: longint): longint;  overload;
function Clamp (const value, min, max: int64): int64;      overload;
function ClampByte (value: longInt): byte;                 overload;
function ClampByte (value: longWord): byte;                overload;
function ClampWord (value: longInt): word;                 overload;
function ClampWord (value: longWord): word;                overload;

// swap values
procedure Swap (var a,b: byte);                            overload;
procedure Swap (var a,b: shortInt);                        overload;
procedure Swap (var a,b: word);                            overload;
procedure Swap (var a,b: smallInt);                        overload;
procedure Swap (var a,b: longInt);                         overload;
procedure Swap (var a,b: cardinal);                        overload;
procedure Swap (var a,b: int64);                           overload;
procedure Swap (var a,b: pointer);                         overload;
procedure SwapMem (var source, dest; len: integer);

// sort values
procedure Sort (var a,b: byte);                            overload;
procedure Sort (var a,b: word);                            overload;
procedure Sort (var a,b: smallInt);                        overload;
procedure Sort (var a,b: longInt);                         overload;
procedure Sort (var a,b: longWord);                        overload;
procedure Sort (var a,b: int64);                           overload;

// conversions
function HexString (value: byte): string;                  overload;
function HexString (value: word): string;                  overload;
function HexString (value: longWord): string;              overload;
function HexString (value: int64): string;                 overload;
function HexString (value: pointer): string;               overload;
function HexString (value: tProcedure): string;            overload;
function HexToLong (text: string): int64;
function OnOff (flag: boolean): string;

// sound handling
procedure Sound (Hz: word);
procedure NoSound;
procedure PlayWaveSound (soundFile: string);

// time handling
procedure GetTime (var Hour, Min, Sec, MSec: word);
procedure GetDate (var Year, Month, Day: word);
function SystemTimeString: string;
function SystemDateString: string;
// MSec time measuring 1
procedure MarkTime;
function GetRuntimeMS: longWord;
function RunTimeMS: longWord;
function FPS: string;
// MSec time measuring 2
procedure MarkTime2;
function GetRuntimeMS2: longWord;
function FPS2: string;

// OS information
function WindowsVersion: string;
// HW information
function CPUSpeed: integer;   // works only for Pentium and higher
function CPUSpeed2: integer;
function ProcessorName: string;

// not useable if Win32PlatForm = VER_PLATFORM_WIN32_NT
procedure SetPort (portAddr: word; value: byte); assembler; stdcall;
function  GetPort (portAddr: word): byte; assembler; stdcall;

// memory information
function MemoryStatus: TMemoryStatus;
function AvailableVirtualMemory: Int64;
function AvailablePhysicalMemory: Int64;
function TotalPhysicalMemory: Int64;
function TotalVirtualMemory: Int64;

// screen information
procedure GetDisplayInfo (var screenInfo: tDeviceInfo);
function DisplayInfoString: string;

// window handling
procedure ShowTaskBar (show: boolean);

// drive and file handling
function GetDriveList: TStringList;
function GetVolumeName (drive: string): string;
function GetCDVolumeName: shortstring;
procedure GetFileList (var sList: TStringList; path, ext: string);
function SubstituteFilename (filename: string): string;

// display message and errors
procedure RaiseObjectNilError  (name: string);
procedure RaiseObjectInitError (name: string);
procedure ShowMessageBox (const caption, infoText: string);
procedure ShowWarning    (const caption, infoText: string);
function  ShowError      (const caption, infoText: string): integer;
function ShowFileError (hide: byte; filename: string): boolean;

//-- global values --
var   picCnt: longWord = 0;       // used for FPS calculation
                                  // reseted on key input and step mode

      screenInfo: tDeviceInfo;    // screen device data

      ioError: Integer;           // last file error

{--- debug file, if compiler option is set to D+ ---}
{$ifopt D+}
  const testFile = 'TEST.TMP';
  var   tst: text;
{$endif}

//*********************************************************
implementation

uses  Dialogs, MMSystem, Registry;

//*********************************************************
// get minimum value
//---------------------------------------------------------
function Min (const a,b: longWord): longWord;
begin
  if a < b   then result := a   else result := b;
end;
//---------------------------------------------------------
{$ifdef useAsm}
function Min (const a,b: longInt): longInt;   assembler;
asm
  cmp  eax, edx
  jle @ret                   // a < b -> result = a
  mov  eax, edx              // result = b
  @ret:
end;
{$else}
//---------------------------------------------------------
function Min (const a,b: longInt): longInt;
begin
  if a < b   then result := a   else result := b;
end;
{$endif}

//*********************************************************
// get maximum value
//---------------------------------------------------------
function Max (const a,b: longWord): longWord;
begin
  if a > b   then result := a   else result := b;
end;
//---------------------------------------------------------
{$ifdef useAsm}
function Max (const a,b: longInt): longInt;   assembler;
asm
  cmp  eax, edx
  jge @ret
  mov  eax, edx
  @ret:
end;
{$else}
//---------------------------------------------------------
function Max (const a,b: longInt): longInt;
begin
  if a > b   then result := a   else result := b;
end;
{$endif}

//=========================================================
// Clamp: return value between min and max
//=========================================================
{$ifdef useAsm}
//---------------------------------------------------------
function Clamp (const value, min, max: byte): byte; assembler;
asm            // AL = value;  DL = min;  CL = max
  cmp  al, dl
  jge @greater               // value > min  -> goto greater
  mov  al, dl                // result = min
  jp  @return
@greater:
  cmp  al, cl
  jle @return                // value <= max -> return
  mov  al, cl                // result = max
@return:
end;
//---------------------------------------------------------
// Achtung: Wert wird nur als Funktionsergebnis zurckgegeben !!!
//---------------------------------------------------------
function Clamp (const value, min, max: longint): longint; assembler;
asm            // EAX = value;  EDX = min;  ECX = max
  cmp  eax, edx
  jge @greater               // value > min  -> goto greater
  mov  eax, edx              // result = min
  jp  @return
@greater:
  cmp  eax, ecx
  jle @return                // value <= max -> return
  mov  eax, ecx              // result = max
@return:
end;
{$else}
//---------------------------------------------------------
function Clamp (const value, min, max: byte): byte;
begin
  if value < min then Result := min
  else if value > max then Result := max
  else Result := value;
end;
//---------------------------------------------------------
function Clamp (const value, min, max: longint): longint;
begin
  if value < min then Result := min
  else if value > max then Result := max
  else Result := value;
end;
{$endif}

//---------------------------------------------------------
function Clamp (const value, min, max: int64): int64;
begin
  if value < min then Result := min
  else if value > max then Result := max
  else Result := value;
end;

//---------------------------------------------------------
// return value between 0 and 255
//---------------------------------------------------------
function ClampByte (value: longInt): byte;
begin
  if value < 0 then Result := 0
  else if value > 255 then Result := 255
  else Result := value;
end;
//---------------------------------------------------------
function ClampByte (value: longWord): byte;
begin
  if value > 255 then Result := 255
  else Result := value;
end;
//---------------------------------------------------------
// return value between 0 and 65535
//---------------------------------------------------------
function ClampWord (value: longInt): word;
begin
  if value < 0 then Result := 0
  else if value > maxWord then Result := maxWord
  else Result := value;
end;
//---------------------------------------------------------
function ClampWord (value: longWord): word;
begin
  if value > maxWord then Result := maxWord
  else Result := value;
end;

//*********************************************************
// swap values
//---------------------------------------------------------
{$ifdef useAsm}
procedure Swap (var a,b: byte);   assembler;
asm
  mov  cl, byte Ptr [EAX]
  xchg cl, byte Ptr [EDX]
  mov  byte Ptr [EAX], cl
end;
//---------------------------------------------------------
procedure Swap (var a,b: shortInt);   assembler;
asm
  mov  cl, byte Ptr [EAX]
  xchg cl, byte Ptr [EDX]
  mov  byte Ptr [EAX], cl
end;
//---------------------------------------------------------
procedure Swap (var a,b: word);   assembler;
asm
  mov  cx, word Ptr [EAX]
  xchg cx, word Ptr [EDX]
  mov  word Ptr [EAX], cx
end;
//---------------------------------------------------------
procedure Swap (var a,b: smallInt);   assembler;
asm
  mov  cx, word Ptr [EAX]
  xchg cx, word Ptr [EDX]
  mov  word Ptr [EAX], cx
end;
//---------------------------------------------------------
procedure Swap (var a,b: longInt);   assembler;
asm
  mov  ecx, [EAX];   xchg ecx, [EDX];   mov  [EAX], ecx
end;
//---------------------------------------------------------
procedure Swap (var a,b: cardinal);   assembler;
asm
  mov  ecx, [EAX];   xchg ecx, [EDX];   mov  [EAX], ecx
end;
//---------------------------------------------------------
procedure Swap (var a,b: pointer);   assembler;
asm
  mov  ecx, [EAX];   xchg ecx, [EDX];   mov  [EAX], ecx
end;

{$else}
//---------------------------------------------------------
procedure Swap (var a,b: byte);
var   c: byte;
begin
  c := a;   a := b;   b := c;
end;
//---------------------------------------------------------
procedure Swap (var a,b: shortInt);
var   c: shortInt;
begin
  c := a;   a := b;   b := c;
end;
//---------------------------------------------------------
procedure Swap (var a,b: word);
var   c: word;
begin
  c := a;   a := b;   b := c;
end;
//---------------------------------------------------------
procedure Swap (var a,b: smallInt);
var   c: smallInt;
begin
  c := a;   a := b;   b := c;
end;
//---------------------------------------------------------
procedure Swap (var a,b: longInt);
var   c: longInt;
begin
  c := a;   a := b;   b := c;
end;
//---------------------------------------------------------
procedure Swap (var a,b: cardinal);
var   c: cardinal;
begin
  c := a;   a := b;   b := c;
end;
//---------------------------------------------------------
procedure Swap (var a,b: pointer);
var   c: pointer;
begin
  c := a;   a := b;   b := c;
end;
{$endif}
//---------------------------------------------------------
procedure Swap (var a,b: int64);
var   c: int64;
begin
  c := a;   a := b;   b := c;
end;

//---------------------------------------------------------
// fast swap routine that swaps the contents of any 2 variables.
// The variables may be of any type but the size of the VARS
// must be passed in Len
// eg.  X1,X2 : integer;    SwapMem(X1,X2,SizeOf(Integer));
//---------------------------------------------------------
procedure SwapMem (var source, dest; len: integer);
begin
  asm
    push edi
    push esi
    mov esi,Source
    mov edi,Dest
    mov ecx,Len
    cld
  @1:
    mov al,[edi]
    xchg [esi],al
    inc si
    stosb
    loop @1
    pop esi
    pop edi
  end;
end;

//=========================================================
// Sort: return values in ascendent order (a <= b)
//=========================================================

//---------------------------------------------------------
procedure Sort (var a,b: smallint);
begin
  if a > b then Swap (a,b);
end;
//---------------------------------------------------------
procedure Sort (var a,b: int64);
begin
  if a > b then Swap (a,b);
end;
{$ifdef useAsm}
//---------------------------------------------------------
procedure Sort (var a,b: byte);   assembler;
asm
  mov  cl, byte Ptr [EAX]
  cmp  cl, byte Ptr [EDX]
  jle @ret
  xchg cl, byte Ptr [EDX]
  mov  byte Ptr [EAX], cl
  @ret:
end;
//---------------------------------------------------------
procedure Sort (var a,b: word);   assembler;
asm
  mov  cx, word Ptr [EAX]
  cmp  cx, word Ptr [EDX]
  jle @ret
  xchg cx, word Ptr [EDX]
  mov  word Ptr [EAX], cx
  @ret:
end;
//---------------------------------------------------------
procedure Sort (var a,b: longInt);   assembler;
asm                          // EAX = a;  EDX = b
  mov  ecx, [EAX]            // ECX = a
  cmp  ecx, [EDX]            // a <> b
  jle @ret                   // a <= b ->  exit
  xchg ecx, [EDX]            // a <-> b
  mov  [EAX], ecx            // a = ECX
  @ret:
end;
//---------------------------------------------------------
procedure Sort (var a,b: longWord);   assembler;
asm                          // EAX = a;  EDX = b
  mov  ecx, [EAX]            // ECX = a
  cmp  ecx, [EDX]            // a <> b
  jle @ret                   // a <= b ->  exit
  xchg ecx, [EDX]            // a <-> b
  mov  [EAX], ecx            // a = ECX
  @ret:
end;
{$else}
//---------------------------------------------------------
procedure Sort (var a,b: byte);
begin
  if a > b then Swap (a,b);
end;
//---------------------------------------------------------
procedure Sort (var a,b: word);
begin
  if a > b then Swap (a,b);
end;
//---------------------------------------------------------
procedure Sort (var a,b: longWord);
begin
  if a > b then Swap (a,b);
end;
//---------------------------------------------------------
procedure Sort (var a,b: longInt);
begin
  if a > b then Swap (a,b);
end;
{$endif}

//=========================================================
// convertions
//=========================================================

const hexChar: array[0..15] of char = '0123456789ABCDEF';

//*********************************************************
// return value as hexadecimal string
//---------------------------------------------------------
function HexString (value: byte): string;
begin
  HexString := hexChar[value shr 4] + hexChar[value and 15];
end;
//---------------------------------------------------------
function HexString (value: word): string;
begin
  HexString := HexString (byte(value shr 8))
             + HexString (byte(value));
end;
//---------------------------------------------------------
function HexString (value: longWord): string;
begin
  HexString := HexString (word(value shr 16))
             + HexString (word(value));
end;
//---------------------------------------------------------
function HexString (value: int64): string;
begin
  HexString := HexString (longWord(value shr 32))
             + HexString (longWord(value));
end;
//---------------------------------------------------------
function HexString (value: pointer): string;
begin
  HexString := HexString (word(longWord(value) shr 16)) + ':'
             + HexString (word(value));
end;
//---------------------------------------------------------
function HexString (value: tProcedure): string;
begin
  HexString := HexString (@value);
end;
//*********************************************************
// convert hexadecimal string to long integer
//---------------------------------------------------------
function HexToLong (text: string): int64;
var   i1: word;
      hv: int64; //longword;
begin
  hv := 0;
  For i1 := 1 to length (text)
  do hv := hv shl 4 + pos (upCase (text[i1]), hexChar) - 1;
  Result := hv;
end;

//*********************************************************
// return on or off string
//---------------------------------------------------------
function OnOff (flag: boolean): string;
begin
  if flag then result := 'on' else result := 'off';
end;

//=========================================================
// sound handling
//=========================================================

//*********************************************************
// play sound with given frequency
//---------------------------------------------------------
procedure Sound (Hz: word);
begin
  if Win32PlatForm = VER_PLATFORM_WIN32_NT
  then begin
    windows.Beep (Hz, 0);
  end
  else begin
    if Hz < 19 then Hz := 19;
    SetPort ($61, GetPort ($61) or 3);      // enable speaker
    hz := word (1193181 div longInt(hz));   // 18.206... * 65536 = 1193180
    SetPort ($43, $B6);      // set timer 2 to square wave generator
    SetPort ($42, Lo(hz));   // set frequence
    SetPort ($42, Hi(hz));
  end;
end;
//*********************************************************
// turn off sound
//---------------------------------------------------------
procedure NoSound;
begin
  if Win32PlatForm = VER_PLATFORM_WIN32_NT
  then windows.Beep (0,0)
  else SetPort ($61, GetPort($61) and $FC);
end;
//*********************************************************
// play wave sound given by name without waiting
// example:    PlayWaveSound ('Utopia Windows starten.WAV');
//---------------------------------------------------------
procedure PlayWaveSound (soundFile: string);
var sf: pchar;
begin
  sf := pChar(soundFile);
  sndPlaySound (sf, SND_ASYNC or SND_FILENAME);
end;

//=========================================================
// time handling
//=========================================================

//*********************************************************
// return actual system time values
//---------------------------------------------------------
procedure GetTime (var Hour, Min, Sec, MSec: word);
var   systemTime: TSystemTime;
begin
  Windows.GetLocalTime (systemTime);
  hour := systemTime.wHour;
  min  := systemTime.wMinute;
  sec  := systemTime.wSecond;
  mSec := systemTime.wMilliSeconds;
end;
//*********************************************************
// return actual system date values
//---------------------------------------------------------
procedure GetDate (var Year, Month, Day: word);
var   systemTime: TSystemTime;
begin
  Windows.GetLocalTime(SystemTime);
  Year  := systemTime.wYear;
  Month := systemTime.wMonth;
  Day   := systemTime.wDay;
end;

//*********************************************************
// return actual system time as string [hh:mm:ss]
// example:    timeStr := SystemTimeString;
//---------------------------------------------------------
function SystemTimeString: string;
var   SystemTime: TSystemTime;
begin
  Windows.GetLocalTime(SystemTime);
  with systemTime
  do Result := Format('%d:%2.2d:%-2.2d',[wHour,wMinute,wSecond]);
end;
//*********************************************************
// return actual system date as string [dd.mm.yyyy]
// example:    dateStr := SystemDateString;
//---------------------------------------------------------
function SystemDateString: string;
var   SystemTime: TSystemTime;
begin
  Windows.GetLocalTime(SystemTime);
  with systemTime
  do Result := Format('%2d.%2.2d.%4.4d',[wDay,wMonth,wYear]);
end;

//=== MSec time measuring 1 ===

var   t1,t2: longWord;
//*********************************************************
// reset picture counter and mark actual time
//---------------------------------------------------------
procedure MarkTime;
begin
  picCnt := 0;
  t1 := t2;
  t2 := MMSystem.TimeGetTime;
end;
//*********************************************************
// return runtime between last 2 MarkTime calls in milliseconds
//---------------------------------------------------------
function RunTimeMS: longWord;
begin
  result := t2 - t1;
  if result <= 0
  then result := 1;
end;
//*********************************************************
// return runtime since last MarkTime call in milliseconds
//---------------------------------------------------------
function GetRuntimeMS: longWord;
begin
  result := MMSystem.TimeGetTime - t2;
  if result <= 0
  then result := 1;
end;
//*********************************************************
// return frames per second
// note: picCnt must be incremented after picture drawing
//---------------------------------------------------------
function FPS: string;
begin
  result := IntToStr (1000 * picCnt div GetRunTimeMS);
end;

//=== MSec time measuring 2 ===

var   ticksPerSecond: int64;      // clockrate = 1193182 tps
      tt1, tt2: int64;
//---------------------------------------------------------
function InitTicker: boolean;
begin
  QueryPerformanceFrequency (ticksPerSecond);
  MarkTime2;
  InitTicker := ticksPerSecond > 0;
end;
//---------------------------------------------------------
procedure MarkTime2;
begin
  QueryPerformanceCounter (tt2);
  tt1 := tt2;
end;
//*********************************************************
// return runtime between last 2 MarkTime calls in milliseconds
//---------------------------------------------------------
function RunTimeMS2: longWord;
begin
  if tt2 > tt1
  then tt2 := tt2 - tt1
  else tt2 := (maxInt64 - tt1) + tt2 + 1;   // timer overflow
  result := longWord ((1000 * tt2) div ticksPerSecond);
  if result <= 0
  then result := 1;
end;
//*********************************************************
// return runtime since last MarkTime call in milliseconds
//---------------------------------------------------------
function GetRuntimeMS2: longWord;
begin
  QueryPerformanceCounter (tt2);
  result := RunTimeMS2;
end;
//*********************************************************
// return frames per second
// note: picCnt must be incremented after picture drawing
//---------------------------------------------------------
function FPS2: string;
begin
  result := IntToStr (1000 * picCnt div GetRunTimeMS2);
end;

//=========================================================
// drive and file handling
//=========================================================

//=========================================================
// OS information
//=========================================================

//*********************************************************
function WindowsVersion: string;

const opSystem: array[0..2] of string[14]
      =('Windows 3.1 ','Windows 95 ','Windows NT ');

var   OSInfo: TOSVersionInfo;
begin
  OSInfo.dwOSVersionInfoSize := sizeof(TOSVERSIONINFO);
  GetVersionEx(OSInfo);
  Result := opSystem  [OSInfo.dwPlatformID]
            + IntToStr(OSInfo.dwMajorVersion) + '.'
            + IntToStr(OSInfo.dwMinorVersion);
end;
//*********************************************************
// return CPU speed [MHz] works only for Pentium and higher
//---------------------------------------------------------
function CPUSpeed: integer; 

const  DelayTime = 100;       // measure time in ms
var    TimerHi, TimerLo: DWORD;
begin
  Sleep(1);
  try
    asm
      dw 310Fh   // rdtsc
      mov TimerLo, eax
      mov TimerHi, edx
    end;
    Sleep (DelayTime);
    asm
      dw 310Fh   // rdtsc
      sub eax, TimerLo
      sbb edx, TimerHi
      mov TimerLo, eax
      mov TimerHi, edx
    end;
    Result := 1 + TimerLo div (1000 * DelayTime);
  Finally
  End;
end;
//*********************************************************
// return CPU speed [MHz] (with priority handling)
//---------------------------------------------------------
function CPUSpeed2: integer;
var   PriorityClass, Priority : Integer;
begin
  PriorityClass := GetPriorityClass(GetCurrentProcess);
  Priority := GetThreadPriority(GetCurrentThread);

  SetPriorityClass(GetCurrentProcess, REALTIME_PRIORITY_CLASS);
  SetThreadPriority(GetCurrentThread, THREAD_PRIORITY_TIME_CRITICAL);

  Result := CPUSpeed;

  SetThreadPriority(GetCurrentThread, Priority);
  SetPriorityClass(GetCurrentProcess, PriorityClass);
end;
//*********************************************************
// return processor name
//---------------------------------------------------------
function ProcessorName: string;
var reg: TRegistry;
begin
  result := '?';
  reg := TRegistry.Create;
  try
    reg.RootKey := HKEY_LOCAL_MACHINE;
    if reg.OpenKeyReadOnly('\Hardware\Description\System\CentralProcessor\0')
    then if reg.ValueExists('ProcessorNameString')
    then result := reg.ReadString('ProcessorNameString')
    else if reg.ValueExists('Identifier')                 // Win-2000
    then result := reg.ReadString('Identifier');
  finally
    reg.free;
  end;
end;

//*********************************************************
// not useable if Win32PlatForm = VER_PLATFORM_WIN32_NT
//*********************************************************
procedure SetPort (portAddr: word; value: byte); assembler;
asm
  mov dx, portAddr
  mov al, value
  out dx, al
end;
//*********************************************************
procedure SetPortW (portAddr: word; value: word); register;
asm
  xchg ax,dx
  out dx,ax
end;
//*********************************************************
function GetPort (portAddr: word): byte; assembler;
asm
  mov dx, portAddr
  in  al, dx
end;
//*********************************************************
function GetPortW (portAddr: word): word; register;
asm
  mov dx,ax
  in  ax,dx
end;

//=========================================================
// memory information
//=========================================================

var   MemStat: TMemoryStatus;

//*********************************************************
function MemoryStatus: TMemoryStatus;
begin
  MemStat.dwLength:= SizeOf(MemStat);
  GlobalMemoryStatus (MemStat);
  Result := MemStat;
end;
//*********************************************************
function AvailableVirtualMemory: Int64;
begin   result := MemoryStatus.dwAvailVirtual;   end;

//*********************************************************
function AvailablePhysicalMemory: Int64;
begin   result := MemoryStatus.dwAvailPhys;   end;

//*********************************************************
function TotalPhysicalMemory: Int64;
begin   result := MemoryStatus.dwTotalPhys;  end;

//*********************************************************
function TotalVirtualMemory: Int64;
begin   result := MemoryStatus.dwTotalVirtual;  end;


//=========================================================
// screen information
//=========================================================

//*********************************************************
// return information about screen and active window
//---------------------------------------------------------
procedure GetDisplayInfo (var screenInfo: tDeviceInfo);
begin
  FillChar (screenInfo, sizeof(screenInfo), 0);
  with screenInfo
  do begin
    dc  := GetDc(GetActiveWindow);
    // screen resolution
    x   := GetDeviceCaps (dc, HORZRES);
    y   := GetDeviceCaps (dc, VERTRES);
    // screen size [mm]
    xSize   := GetDeviceCaps (dc, HORZSIZE);
    ySize   := GetDeviceCaps (dc, VERTSIZE);
    // bits per pixel
    pPlanes := GetDeviceCaps (dc, PLANES);
    bitsPerPixel := GetDeviceCaps (dc, BITSPIXEL) * pPlanes;
    // get aspect ratio
    xAspect := GetDeviceCaps (dc, ASPECTX);
    yAspect := GetDeviceCaps (dc, ASPECTY);

    colRes := GetDeviceCaps (dc, COLORRES);
    dVersion := GetDeviceCaps (dc, DRIVERVERSION);

//    if Win32Platform = VER_PLATFORM_WIN32_NT then
//    begin
      refreshRate := GetDeviceCaps (dc, VREFRESH);
      xDesktop := GetDeviceCaps (dc, DESKTOPHORZRES);
      yDesktop := GetDeviceCaps (dc, DESKTOPVERTRES);
//    end;

    case bitsPerPixel of
      1:   screenInfo.pixelFormat := pf1bit;
      4:   screenInfo.pixelFormat := pf4bit;
      8:   screenInfo.pixelFormat := pf8bit;
      15:  screenInfo.pixelFormat := pf15bit;
      16:  screenInfo.pixelFormat := pf16bit;
      24:  screenInfo.pixelFormat := pf24bit;
      32:  screenInfo.pixelFormat := pf32bit;
      else screenInfo.pixelFormat := pfDevice;
    end;
  end;
end;
//*********************************************************
// return size and color deep of active window as X*Y*Col
//---------------------------------------------------------
function DisplayInfoString: string;
var   devInfo: tDeviceInfo;
begin
  GetDisplayInfo (devInfo);
  with devInfo
  do result := Format('%dx%dx%d', [x,y, bitsPerPixel]);
end;

//=========================================================
// window handling
//=========================================================

//*********************************************************
// show or hide system task bar (mostly at the screen bottom)
//---------------------------------------------------------
procedure ShowTaskBar (show: boolean);
begin
  if show
  then ShowWindow(FindWindow('Shell_TrayWnd', nil), SW_RESTORE)
  else ShowWindow(FindWindow('Shell_TrayWnd', nil), SW_HIDE);
end;
//*********************************************************
// return list of existing drives
//---------------------------------------------------------
function GetDriveList: TStringList;
const bufSize = 256;
var   bufferPtr, helpPtr: PChar;
      strings: TStringList;
begin
  strings := TStringList.Create;
  strings.BeginUpdate;
  bufferPtr := nil;
  try
    GetMem (bufferPtr, bufSize);
    Strings.Clear;
    if GetLogicalDriveStrings (bufSize, bufferPtr) <> 0
    then begin
      helpPtr := bufferPtr;
      while helpPtr^ <> #0 do begin
        strings.Add (helpPtr);
        Inc (helpPtr, StrLen(helpPtr) + 1);
      end;
    end;
  finally
    strings.EndUpdate;
    FreeMem (bufferPtr, BufSize);
  end;
  Result := strings;
end;
//*********************************************************
// return volume name of given drive
//---------------------------------------------------------
function GetVolumeName (drive: string): string;
var
  VolumeName    : array[0..255] of char;
  FileSystemType: array[0..255] of char;
  SerialNum, MaxFilenameLength, Flags: DWORD;
begin
  Result := '';
  if (GetVolumeInformation(@drive,
                           VolumeName,256,
                           @SerialNum,
                           MaxFilenameLength,
                           Flags,
                           FileSystemType, 256))
  then Result := string(VolumeName);
end;
//*********************************************************
// return volume name of CD or 'not present'
//---------------------------------------------------------
function GetCDVolumeName: shortstring;
var
  AList  : TStringList;
  Counter: integer;
  cddrive: shortstring;
begin
  Result := 'not present';
  AList := GetDriveList;
  for Counter := 0 to AList.Count-1
  do if GetDriveType(PChar(Alist.Strings[Counter])) = DRIVE_CDROM
  then begin
    cddrive := Alist.Strings[Counter];
    Result := cddrive + GetVolumeName (cddrive);
  end;
end;
//*********************************************************
// get all file names of given path and extension
// example:    GetFileList (filenameList, 'Data\', '*.xyz');
//---------------------------------------------------------
procedure GetFileList (var sList: tStringList; path, ext: string);
var   sRec: TSearchRec;
begin
  if sList = nil
  then sList := tStringList.Create;
  try
    if FindFirst(path + ext, faAnyFile - faDirectory, sRec) = 0
    then repeat
      sList.Add(path + sRec.Name);
{$ifdef test2}
    writeln (tst, '  file=', path + sRec.Name);
{$endif}
    until FindNext(sRec) <> 0;
  finally
    SysUtils.FindClose(sRec);
  end;
end;
//*********************************************************
// if drive or file is write protected for writing a file,
// a substitute filename will be generated
//---------------------------------------------------------
function SubstituteFilename (filename: string): string;
begin
  result := GetEnvironmentVariable ('temp');
  if result = ''
  then result := GetEnvironmentVariable ('tmp');
  if result = ''
  then result := 'C:\' + filename
  else result := result + '\' + filename;
end;


//=========================================================
// display message and errors
//=========================================================

//*********************************************************
// Ausgeben der Fehlermeldung, da Objektzeiger = 0 ist
// (entspricht RuntimeError 222)
//---------------------------------------------------------
procedure RaiseObjectNilError (name: string);
begin
  Raise EInvalidPointer.Create (#13#10'Object pointer of ' +
    name + ' = nil'#13#10);
end;
//*********************************************************
// Ausgeben der Fehlermeldung, da Objekt nicht initialisiert ist
// (entspricht RuntimeError 210)
//---------------------------------------------------------
procedure RaiseObjectInitError (name: string);
begin
  Raise EInvalidPointer.Create (#13#10'Object of type "' +
    name + '" is not initialized'#13#10);
end;

{*********************************************************}
{ Meldung in Fenster anzeigen
{---------------------------------------------------------}
procedure ShowMessageBox (const caption, infoText: string);
begin
  Application.MessageBox (PChar(infoText), PChar(caption));
end;
{*********************************************************}
{ Meldung in Fenster anzeigen
{---------------------------------------------------------}
procedure ShowWarning (const caption, infoText: string);
begin
  if Application.MessageBox (PChar(infoText), PChar(' Warning ' + caption)
                            , mb_OkCancel) = IDCANCEL
  then Halt (errUserBreak);
end;
{*********************************************************}
{ Fehlermeldungen auf beiden Seiten ausgeben
{ Rckgabe:  Buttonwert von Application.MessageBox
{---------------------------------------------------------}
function ShowError (const caption, infoText: string): integer;
begin
{$ifopt D+}
  writeln (tst,'>>> ShowError: ', caption, #13#10'    ', infoText);
  flush (tst);
{$endif}
  result := Application.MessageBox (PChar(infoText)
            ,PChar('Error in ' + caption)
            ,MB_ABORTRETRYIGNORE + MB_DEFBUTTON2);
  case result of                       // set debug stop here !!!
    //IDCANCEL: Application.Terminate;
    IDCANCEL, IDABORT: Halt (errUserBreak);
    // IDNO, IDRETRY, IDIGNORE: ;
  end;
end;
//*********************************************************
const fileErrorMsg: array [100..107] of string
      =('Fehler beim Lesen vom Datentrger',          // 100
        'Fehler beim Schreiben auf Datentrger',      // 101
        'Dateivariable ist keiner Datei zugewiesen',  // 102
        'Datei nicht geffnet',                       // 103
        'Datei nicht fr Eingabe geffnet',           // 104
        'Datei nicht fr Ausgabe geffnet',           // 105
        'Ungltiges Eingabeformat',                   // 106
        'Datei ist bereits geffnet');                // 107

//*********************************************************
// Anzeige des Dateifehlers, wenn der Fehler ungleich hide ist.
// Delphi-Hilfe: EInOutError, IOResult-->E/A-Fehler
//---------------------------------------------------------
function ShowFileError (hide: byte; filename: string): boolean;
var   ft: string;
begin
  ShowFileError := false;
  ioError := ioResult;
  if ioError = 0 then exit;            { ===> }
  if ioError <> hide
  then begin
    str (ioError, ft);
    ft := 'error ' + ft +' at file "' + filename + '"'#13#10;
    if (ioError < 100) or (ioError > 107)
    then ft := ft + SysErrorMessage (ioError)
    else ft := ft + fileErrorMsg[ioError];
    if MessageDlg (PChar(ft), mtError, [mbOk,mbCancel], 0) = IDCANCEL
    then Halt (errUserBreak);

    ShowFileError := true;
  end;
end;

//*********************************************************
// open test file for debug output
//---------------------------------------------------------
{$ifopt D+}
procedure OpenTestFile;
var filename: string;
begin
  filename := testFile;
  Assign (tst, filename);
  {$I-} ReWrite (tst); {$I+}
  if ShowFileError (0, testFile)
  then begin
    filename := SubstituteFilename(filename);
    MessageDlg (PChar('using testfile ' + filename)
               , mtInformation, [mbOk], 0);
    Assign (tst, filename);
    {$I-}   ReWrite (tst);   {$I+}
  end;
  if ShowFileError (0, filename) then Halt (ioError);   // Start abbrechen

  writeln (tst, testfile +': Delphi Debug File  '
              , DateTimeToStr(Now), #13#10);
  Flush (tst);
end;
{$endif}
//*********************************************************
initialization

  GetDisplayInfo (screenInfo);   // get screen data

  {$ifopt D+}
    OpenTestFile;
  {$endif}

  TimeBeginPeriod (1);   // set minimum timer resolution in ms
  MarkTime;
  InitTicker;

//---------------------------------------------------------
finalization
  {$I-}
  {$ifopt D+}
    Close (tst);             {close debug file}
    ioError := ioResult;
  {$endif}
end.

