//*********************************************************
//
//  UNIT:  T i t l e                   15.10.99, 05.04.2002
//
//  title animation, coded by Gerd Platl
//
//*********************************************************
{$I comp_opt.inc}

unit Title;

{$ifopt D+}
  {$define xtest1    main functions}
  {$define xtest2    sub functions}
{$endif}

interface

//*********************************************************
implementation

uses  SysUtils, Graphics,
      SysUtil2, Graph, Math2, ColorRGB, Palette, Fonts, Values,
      Blur, GrForm;

procedure NextData; forward;

var   origBm: tBitmap = nil;

//*********************************************************
// start new animation
//---------------------------------------------------------
procedure AnimStart;
begin
  {$ifdef test1}
    writeln (tst, #13#10'Title.AnimStart: ');
  {$endif}

  with valuePtr^
  do if ValueCount = 0
  then valuePtr^.AppendValue ('attraction',  20, 220);

  SetGraphMode (pf8bit);
end;
//*********************************************************
// resize animation window
//---------------------------------------------------------
procedure AnimResize;
begin
  {$ifdef test1}
    writeln (tst, 'Title.AnimResize: ',pageX,'*',pageY);
  {$endif}

  origBm.Width  := pageX;
  origBm.Height := pageY;
  NextData;
end;

//=========================================================
// palette handling
//=========================================================

var   cPal: tPalette;
      spreadValues: array[0..2] of tColorRGBI;

//---------------------------------------------------------
procedure PalettePart (ci,int: byte);

begin
  ci := ci * 64;
  spreadValues[0].rb := 0;
  spreadValues[0].gb := 0;
  spreadValues[0].bb := 0;
  spreadValues[0].ii := ci;

  spreadValues[1].rb := byte ((int + sinB[byte(picNo)   ]) div 2);
  spreadValues[1].gb := byte ((int + sinB[byte(picNo+40)]) div 2);
  spreadValues[1].bb := byte ((int + sinB[byte(picNo+55)]) div 2);
  spreadValues[1].ii := ci+58;

  spreadValues[2].rb := 255;
  spreadValues[2].gb := 255;
  spreadValues[2].bb := 255;
  spreadValues[2].ii := ci+63;
  SetSpreadPalette (cPal, spreadValues, 3);
end;
//---------------------------------------------------------
procedure GlitterColorMap;
var   ci: integer;
begin
  GetAllPalette (cPal);
  for ci := 64 to 255
  do begin
    cPal[ci] := cPal[20 + random(20)];
    if random(6) = 0
    then cPal[ci] := tColorRGBI(RandomColor(200,255));
  end;
  SetAllPalette (cPal);
end;
//---------------------------------------------------------
procedure AnimateColorMap;
begin
 {$ifdef test2}
    writeln (tst, 'Title.AnimateColorMap: ');
 {$endif}

  PalettePart (0, 255);
//  PalettePart (1, 255);
//  PalettePart (2, 255);
//  PalettePart (3, 255);
  SetAllPalette (cPal);
  GlitterColorMap;
end;

//=========================================================
// animation
//=========================================================

const titleText
      = 'GrAnDe that''s my first interactive intro made with delphi'
      + ' from Borland Inprise ~.~.~.~ hope you like it ;-) ';

const maxParticle = 600;
      sf = 6;                  // shift factor

type  tParticle = record
        xp,yp: integer;
        dx,dy: integer;
        pixColor: byte;
      end;

var   attraction: integer = 100;
      titlePos: word = 0;
      titleWord: string;
      particleNo: integer = 0;
      particle: array [0..maxParticle] of tParticle;
      pixPtr: ^byte;
      pixCnt: integer;
      xOff, yOff, pLen, xSpeed: integer;

//*********************************************************
// add particle to moving list
//---------------------------------------------------------
procedure AddParticle (x,y: integer; col: byte);
begin
  with particle [particleNo]
  do begin
    xp := x shl sf;
    yp := y shl sf;
    pixColor := 128 + random(128);
    dx := integer(random (140)) - 70;
    dy := integer(-random (80));
  end;
  inc (particleNo);
end;
//*********************************************************
procedure SearchParticles;
var   col: byte;
begin
  while (pixCnt > 0)
  and (particleNo < maxParticle)
  do begin
    col := pixPtr^;
    if col <> iBlack
    then begin
      AddParticle (pageX - (pixCnt mod pageX),
                   pixCnt div pageX, col);
      inc (pixPtr);
      dec (pixCnt);
      if pixCnt and $1F = 0 then exit;
    end
    else begin
      inc (pixPtr);
      dec (pixCnt);
    end;
  end;
  if particleNo = 0
  then NextData;
end;
//*********************************************************
procedure MoveParticles;
var  ni: integer;
begin
  ni := 0;
  while ni < particleNo
  do with particle[ni]
  do begin
    PutPixel (xp shr sf, yp shr sf, 0);  // clear pixel
    // move pixel
    xp := xp + dx;
    yp := yp + dy;
    if (xp < 0)
    or (xp >= pageX shl sf)
    or (yp >= pageY shl sf)
    then begin
      particle[ni] := particle[particleNo-1];
      dec (particleNo);
    end
    else begin
      XorPixel (xp shr sf, yp shr sf, pixColor);
      pixColor := byte (pixColor + 1);
    end;
    inc (dy, attraction shr sf);
    inc (ni);
  end;
end;
//*********************************************************
// draw one animation picture
//---------------------------------------------------------
procedure AnimDraw;
begin
  {$ifdef test2}
    writeln (tst, 'Title.AnimDraw: ');
  {$endif}

  dec (xSpeed);
  if xSpeed > 0
  then begin
    xOff := xOff - xSpeed;
    SetBkColor (0);
    ClearViewport;
    TextColor (GetPaletteColor(63));
    DisplayText (xOff, yOff, titleWord);
    Blur4Q (bgBitmap, bgBitmap, 0,GetMaxY);
  end
  else if xSpeed <= 0
  then begin
    if xSpeed = 0
    then begin
      MarkTime;
      inc (picCnt);
      CopyBitmap (bgBitmap, origBm);
      pixCnt := pageX * pageY;
      pixPtr := GetStartPointer (origBm);
      particleNo := 0;
    end;
    SearchParticles;
    MoveParticles;
  end;
end;
//*********************************************************
procedure NextData;
var   xs, bl: integer;
begin
  xOff := pageX;
  AnimateColorMap;

  SelectBitmap ('bg', bgBitmap);
  SetFont ('Desdemona', pageY div 6, [fsBold]);

  if titlePos >= length (titletext)
  then titlePos := 0;
  bl := pos (' ', copy (titleText, titlePos, 20));
  titleWord := copy (titleText, titlePos, bl-1);
  titlePos := titlePos+bl;
  pLen := TextWidth (titleWord);
  yOff := pageY div 3;
  xs   := (pageX - pLen) div 2;
  xSpeed := 0;
  while xs < pageX
  do begin
    inc (xSpeed);
    inc (xs, xSpeed);
  end;
end;
//*********************************************************
// handle animation input
//---------------------------------------------------------
procedure AnimInput;
begin
  {$ifdef test2}
    writeln (tst, 'Title.AnimInput: ', ord(cb));
  {$endif}

  with valuePtr^ do begin
    attraction := IntValue (1);
  end;
  if animColors then AnimateColorMap;
  case cb of
    CR : NextData;
    'n': valuePtr^.RandomValues;
  end;
end;
//*********************************************************
// display animation specific help
//---------------------------------------------------------
procedure AnimHelp;
begin
  HelpText (1, '<ret>   next word');
  HelpText (2, '  n     random values');
end;

//*********************************************************
initialization

  {$ifdef InitTest}
    writeln (tst, 'Title.initialization: ');
    flush (tst);
  {$endif}

  origBm := tBitmap.Create;
  origBm.PixelFormat := pf8Bit;

  RegisterAnimation (' Welcome to GrAnDe v1.32 ',
    AnimStart, AnimResize, AnimDraw, AnimInput, AnimHelp);

//---------------------------------------------------------
finalization

  {$ifdef ExitTest}
    writeln (tst, 'Title.finalization: ');
    flush (tst);
  {$endif}

  origBm.Free;
end.

