
//*********************************************************
//
//  UNIT:     V o x e l              29.12.2000, 04.10.2002
//
//  voxel demo with loaded or generated fractal landscape
//  coded by Gerd Platl
//
//*********************************************************
{$I comp_opt.inc}

unit Voxel;

{$ifopt D+}
  {$define xtest1    main functions}
  {$define xtest2    sub functions}
  {$define xtest3    game main functions}
  {$define xtest4    game sub functions}
{$endif}

interface

//*********************************************************
implementation

uses  SysUtils, Windows, Classes, Graphics,
      SysUtil2, Graph, Images, Values, Palette, Blur, GrForm, FractMap;

procedure AnimResize; forward;

const maxDist = 1000;             // max. view distance
      maxT    = $0FFF;            // max. table length = 4095

type  tPlayer = record
        id, color: byte;
        xc,yc: word;         {position: 0..65535}
        zc: longint;         {height }
        angle: word;         {direction 0..maxT}
        speed: longint;      {moving speed}
        xSpeed, ySpeed: longint;
        score: word;
        yText: word;         {y-test position}
        xd,yd: longint;      {angle direction}
        ats: longint;        {turn speed: 0..31}
        viewY: word;         {first view line}
        viewH: word;         {view height}
      end;

type  tPointerArray     = array[0..1023] of pointer;
      tBytePointerArray = array[0..1023] of pByteArray;
      tWordPointerArray = array[0..1023] of pWordArray;

var   cost,sint: array[0..maxT] of integer; {cosin and sin tables}
      dct: array[1..maxDist] of integer;    {distance compensation table}

      heightMap:   tBitmap = nil;      // maps 2^n * 2^n
      colorMap:    tBitmap = nil;
      hMap:        tBytePointerArray;  // height map pointer list
      cMap:        tBytePointerArray;  // color map pointer list
      bMap:        tWordPointerArray;  // background map pointer list
      hMapInc:     longint;
      cMapInc:     longint;
      bMapInc:     longint;
      showMap:     boolean = false;
      viewDist:    integer = 140;      //setting higher means further sight
      viewHeight:  integer =  20;
      skyColor:    byte    =  53;
      fractalDim:  word    =  50;      //% Rauhheit
      pl1: tPlayer;
      //pl2: tPlayer;

//--- file handling ---

const picFileName  = 'Images\Voxel';   // name of picture files
      picFileExt   = 'C.BMP';          // file extension

var   picFiles     :word = 0;          // # of picture files
      picFileNo    :word = 0;          // actual picture number
      picName      :string[40] = '';   // actual file name

procedure UpdatePlayer (var player: tPlayer); forward;
procedure SelectPlayers (pno: byte); forward;

//*********************************************************
// draw vertical 2 pixel thick line for 256 color mode
// attention: no clipping !!!
//---------------------------------------------------------
procedure _DrawVLine256 (x,y1,y2: integer; col: word);
var   yi: longint;
      pixPtr: pWord;
begin
{$ifdef test5}
  writeln (tst, '_DrawVLine256: y2=',y2);

{$endif}
  inc (col, col shl 8);

//  for yi := y1 to y2 do bMap[yi]^[x] := col;

  pixPtr := addr(bMap[y1]^[x]);
  for yi := y1 to y2
  do begin
    pixPtr^ := col;
    inc (longint(pixPtr), bMapInc);
  end;
end;
{********************************************************}
procedure CreateValues;
begin
  with valuePtr^
  do if ValueCount = 0
  then begin
    AppendValue ('view height',        0, 200);
    AppendValue ('view distance',    100, maxDist);
    AppendValue ('sky color',          1, 80);
    AppendValue ('fractal dimension', 30, 70);
    valuePtr^.SetValue (1, viewHeight);
    valuePtr^.SetValue (3, skyColor);
  end;
end;

//---------------------------------------------------------
procedure MakeMap (delta: integer);
begin
{$ifdef test1}
  writeln (tst, #13#10'Voxel.MakeMap: ', delta);
{$endif}

  //### test
//  SelectBitmap (MainForm.imageBm);
//  SetBkcolor (clOlive);
//  ClearViewPort;

  valuePtr^.IncValue (4,delta);
  fractalDim := fractalDim + delta;
  GenerateMap (fractalDim);
end;

//*********************************************************
// load next map files
//---------------------------------------------------------
procedure LoadNextPicture;
begin
  Sound (444);
  if picFiles = 0
  then MainForm.FlashInfo ('no voxel files found')
  else begin
    picFileNo := 1 + picFileNo mod picFiles;
    picName := picFileName + IntToStr(picFileNo);
    LoadMap (picName);
    MainForm.FlashInfo (picName);
    AnimResize;
  end;
  MarkTime;
end;

//*********************************************************
// start new animation
//---------------------------------------------------------
procedure AnimStart;
begin
  {$ifdef test1}
    writeln (tst, #13#10'Voxel.AnimStart: ');
  {$endif}

  if heightMap = nil
  then begin
    FractMap.CreateMap (heightMap, colorMap);
    MakeMap (0);
  end;

  Randomize;
  CreateValues;
  SetGraphMode (pf8bit);

  SelectMap (heightMap, colorMap);
  SelectPlayers (1);
  mainForm.FlashInfo ('press L to load landscape');
  mainForm.FlashInfo2 ('');
end;
//*********************************************************
// calculates lookup tables
//---------------------------------------------------------
procedure InitTables;
var a: word;
begin
  {precalculate sinus and cosinus table [-256 .. +256]}
  for a := 0 to maxT
  do begin
    cost[a]:=trunc(cos(a*pi/1024)*256);
    sint[a]:=trunc(sin(a*pi/1024)*256);
  end;
end;
//*********************************************************
// calculates table with height of given distance
//---------------------------------------------------------
procedure InitDistanceHeight (yHor: longint);
var i1: word;
begin
  {precalculate distance table}
  for i1 := 1 to maxDist
  do begin
    dct[i1] := (2 * maxDist) div i1 + yHor;
    //writeln (tst, i1:4, dct[i1]: 5);
  end;
end;
//*********************************************************
// set pointer array with scanline pointer values
//---------------------------------------------------------
function CreateScanlinePointerList (var bm: tBitmap;
                                    var pa: tPointerArray): longint;
var   yi: word;
begin
{$ifdef test3}
  writeln (tst, 'Voxel.CreateScanlinePointerList: ', HexString(bm));
  writeln (tst, '   size: ', bm.width, '*', bm.Height);
{$endif}
  for yi := 0 to bm.Height - 1
  do pa[yi] := bm.ScanLine[yi];
//  writeln (tst, '  sl[0]: ', HexString(pa[0]));
  CreateScanlinePointerList := longint(pa[1]) - longint(pa[0]);
end;
//*********************************************************
procedure AnimResize;
begin
{$ifdef test1}
  writeln (tst, 'Voxel.AnimResize: ', pageX, '*', pageY);
{$endif}
  // ClearBitmap (bgBitmap);
  hMapInc := CreateScanlinePointerList (heightMap, tPointerArray(hMap));
  cMapInc := CreateScanlinePointerList (colorMap,  tPointerArray(cMap));
  bMapInc := CreateScanlinePointerList (bgBitmap,  tPointerArray(bMap));
  InitDistanceHeight (pageY div 3);
  UpdatePlayer (pl1);
//  UpdatePlayer (pl2);
end;

//---------------------------------------------------------
{draws map at xy-position with given map size
 with player and boni}
procedure DrawMapAndPlayer (x,y: word);

var   xa,ya:word;

  {draws position of a player}
  procedure DrawPlayer (var player: tPlayer);

  begin
    with player
    do begin
      {calc pixel position of map}
      GetViewMapPosition (xc,yc, xa,ya);
      SetColor (iBlack);
      Bar(x+xa-1, y+ya-1, x+xa+1, y+ya+1);
      PutPixel(x+xa, y+ya, color);
      PutPixel(x+(xa+cost[(angle+ 768)and maxT] div 130),
               y+(ya+sint[(angle+ 768)and maxT] div 130), color);
      PutPixel(x+(xa+cost[(angle+1280)and maxT] div 130),
               y+(ya+sint[(angle+1280)and maxT] div 130), color);
    end;
  end;

begin
  DrawBitmap (x,y, viewMap, bgBitmap);
  DrawPlayer (pl1);
end;

//*********************************************************
procedure ChangeSpeed (var player: tPlayer; delta: integer);
begin
  with player
  do begin
    speed := speed + delta;
    MainForm.FlashInfo ('speed ' + IntToStr(speed));
  end;
end;
//*********************************************************
const noTurn = 32;
      turnMask = $FFFF - noTurn - 1;

procedure ChangeAngle (var player: tPlayer; add: integer);
begin
  with player
  do begin
    if (ats + add) and $FFC0 = 0
    then ats := (ats + add);
    angle := (angle + (ats - noTurn) div 2) and maxT;
    xd := cost[angle];
    yd := sint[angle];
    MainForm.FlashInfo ('angle ' + IntToStr(angle));
    MainForm.FlashInfo2 ('turn ' + IntToStr(ats - noTurn));
  end;
end;
//---------------------------------------------------------
procedure InitPlayer (var player: tPlayer;
                      i,c,x,y,z,a,t, vy: word);
begin
{$ifdef test3}
  writeln (tst, 'Voxel.InitPlayer: ');
{$endif}

  with player
  do begin
    id := i;
    color := c;
    xc := x;                      {init position}
    yc := y;
    zc := z;                      {init height}
    angle := a;
    speed := 64;                  {init speed}
    score := 0;
    yText := t;
    ats := noTurn;                {init turn speed}
    ChangeAngle(player, 0);       {init direction}
    viewY := vy;                  {y start line}
    viewH := pageY - viewY - 1;   {view height}
  end;
end;
//---------------------------------------------------------
procedure UpdatePlayer (var player: tPlayer);
begin
{$ifdef test3}
  writeln (tst, 'Voxel.UpdatePlayer: ');
{$endif}

  with player
  do begin
    viewH := pageY - viewY - 1;   {view height}
  end;
end;
//---------------------------------------------------------
procedure MovePlayer (var player: tPlayer);

var   xx,yy: word;
begin
  with player
  do begin

    {move}
    xSpeed := xd * speed;
    ySpeed := yd * speed;
    xc := word(xc + xSpeed shr 8);
    yc := word(yc + ySpeed shr 8);

    {calc camera height according to height map}
    zc := zc*5 + hMap[yc shr yShift]^[xc shr xShift];
    xx := word(xc + xSpeed shr 4);
    yy := word(yc + ySpeed shr 4);
    zc := zc + hMap[yy shr yShift]^[xx shr xShift];
    xx := word(xx + xSpeed);
    yy := word(yy + ySpeed);
    zc := zc + hMap[yy shr yShift]^[xx shr xShift];
    zc := zc div 8 + viewHeight;
  end;
end;
//---------------------------------------------------------
procedure ChangeHeight (var player: tPlayer; delta: longint);
begin
  with player do begin
    valuePtr^.IncValue (1,delta);
    MainForm.FlashInfo ('height ' + IntToStr(valuePtr^.IntValue (1)));
  end;
end;
//---------------------------------------------------------
procedure CheckPlayer1Input;
var  step: integer;
begin
  with pl1 do begin
    if      shiftState = [ssShift] then step := 10
    else if shiftState = [ssCtrl]  then step := 5
    else                                step := 1;
    case vKey of
      vk_Prior: ChangeHeight(pl1, +step); {33: page up}
      vk_Next:  ChangeHeight(pl1, -step); {34: page down}
      vk_Up:    ChangeSpeed (pl1, +step); {38: forward}
      vk_Down:  ChangeSpeed (pl1, -step); {40: backward}
      vk_Left:  ChangeAngle (pl1, -step); {37}
      vk_Right: ChangeAngle (pl1, +step); {39}
      else if ats > noTurn  then ChangeAngle(pl1, -1)
        else if ats < noTurn  then ChangeAngle(pl1, +1);
    end;
  end;
end;
//---------------------------------------------------------
procedure CheckPlayer2Input;
begin
end;
//*********************************************************
procedure ChangeDeep (add: integer);
begin
  valuePtr^.IncValue (2,add);
  MainForm.FlashInfo ('deep ' + IntToStr(valuePtr^.IntValue (2)));
end;
//*********************************************************
// Clear view with shaded sky colors
//---------------------------------------------------------
procedure ClearView (yp, h: word);
var   yi: word;
begin
  SetColor (iBlue);
  for yi := 0 to h - 1
  do begin
    //skyColor := byte(yi * maxColor div h);
    // SetColor (skyColor);
    _DrawHLine256 (0, GetMaxX, yp+yi, skyColor);
  end;
end;
//*********************************************************
// calculates whole view: y=y_position, h=heigth
//---------------------------------------------------------
{$Q-}
procedure DrawView (var player: tPlayer);
var ang,xi:integer;

  //-------------------------------------------------------
  // casts a verticle ray and renders result
  //   x,y:   player position   0..maxWord
  //   dir:   eye direction     0..maxT
  //-------------------------------------------------------
  procedure DrawRay (x,y, dir: word);
  var dx,dy,p,o,d, lastY:integer;
      h2,y2:longint;    col, dd: word;
  begin

    {$ifdef test4}
      writeln (tst, '  DrawRay: ',x:6,y:6,a:6);
    {$endif}

    dx := cost[dir];
    dy := sint[dir];
    d  := 10;                        {start distance}
    dd := 1;                         {distance step}
    with player
    do begin
      lastY := viewH;
      repeat
        inc(x,dx);                   {new position}
        inc(y,dy);
        inc(d,dd);                   {new distance}
        p  := y shr yShift;          {map position}
        o  := x shr xShift;

        h2 := hMap[p]^[o] - zc;      {get height}
        y2 := dct[d] - (h2 shl 7) div d;  {calculate height}

        if y2 < 0 then y2 := 0;
        if y2 < lastY
        then begin
//#          if gouraud and (lastY-y2 > 6)
//#          then begin
//#            if col <> bonusColor
//#            then _GouraudVLineAsm (xi, viewY+y2, viewY+lastY-1, col, lastC)
//#            else _DrawVLineAsm (xi, viewY+y2, viewY+lastY-1,col);
//#          end
//#       else
          col := cMap[p]^[o];          {get color}
          _DrawVLine256 (xi, viewY+y2, viewY+lastY - 1, col);
          lastY := y2;
        end;
{
        if d = 300
        then begin
          dd := dd * 2;
          dx := dx * 2;
          dy := dy * 2;
        end;
        if d = 600
        then begin
          dd := dd * 2;
          dx := dx * 2;
          dy := dy * 2;
        end;
}
      until d >= viewDist;
    end;
  end;

const ps = 2;                          // pixel size

begin
  with player
  do begin
    ClearView (viewY, viewH);          {clear background}

    for xi := 0 to GetMaxX div ps      {fill every column}
    do begin                           {calculate ray angle with view angle}
      ang := (angle+xi+1888 {+ GetMaxX div 4}) and maxT;
      DrawRay (xc,yc,ang);             {cast the ray}
    end;

    MovePlayer (pl1);
  end;
end;
//---------------------------------------------------------
procedure SelectPlayers (pno: byte);
begin
  InitPlayer (pl1, 1, 100, 57000,10000, 100,  1024, 4, 0);
//  InitPlayer (pl2, 2, 200, 16000,16000, 100,  0, 4, 0);
end;
//*********************************************************
// switch map on and off
//---------------------------------------------------------
procedure ToggleMap;
begin
  showMap := not showMap;
  MainForm.Restart;
end;
//*********************************************************
// draw one animation picture
//---------------------------------------------------------
procedure AnimDraw;
begin
  {$ifdef test2}
    writeln (tst, 'Voxel.AnimDraw: ');
  {$endif}

  SelectBitmap ('bg', bgBitmap);
  if GetGraphMode <> pf8Bit then SetGraphMode (pf8Bit);

  DrawView (pl1);

  if showMap and (pageX > 2*viewMap.Width)
  then DrawMapAndPlayer (pageX - viewMap.Width, 0);

  if not showValues then CheckPlayer1Input;
(*
  case picNo and 1 of
    0: if not showValues then CheckPlayer1Input;
    1: if not showValues then CheckPlayer2Input;
  end;
*)
end;

//*********************************************************
// handle animation input
//---------------------------------------------------------
procedure AnimInput;
begin
  {$ifdef test2}
    writeln (tst, 'Voxel.AnimInput: ', cb, ord(cb):4, '  ss=',Byte(shiftState));
  {$endif}

  with valuePtr^ do begin
    viewHeight   :=  IntValue (1);
    viewDist     :=  IntValue (2);
    skyColor     :=  IntValue (3);
    fractalDim   :=  IntValue (4);
  end;

  case cb of
    'n': valuePtr^.RandomValues;
    'r': begin valuePtr^.Reset; SelectPlayers (1); end;
    'm': ToggleMap;
    CR:  MakeMap (0);        {new map}
    '*': MakeMap (+1);       {fractal +}
    '/': MakeMap (-1);       {fractal -}
    '+': ChangeDeep (+10);
    '-': ChangeDeep (-10);
    's': begin
           pl1.speed := 0;
           MainForm.FlashInfo ('x='+IntToStr(pl1.xc)+' y='+IntToStr(pl1.yc));
         end;
    'l': LoadNextPicture;
    'W': SaveMap (picFileName);
    'b': begin
           EndlessBitmap (heightMap);
           Blur4C (heightMap, heightMap, 0, heightMap.height - 1);
         end;

    'f': GenerateFractal (random(3));
  end;
end;

//*********************************************************
// display animation specific help
//---------------------------------------------------------
procedure AnimHelp;
var  state: string;
begin
  if showMap then state := 'on' else state := 'off';
  HelpText (1, ' <ret> create new map');
  HelpText (2, '   n   random values');
  HelpText (3, '   r   reset values');
  HelpText (4, '   l   load map');
  HelpText (5, '   m   map = ' + state);
  HelpText (6, '  +/-  view distance ' + IntToStr (viewDist));
  HelpText (7, ' page +/-  view height ' + IntToStr (viewHeight));
  HelpText (8, '   s   stop moving');
  HelpText (9, '   steer with cursor');
end;

//*********************************************************
initialization

  {$ifdef InitTest}
    writeln (tst, 'Voxel.initialization: ');
    flush (tst);
  {$endif}

  InitTables;
  picFiles := GetFileCount (picFileName, picFileExt);

  RegisterAnimation (' Voxel Flight ',
    AnimStart, AnimResize, AnimDraw, AnimInput, AnimHelp);

//----------------------------------------------------------
finalization

  {$ifdef ExitTest}
    writeln (tst, 'Voxel.finalization: ');
    flush (tst);
  {$endif}

  FractMap.DeleteMap (heightMap, colorMap);
end.

