//*******************************************************************
//
//  UNIT:     W a t e r                        20.05.2000, 04.10.2002
//
//  animated water effects, coded by Gerd Platl
//
//*******************************************************************
{$I comp_opt.inc}

unit Water;

{$ifopt D+}
  {$define test1    main functions}
  {$define test2    sub functions}
  {$define test3    water functions}
{$endif}

interface

//*********************************************************
implementation

uses  SysUtils, Windows, Graphics,
      SysUtil2, Graph, Palette, Values, Images, GrForm;

//=========================================================
// Background picture stretching
//=========================================================

const modes = 2;                       // # of effects
      frame = 40;
      waterW = 256;                    // 2^n
      waterH = 256;                    // 2^n
      waterMask = waterH - 1;          // y value mask

var   modeNo       :word = 2;          // actual demo mode
      origBm       :tBitmap = nil;     // original background picture
      backBm       :tBitmap = nil;     // transformed background picture
      water1Bm     :tBitmap = nil;     // water height bitmap 1
      water2Bm     :tBitmap = nil;     // water height bitmap 2
      water3Bm     :tBitmap = nil;     // water height bitmap 3
      cPal         :tPalette;          // water bitmap color palette
      picFileNo    :word = 0;          // actual picture number

var   scanline: array [0..2048] of pointer;

//*********************************************************
// calculate all scanline pointers
//---------------------------------------------------------
procedure CreateScanlinePointerArray (bm: tBitmap);
var   yi: integer;
begin
  for yi := 0 to bm.height - 1
  do scanline[yi] := bm.ScanLine[yi];
end;

//*********************************************************
// stretch loaded picture to oversized background picture
//---------------------------------------------------------
procedure StretchPicture;
begin
  backBm.Width  := bgBitmap.Width  + 2 * frame;
  backBm.Height := bgBitmap.Height + 2 * frame;
  StretchBitmap (origBm, backBm);
  CreateScanlinePointerArray (backBm);
end;

//*********************************************************
// load next picture from file
//---------------------------------------------------------
procedure GetNextPicture;
begin
  picFileNo := LoadDemoPicture (origBm);
  if picFileNo = 0
  then MainForm.FlashInfo ('no pictures found')
  else StretchPicture;
  MarkTime;
end;
//*********************************************************
//  calculate water palette for height values from -127 .. 127
//---------------------------------------------------------
procedure CalcHeightPalette (var pal: tPalette);
var   ci: Integer;
begin
  for ci := 0 to 255
  do with pal[ci]
  do begin
    rb := 0;
    gb := byte (ci + 128);
    bb := byte (ci + 128);
  end;
  pal[128] := cYellow;
end;

//=========================================================
// Water
//=========================================================

var   dropSize   : word    =  20;     // drop size
      dropHeight : word    =  40;     // drop height
      rainFreq   : word    =  25;     // % drops per picture
      rainDelta  : word    =   4;     // pictures per drop

{********************************************************}
procedure GetValues;
begin
  with valuePtr^ do begin
    dropSize   := IntValue (1);
    dropHeight := IntValue (2);
    rainFreq   := IntValue (3);
    rainDelta  := 400 div (rainFreq + 10);
  end;
end;
{********************************************************}
procedure CreateValues;
begin
  with valuePtr^
  do if ValueCount = 0
  then begin
    AppendValue ('drop size',   1,  40);
    AppendValue ('drop height', 1, 100);
    AppendValue ('rain drops',  1, 100);
  end;
  GetValues;
end;
//*********************************************************
procedure ClearWaves;
begin
{$ifdef test2}
  writeln (tst, 'ClearWaves: ');
{$endif}

  SetBkColor (15);
  SelectBitmap ('water1', water1Bm);   ClearBitmap (water1Bm);
  SelectBitmap ('water2', water2Bm);   ClearBitmap (water2Bm);
  SelectBitmap ('water3', water3Bm);   ClearBitmap (water3Bm);
end;
//*********************************************************
// start new animation
//---------------------------------------------------------
procedure AnimStart;
begin
  {$ifdef test1}
    writeln (tst, #13#10'Water.AnimStart: ', modeNo);
    flush (tst);
  {$endif}

  CreateValues;
  case modeNo of
    1:   SetGraphMode (pf8bit);  // draw only wave heights
    else SetGraphMode (pf16bit);
  end;
  if GetGraphMode > pf8Bit then GetNextPicture;
end;
//*********************************************************
// create / resize background bitmap
//---------------------------------------------------------
procedure AnimResize;
var  wh,ww: integer;
begin
{$ifdef test1}
  writeln (tst, 'AnimResize: ',pageX,'*',pageY);
{$endif}

  StretchPicture;
  wh := Min (waterH, bgBitmap.Height);
//  ww := Min (waterW, bgBitmap.Width);
  ww := bgBitmap.Width;
  water1Bm.Height := wh;       water1Bm.Width  := ww;
  water2Bm.Height := wh;       water2Bm.Width  := ww;
  water3Bm.Height := wh * 2;   water3Bm.Width  := ww;
  SelectBitmap ('bg', bgBitmap);
  case modeNo of
    1: SetAllPalette (cPal);
  end;
end;
//*********************************************************
// switch to next mode
//---------------------------------------------------------
procedure NextMode;
begin
  modeNo := 1 + modeNo mod modes;

{$ifdef test2}
  writeln (tst, 'NextMode: ',modeNo);
  flush (tst);
{$endif}

  SelectBitmap ('bg', bgBitmap);
  case modeNo of
    1: SetGraphMode (pf8bit);  // draw only wave heights
    else begin
      SetGraphMode (pf16bit);
      GetNextPicture;
    end;
  end;
  AnimResize;
end;

//*********************************************************
// animate water height with endless window frame:
// calculate middle height          s1:      o
// of cross figure                  s2:   o  x  o
//                                  s3:      o
//                                        p1 p2 p3
// input:  y1,y2     first and last line
//---------------------------------------------------------
procedure WaterWave (source, dest: tBitmap; y1,y2: cardinal);

var  xi, yi: cardinal;      // bitmap coordinates
     l1, l2, l3,            // line pointers
     s1, s2, s3,            // source pointer
     dp: ^shortint;         // destination pointer
     p1, p2, p3: shortint;  // source height values
     po: longint;           // result height value
begin
{$ifdef test3}
  writeln (tst, 'WaterWave: ');
{$endif}

  EndlessBitmap (source);
  EndlessBitmap (dest);
  l1 := source.ScanLine[y1];
  l2 := source.ScanLine[y1+1];
  for yi := y1+1 to y2-1
  do begin
    l3 := source.ScanLine[yi+1];
    s1 := l1;
    s2 := l2;
    s3 := l3;
    dp := dest.ScanLine[yi];
    p1 := s2^;   inc (s2);
    p2 := s2^;
    for xi := 2 to source.Width
    do begin
      inc (s1);
      inc (s2);
      inc (s3);
      inc (dp);
      p3 := s2^;
      po := (s1^ + p1 + p3 + s3^) div 2 - dp^;
      if po = 0
      then dp^ := 0
      else begin
        po := po * 63 div 64;
        if      po >  127 then po :=  127
        else if po < -127 then po := -127;
        dp^ := po;
      end;
      p1 := p2;
      p2 := p3;
    end;
    l1 := l2;
    l2 := l3;
  end;
end;
//*********************************************************
// DrawWavePicture
// input:   waterBm       water height bitmap (shortint)
//---------------------------------------------------------
procedure DrawWavePicture (waterBm: tBitmap);

var   xi,yi, lx,ly, maxX, maxY, wy: longint;
      p1: shortint;
      s1,s2,s3: ^shortint;   // water height pointer
      sp,b1: pWordArray;     // picture source line pointer
      dp: ^word;             // destination bitmap pointer
begin
{$ifdef test3}
  writeln (tst, 'DrawWavePicture: ');
{$endif}

  CopyBitmap (backBm, bgBitmap);
  SelectBitmap ('bg', bgBitmap);
  s1 := waterBm.ScanLine[0];           //         o  s1^
  s2 := waterBm.ScanLine[1];           //   P1  o o  s2^
  s3 := waterBm.ScanLine[2];           //         o  s3^
  p1 := s2^;
  inc (s2);
  b1 :=   bgBitmap.ScanLine[0];
  maxX := bgBitmap.Width - 1;
  maxY := bgBitmap.Height - 1;
  for yi := 0 to maxY - 1
  do begin
    dp := GetLinePointer (yi);         // draw pointer
    for xi := 0 to maxX
    do begin
      ly := (longint(s3^) - s1^) div 8;
      lx := (longint(s2^) - p1)  div 8;

      if (lx <> 0) or (ly <> 0)
      then begin
        // get picture source line pointer
        inc (ly, yi);
        if (ly >= 0) and (ly < maxY)
        then begin
          sp := bgBitmap.ScanLine[ly];

          // draw pixel
          inc (lx, xi);
          if (lx >= 0) and (lx <= maxX) then dp^ := sp^[lx];
        end;
      end;
      p1 := s2^;
      inc (s1);
      inc (s2);
      inc (s3);
      inc (dp);
    end;
    wy := yi mod waterH;
    s1 := waterBm.ScanLine[wy  ];
    s2 := waterBm.ScanLine[wy+1];
    s3 := waterBm.ScanLine[wy+2];
  end;
end;

//*********************************************************
// draw water height line
//---------------------------------------------------------
procedure DrawHeightLine (bm: tBitmap; color: byte);

var   x: integer;
      s1: ^shortint;
      col32: tColor;
begin
    If color and 1 = 0 then col32 := clYellow else col32 := clMagenta;
    SelectBitmap ('bg', bgBitmap);
    s1 := bm.ScanLine[pageY div 2];
    if bgBitmap.PixelFormat = pf8Bit
    then for x := 0 to pageX
    do begin
      PutPixel (x, pageY div 2 - s1^ div 4, color);
      inc (s1);
    end
    else for x := 0 to pageX
    do begin
      PutPixel (x, pageY div 2 - s1^ div 4, col32);
      inc (s1);
    end;
    for x := 0 to 255
    do begin
      PutPixel (255, x, clRed);
      PutPixel (x, 255, clRed);
      inc (s1);
    end;
  end;
//*********************************************************
// draw one animation picture
//---------------------------------------------------------
procedure AnimDraw;

var   sp,dp: tBitmap;        // source + destination pointer
      x,y: integer;
begin
  if picNo and 1 = 0
  then begin
    sp := water1Bm;
    dp := water2Bm;
  end
  else begin
    sp := water2Bm;
    dp := water1Bm;
  end;
  SelectBitmap ('sp', sp);
  x := dropSize div 2 + random(pageX - dropSize);
  y := dropSize div 2 + random(waterH - dropSize);
  if picNo mod rainDelta = 0
  then begin
     SetColor (dropHeight);
     sp.Canvas.Ellipse (x,y, x+dropSize, y+dropSize);
  end;

  // calculate new water heights
  WaterWave (sp,dp, 0,waterH - 1);
  // copy upper lines to overscan lines
  CopyLines (dp, water3Bm, 0,0, waterH);
  CopyLines (dp, water3Bm, 0,waterH, 20);

  if bgBitmap.PixelFormat = pf8Bit
  then CopyBitmap (sp, bgBitmap)    // only for testing
  else DrawWavePicture (water3Bm);

  if demoHelp
  then begin
    DrawHeightLine (sp, 127);
    DrawHeightLine (dp, 128);
  end;
end;
//*********************************************************
// handle animation input
//---------------------------------------------------------
procedure AnimInput;
begin
  {$ifdef test2}
    writeln (tst, 'Water.AnimInput: ', ord(cb));
  {$endif}

  GetValues;

  case cb of
    'r': valuePtr^.Reset;
    CR : begin   ClearWaves;   valuePtr^.RandomValues;   end;
    'e': NextMode;
    '#': GetNextPicture;
    '': begin   PrevDemoPicture;   GetNextPicture;   end;
    'c': ClearWaves;
  end;
end;
//*********************************************************
// display animation specific help
//---------------------------------------------------------
procedure AnimHelp;
{var col: Shortint;}
begin
  HelpText (1, ' <ret>  random values ');
  HelpText (2, '  e   effect '  + IntToStr(modeNo));
  HelpText (3, '  #   picture ' + IntToStr(picFileNo));
  HelpText (4, '  c   clear waves ');
{
  col := pShortInt(GetPixelPointer (water1Bm, pageX div 2, pageY div 2))^;
  HelpText (5, '  h = ' + IntToStr(col));
}
end;
//*********************************************************
initialization

  {$ifdef InitTest}
    writeln (tst, 'Water.initialization: ');
    flush (tst);
  {$endif}

  backBm := tBitmap.Create;
  backBm.PixelFormat := pf16Bit;

  water1Bm := tBitmap.Create;
  water2Bm := tBitmap.Create;
  water3Bm := tBitmap.Create;
  water1Bm.PixelFormat := pf8Bit;
  water2Bm.PixelFormat := pf8Bit;
  water3Bm.PixelFormat := pf8Bit;
  ClearWaves;
  CalcHeightPalette (cPal);
  SetBitmapPalette (water1Bm, cPal);
  SetBitmapPalette (water2Bm, cPal);
  SetBitmapPalette (water3Bm, cPal);

  RegisterAnimation (' Water Waves',
    AnimStart, AnimResize, AnimDraw, AnimInput, AnimHelp);

//----------------------------------------------------------
finalization

  {$ifdef ExitTest}
    writeln (tst, 'Water.finalization: ');
    flush (tst);
  {$endif}

  backBm.Free;
  water1Bm.Free;
  water2Bm.Free;
  water3Bm.Free;
end.

