//*******************************************************************
//
//  UNIT:     W A V Y                            11.10.98, 28.01.2002
//
//  overlapping wave animation, coded by Gerd Platl
//
//*******************************************************************
{$I comp_opt.inc}

unit Wavy;

{$ifopt D+}
  {$define xtest1    main functions}
  {$define xtest2    sub functions}
{$endif}

interface

//*********************************************************
implementation

uses  SysUtils, SysUtil2, Graphics,
      Graph, Palette, GrForm;

const fac = 8;
      dimX = 1 shl fac;  {must be 2^n !!!}
      dimY = 1 shl fac;
      maxX = dimx - 1;
      maxY = dimY - 1;
      oversizeY = maxY + 10;
      mask = dimX * dimY - 1;

var   wavBm: tBitmap = nil;

//*********************************************************
// create dimX * oversizeY / 256 2d sine wave field
//---------------------------------------------------------
procedure InitWavy;
var   xi,yi: longint;
      df: double;
      linePtr: pByteArray;
begin
  wavBm := TBitmap.Create;
  wavBm.PixelFormat := pf8Bit;
  wavBm.height := oversizeY+1;
  wavBm.width  := dimX;
  for yi := 0 to overSizeY
  do begin
    linePtr := wavBm.ScanLine[yi];
    for xi := 0 to maxX
    do begin
      df := sqrt (sqr(128-xi) + sqr(128-yi));
      linePtr^[xi] := round(sin(df*0.25)*63) + 64;
    end;
  end;
end;
//*********************************************************
// create new color palette
//---------------------------------------------------------
procedure CreateColorMap;
var   cPal: tPalette;
begin
  SetSineWavePalette (cPal, 65000, 5+Random(250)
                                 , 5+Random(250)
                                 , 5+Random(250));
  SetBitmapPalette (bgBitmap, cPal);
end;
//*********************************************************
// start new animation
//---------------------------------------------------------
procedure AnimStart;
begin
  {$ifdef test1}
    writeln (tst, #13#10'Wavy.AnimStart: ');
  {$endif}

  SetGraphMode (pf8Bit);
  CreateColorMap;
end;
//*********************************************************
// draw one animation picture
//---------------------------------------------------------
procedure AnimDraw;

var   xi,yi, d1,d2: longint;
      p0,p1,p2: ^byte;       // window position
      v1: word;
      w1,w2,wb: ^word;
begin
  {$ifdef test2}
    writeln (tst, 'Wavy.AnimDraw: ');
  {$endif}

  p0 := wavBm.scanline[oversizeY];
  d1 := round((cos(picNo*0.041)*50)+50)*dimX     // offset 1.source
       +round((sin(picNo*0.037)*50)+50);
  d2 := round((sin(picNo*0.021)*50)+50)*dimX     // offset 2.source
       +round((cos(picNo*0.053)*50)+50);

  for yi := 0 to (GetMaxY div 2)
  do begin
    p1 := pointer (longint (p0) + d1);
    p2 := pointer (longint (p0) + d2);
    w1 := GetLinePointer(yi*2);
    if (yi*2) >= GetMaxY
    then w2 := GetLinePointer(yi*2)
    else w2 := GetLinePointer(yi*2+1);
    wb := w1;
    for xi := 0 to GetMaxX div 2
    do begin
      v1 := byte(p1^ + p2^);
      wb^ := v1 + v1 shl 8;
      inc (wb);
      inc (p1);
      inc (p2);
    end;
    move (w1^, w2^, pageX);       // copy line
    d1 := (d1+dimX) and mask;     // correct offset to next line
    d2 := (d2+dimX) and mask;
  end;
end;

//*********************************************************
// handle animation input
//---------------------------------------------------------
procedure AnimInput;
begin
  {$ifdef test2}
    writeln (tst, 'Wavy.AnimInput: ', ord(cb));
  {$endif}

  case cb of
    CR:  CreateColorMap;
  end;
end;

//*********************************************************
// display animation specific help
//---------------------------------------------------------
procedure AnimHelp;
begin
  HelpText (1, '<ret>  random colors');
end;
//*********************************************************
initialization

  {$ifdef InitTest}
    writeln (tst, 'Wavy.initialization: ');
    flush (tst);
  {$endif}

  InitWavy;

  RegisterAnimation (' Overlapping Sinus Waves ',
    AnimStart, nil, AnimDraw, AnimInput, AnimHelp);

//---------------------------------------------------------
finalization

  {$ifdef ExitTest}
    writeln (tst, 'Wavy.finalization: ');
    flush (tst);
  {$endif}

  wavBm.Free;
end.

