//*******************************************************************
//
//  UNIT:     B u m p L i                        06.08.99, 12.10.2001
//
//  bump lightning demo coded by Gerd Platl
//
//*******************************************************************
{$I 'comp_opt.inc'}

unit BumpLi;

{$ifopt D+}
  {$define test1    main functions}
  {$define test2    sub functions}
{$endif}

interface

//*********************************************************
implementation

uses  SysUtils, Windows, Graphics,
      SysUtil2, Graph, ColorRGB, Palette, Images, Values,
      GrForm, FractMap, Dance, Blur;

procedure AnimResize; forward;

const modes  = 3;            // # of lightning modes
      lights = 4;            // # of lightning sources
      minLI  = 33;           // minimum of light intensity
      maxLI  = 255;          // maximum of light intensity

var   modeNo:      word    =  1;       // actual demo mode
      lightNo:     word    =  1;       // actual light source
      moveSpeed:   integer = 30;       // moving speed
      moveRadius:  integer = 70;       // moving radius
      lightInt:    integer = 16;       // light intensity
      roughness:   integer = 55;       // map roughness: 1..100

      heightMap:   tBitmap = nil;      // 2^n maps
      colorMap:    tBitmap = nil;

      heightBm:    tBitmap = nil;      // height bitmap (0..255)
      colorBm:     tBitmap = nil;      // color index bitmap (0..255)
      colIntBm:    tBitmap = nil;      // intensity map
      colIntPtr:   array[0..255] of pLongIntArray;   //color array (0..255)

      demoValues:  tValues;

{********************************************************}
procedure CreateValues;
begin
  demoValues := tValues.Create;
  with demoValues
  do begin
    Init;
    Add ('moving speed',      10,  70);
    Add ('moving radius',     20, 160);
    Add ('light intensity',    1,  32);
    Add ('fractal dimension', 40,  70);
  end;
end;

//*********************************************************
// start new animation
//---------------------------------------------------------
procedure AnimStart;
begin
  {$ifdef test1}
    writeln (tst, #13#10'BumpLi.AnimStart: ');
  {$endif}

  SetGraphMode (pf32bit);

  modeNo := 1 + modeNo mod modes;
  GenerateMap (roughness);

  SelectMap (heightMap, colorMap);
  AnimResize;
end;

//*********************************************************
// create intensity color map of palette color
//---------------------------------------------------------
procedure ColorIntensityMap;
var   xi,yi: longint;
      palCol: tCol;
begin
  // create intensity map
  if colIntBm <> nil then exit;

  colIntBm := tBitmap.Create;
  colIntBm.PixelFormat := pf32Bit;
  colIntBm.Width  := 256;
  colIntBm.Height := 256;

  DebugBitmap (colIntBm, 'color intesity');

  // calculate values
  for yi := 0 to 255
  do begin
    palCol := tCol(GetPaletteColor(yi));
    for xi := 0 to 255
    do with palCol
    do colIntBm.Canvas.Pixels[xi,yi] := RGB(rb * xi shr 8
                                           ,gb * xi shr 8
                                           ,bb * xi shr 8);
  end;

  // calculate scanline pointer
  for yi := 0 to 255
  do colIntPtr[yi] := colIntBm.ScanLine[255];
end;
//*********************************************************
// create / resize height and color bitmaps
//---------------------------------------------------------
procedure AnimResize;
var hpal: tpalette;
begin
{$ifdef test1}
  writeln (tst, 'ResizeBumpLightning: ', pageX, '*', pageY);
{$endif}

  // create heightBm with window size
  if heightBm = nil
  then begin
    heightBm := tBitmap.Create;
    heightBm.PixelFormat := pf8Bit;
  end;
  heightBm.Width  := pageX;
  heightBm.Height := pageY;
  SetIntensityPalette (hPal);
  SetBitmapPalette (heightBm, hPal);
  DrawTilesBackground (0,0, heightMap, heightBm);

  // create colorBm with window size
  if colorBm = nil
  then begin
    colorBm := tBitmap.Create;
    colorBm.PixelFormat := pf8Bit;
  end;
  colorBm.Width  := pageX;
  colorBm.Height := pageY;
//  InitLandscapeColors (hPal);  // create landscape palette
  SetBitmapPalette (colorBm, hPal);
  DrawTilesBackground (0,0, colorMap, colorBm);
end;
//*********************************************************
// Bump-Lightning Animation 0
// copy fractal map to dm bitmap
// input:      sm: height bitmap  n*n*256
//             cm: color bitmap   n*n*256
//             dm: destination bitmap (24bit)
//---------------------------------------------------------
procedure BL0Anim (var sm,cm,dm: tBitmap);
begin
  if GetGraphmode <> pf8bit then SetGraphMode (pf8bit);
  sm.transparentcolor := 0;
//  DrawTilesBackground (0,0, sm, dm);
  CopyBitmap (sm, dm);
end;
//*********************************************************
// Bump-Lightning Animation 1
// uses only directional light
// input:      sm: height bitmap  n*n*256
//             cm: color bitmap   n*n*256
//             dm: destination bitmap (24bit)
//---------------------------------------------------------
{$Q-,R-}
procedure BL1Anim (var sm,cm,dm: tBitmap);
var
    lx,ly,                   // light position
    xi,yi,                   // pixel position
    li: longint;             // light intensity
    hLeft,hMid,hRight: byte; // actual line heights
    h1, h2, h3: ^byte;       // height line pointer
    cp: ^Byte;               // color index pointer
    rgbPtr: pColor;          // color pixel pointer
begin
{$ifdef test2}
  writeln (tst, ' BL1Anim:');
{$endif}

  // calculate light position
  lx := round(cos(picNo * moveSpeed / 256) * lightInt);
  ly := round(sin(picNo * moveSpeed / 256) * lightInt);

  for yi := GetMaxY downto 2
  do begin
    h1 := sm.ScanLine[byte(yi)];        // set height line pointer
    h2 := sm.ScanLine[byte(yi-1)];
    h3 := sm.ScanLine[byte(yi-2)];
    hLeft  := h2^;
    hMid   := h2^;
    hRight := h2^;
    cp     := cm.ScanLine[yi];
    rgbPtr := dm.ScanLine[yi - 1];
    for xi := 0 to GetMaxX
    do begin
      // calculate light intensity
      li := 160 - ly*(h1^ - h3^)
                - lx*(hRight - hLeft);
      if      li > maxLI then li := maxLI
      else if li < minLI then li := minLI;

      // set pixel color to intensity palette color
      rgbPtr^ := colIntPtr[cp^]^[li];

      inc (rgbPtr);
      inc (h1);              // inc height pointer
      inc (h2);
      inc (h3);
      inc (cp);

      hLeft  := hMid;        // shift pixel value
      hMid   := hRight;
      hRight := h2^;
    end;
  end;

  // draw light position
  if demoHelp
  then PutPixel (pageX div 2 + lx, pageY div 2 - ly, clYellow);
end;

//*********************************************************
// Bump-Lightning Animation 2
// uses online calculated light intensity
// input:      sm: height bitmap  n*n*256
//             cm: color bitmap   n*n*256
//             dm: destination bitmap (24bit)
//---------------------------------------------------------
procedure BL2Anim (var sm,cm,dm: tBitmap);
var
    lx,ly,                   // light position
    vx,vy,                   // distance to light
    xi,yi,                   // pixel position
    li: integer;             // light intensity
    hLeft,hMid,hRight: byte; // actual line heights
    h1, h2, h3: ^byte;       // height line pointer
    cp: ^byte;               // color index pointer
    rgbPtr: pColor;          // color pixel pointer
begin
{$ifdef test2}
  writeln (tst, ' BL2Anim:');
  DebugBitmap (sm, 'height');
  DebugBitmap (cm, 'color');
  DebugBitmap (sm, 'destination');
  flush (tst);
{$endif}

  if GetGraphmode <> pf32bit then SetGraphMode (pf32bit);

  // calculate light position
  lx := pageX div 2 + round(cos(picNo*moveSpeed / 256) * moveRadius);
  ly := pageY div 2 + round(sin(picNo*moveSpeed / 200) * moveRadius);

  vy := -ly;                 // init y-distance to light

  for yi := GetMaxY downto 2
  do begin
    h1 := sm.ScanLine[byte(yi)];        // set height line pointer
    h2 := sm.ScanLine[byte(yi-1)];
    h3 := sm.ScanLine[byte(yi-2)];
    vx := -lx;
    hLeft  := h2^;
    hMid   := h2^;
    hRight := h2^;
    cp     := cm.ScanLine[yi];
    rgbPtr := dm.ScanLine[yi - 1];
    for xi := 0 to GetMaxX
    do begin
      // calculate light intensity
      li := 255 - (abs(vy + (h1^ - h3^) shl 4)
                  +abs(vx + (hRight - hLeft) shl 4)) shl 5 div lightInt;
      if li < minLI then li := minLI;

      // set pixel color to intensity palette color
      rgbPtr^ := colIntPtr[cp^]^[li];

      inc (rgbPtr);
      inc (vx);              // update y-distance
      inc (h1);              // inc height pointer
      inc (h2);
      inc (h3);
      inc (cp);              // color index pointer

      hLeft  := hMid;        // shift pixel value
      hMid   := hRight;
      hRight := h2^;
    end;
    inc(vy);                 // update y-distance
  end;
//  CopyLine (dm, 0,GetMaxY-1, 0,GetMaxY, pageX);
  // draw light position
  // PutPixel (lx, ly, clOlive);
end;

//*********************************************************
// Bump-Lightning Animation 3
// uses precalculated intensity light map
// input:      sm: height bitmap  n*n*256
//             cm: color bitmap   n*n*256
//             dm: destination bitmap (24bit)
//---------------------------------------------------------
procedure BL3Anim (var sm,cm,dm: tBitmap);
var
    lx,ly,                   // light position
    xi,yi,                   // pixel position
    li: longint;             // light intensity
    hLeft,hMid,hRight: byte; // actual line heights
    h1, h2, h3: ^byte;       // height line pointer
    cp: ^byte;               // color index pointer
    lp: ^byte;               // light intensity pointer
    rgbPtr: pColor;          // color pixel pointer
begin
{$ifdef test2}
  writeln (tst, ' BL3Anim:');
{$endif}

  // calculate light position
  lx := round(cos(picNo*moveSpeed / 256) * moveRadius);
  ly := round(sin(picNo*moveSpeed / 201) * moveRadius);

  for yi := GetMaxY downto 2
  do begin
    h1 := sm.ScanLine[byte(yi)];        // set height line pointer
    h2 := sm.ScanLine[byte(yi-1)];
    h3 := sm.ScanLine[byte(yi-2)];
    hLeft  := h2^;
    hMid   := h2^;
    hRight := h2^;
    cp     := cm.ScanLine[yi];
    rgbPtr := dm.ScanLine[yi - 1];
    lp := LMLinePtr (yi + pageY div 2 - ly);
    inc (lp, pageX div 2 - lx);
    for xi := 0 to GetMaxX
    do begin
      // calculate light intensity
      li := lp^ * lightInt shr 5
          - 16 * ((h1^ - h3^)+(hRight - hLeft));
      if      li > maxLI then li := maxLI
      else if li < minLI then li := minLI;

      // set pixel color to intensity palette color
      rgbPtr^ := colIntPtr[cp^]^[li];
      inc (rgbPtr);
      inc (h1);              // inc height pointer
      inc (h2);
      inc (h3);
      inc (cp);              // color index pointer
      inc (lp);              // intensity map pointer

      hLeft  := hMid;        // shift pixel value
      hMid   := hRight;
      hRight := h2^;
    end;
  end;
//  CopyLine (dm, 0,GetMaxY-1, 0,GetMaxY, pageX);
end;
//*********************************************************
// Bump-Lightning Animation 4
// uses online calculated light intensity
// input:      sm: height bitmap  n*n*256
//             cm: color bitmap   n*n*256
//             dm: destination bitmap (24bit)
//---------------------------------------------------------
procedure BL4Anim (var sm,cm,dm: tBitmap);
var
    lx,ly,                   // light position
    xi,yi,                   // pixel position
    li: longint;             // light intensity
    hLeft,hMid,hRight: byte; // actual line heights
    h1, h2, h3: ^byte;       // height line pointer
    cp: ^byte;               // color index pointer
    lp: ^byte;               // light intensity pointer
    rgbPtr: pColor;          // color pixel pointer
begin
{$ifdef test2}
  writeln (tst, ' BL4Anim:');
{$endif}

  // calculate light position
  lx := pageX div 2 + round(cos(picNo*moveSpeed / 256) * moveRadius);
  ly := pageY div 2 + round(sin(picNo*moveSpeed / 200) * moveRadius);

  h1 := sm.ScanLine[GetMaxY];      // init height line pointer
  h2 := sm.ScanLine[GetMaxY - 1];
  h3 := sm.ScanLine[GetMaxY - 2];
  for yi := GetMaxY downto 1
  do begin
    hLeft  := h2^;
    hMid   := h2^;
    hRight := h2^;
    cp      := cm.ScanLine[yi];
    rgbPtr  := dm.ScanLine[yi - 1];
    lp := LMLinePtr ((yi + pageY div 2 - ly) and $FF);
    inc (lp, pageX div 2 - lx);
    for xi := 0 to GetMaxX
    do begin
      // calculate light intensity
      li := ((lp^ + (h1^ - h3^) shl 3)
            +lp^ + (hRight - hLeft) shl 3)  shl 5 div lightInt;
      if      li > maxLI then li := maxLI
      else if li < minLI then li := minLI;

      // set pixel color to intensity palette color
      rgbPtr^ := colIntPtr[cp^]^[li];
      inc (rgbPtr);
      inc (h1);              // inc source pointer
      inc (h2);
      inc (h3);
      inc (cp);              // color index pointer
      inc (lp);              // intensity map pointer

      hLeft  := hMid;        // shift pixel value
      hMid   := hRight;
      hRight := h2^;
    end;
    h1 := sm.ScanLine[yi];
    h2 := sm.ScanLine[yi+1];
    h3 := sm.ScanLine[yi+2];
  end;
//  CopyLine (dm, 0,GetMaxY-1, 0,GetMaxY, pageX);
  // draw light position
  // PutPixel (lx, ly, clOlive);
end;
//*********************************************************
// create new maps
//---------------------------------------------------------
procedure NewMap;
begin
  GenerateMap (roughness);
  AnimResize;
//  DrawViewMap (0,0);
end;
//*********************************************************
procedure FractalMap;
begin
  GenerateFractal (modeNo);
  AnimResize;
end;
//*********************************************************
procedure SmoothHeightMap;
begin
  Blur4C (heightBm, heightBm, 0,heightBM.height-1);
end;
//*********************************************************
// draw one animation picture
//---------------------------------------------------------
procedure AnimDraw;

begin
  {$ifdef test2}
    writeln (tst, 'BumpLi.AnimDraw: ');
  {$endif}

  case modeNo of
    1: BL1Anim (heightBm, colorBm, bgBitmap);
    2: BL2Anim (heightBm, colorBm, bgBitmap);
    3: BL3Anim (heightBm, colorBm, bgBitmap);
    4: BL4Anim (heightBm, colorBm, bgBitmap);
  end;
end;

//*********************************************************
// handle animation input
//---------------------------------------------------------
procedure AnimInput;
begin
  {$ifdef test2}
    writeln (tst, 'BumpLi.AnimInput: ', ord(cb));
  {$endif}

  with demoValues do begin
    moveSpeed  := IntValue (1);
    moveRadius := IntValue (2);
    lightInt   := IntValue (3);
    roughness  := IntValue (4);
  end;
  if showValues
  then begin
    demoValues.Paint;
    demoValues.KeyDown (vKey, shiftState);
  end;

  case cb of
    CR : begin
           modeNo := 1 + modeNo mod modes;
           MarkTime;
         end;
    'r': demoValues.Reset;
    'n': demoValues.RandomValues;
    '#': lightNo := UseLightmap (lightNo + 1);
    'a': FractalMap;
    'm': NewMap;
    'o': SmoothHeightMap;
    'R': begin LoadMap ('BumpLi'); AnimResize; end;
    'W': SaveMap ('BumpLi');
  end;
end;
//*********************************************************
// display animation specific help
//---------------------------------------------------------
procedure AnimHelp;
begin
  HelpText (1, ' <ret>  effect ' + IntToStr(modeNo));
  case modeNo of
    1: HelpText (2, '  directonal light:');
    2: HelpText (2, '  fix light map:');
    3: HelpText (2, '  #  light map ' + IntToStr(lightNo));
    4: HelpText (2, '  fix light map:');
  end;
  HelpText (3, '  a   apple map');
  HelpText (4, '  n   random values');
  HelpText (5, '  m   create new map');
  HelpText (6, '  o   smooth heights');
  HelpText (7, ' R/W  read/write map');
end;
//*********************************************************
initialization

  FractMap.CreateMap (heightMap, colorMap);

  {$ifdef InitTest}
    writeln (tst, 'BumpLi.initialization: ', HexString(heightMap)
                            , ',', HexString(colorMap));
  {$endif}

  ColorIntensityMap;
  CreateValues;

  RegisterAnimation (' Bump Lightning ',
    AnimStart, AnimResize, AnimDraw, AnimInput, AnimHelp);

//---------------------------------------------------------
finalization

  {$ifdef ExitTest}
    writeln (tst, 'BumpLi.finalization: ');
    flush (tst);
  {$endif}

  DeleteMap (heightMap, colorMap);
  demoValues.Free;
end.
